(******************************************************************************)
(*       _     *                                                              *)
(*      <o>    *   The Coccinelle Library / Evelyne Contejean                 *)
(*    -/@|@\-  *          CNRS-LRI-Universite Paris Sud                       *)
(*    -@ | @-  *                                                              *)
(*    -\@|@/-  *   This file is distributed under the terms of the            *)
(*      -v-    *   CeCILL-C licence                                           *)
(*             *                                                              *)
(******************************************************************************)

Add LoadPath "basis". 

(**  *Dickson Lemma: the multiset extension of a well-founded ordering is well-founded.
 *)

Set Implicit Arguments. 

Require Import Relations Wellfounded.
Require Import List.
Require Import closure.
Require Import more_list.
Require Import Multiset.
Require Import list_permut.
Require Import ordered_set.
Require Import Arith.

Module Make (DS1 : decidable_set.S).

Module DS := DS1.
Module LP := list_permut.Make (DS1).
Import LP.

(** ** Definition of the multiset extension of a relation. *)
Inductive multiset_extension_step (R : relation elt) : list elt -> list elt -> Prop :=
  | rmv_case : 
     forall l1 l2 l la a, (forall b, In b la -> R b a) -> 
      list_permut l1 (la ++ l) -> list_permut l2 (a :: l) ->
      multiset_extension_step R l1 l2.

(** [multiset_extension_step] is compatible with permutation. *)
Lemma list_permut_multiset_extension_step_1 :
  forall R l1 l2 l, list_permut l1 l2 -> 
  multiset_extension_step R l1 l -> multiset_extension_step R l2 l.
Proof.
intros R l1 l2 l P M1; inversion M1; subst.
apply (rmv_case (l1:=l2) (l2:=l) (l:=l4) R la H); trivial;
setoid_rewrite <- P; trivial.
Qed.

Lemma list_permut_multiset_extension_step_2 :
  forall R l1 l2 l, list_permut l1 l2 -> 
  multiset_extension_step R l l1 -> multiset_extension_step R l l2.
Proof.
intros R l1 l2 l P M1; inversion M1; subst.
apply (rmv_case (l1:=l) (l2:=l2) (l:=l4) R la H); trivial;
setoid_rewrite <- P; trivial.
Qed.

Add Morphism multiset_extension_step : mult_morph.
Proof.
intros R1 R2 H l1 l2 P12 l3 l4 P34; split; [intro R13 | intro R24].
apply list_permut_multiset_extension_step_2 with l3; trivial.
apply list_permut_multiset_extension_step_1 with l1; trivial.
inversion R13 as [_l1 _l3 lc la a Ha P1 P3 H1 H2]; subst.
revert P1 P3; apply rmv_case; do 2 intro; apply H; apply Ha; assumption.
apply list_permut_multiset_extension_step_2 with l4.
apply meq_sym; assumption.
apply list_permut_multiset_extension_step_1 with l2.
apply meq_sym; assumption.
inversion R24 as [_l2 _l4 lc la a Ha P2 P4 H1 H2]; subst.
revert P2 P4; apply rmv_case; do 2 intro; apply H; apply Ha; assumption.
Qed.

(** If n << {a} U m, then 
      either, there exists n' such that n = {a} U n' and n' << m,
      or, there exists k, such that n = k U m, and k << {a}. *)
Lemma two_cases :
 forall R a m n, 
 multiset_extension_step R n (a :: m) ->
 (exists n', list_permut n (a :: n') /\ 
             multiset_extension_step R n' m) \/
 (exists k, (forall b, In b k -> R b a) /\ 
            list_permut n (k ++ m)).
Proof.
intros R a m n M; inversion_clear M as [x1 x2 l la a0 H H0 H1];
elim (eq_elt_dec a a0); intro a_eq_a0; subst.

assert (H2 : list_permut m l).
setoid_rewrite (permut_cons a0); trivial.
right; exists la; split; trivial; setoid_rewrite H0; setoid_rewrite H2; auto.

left; elim (In_dec eq_elt_dec a0 m); intro In_a0.
destruct (In_split _ _ In_a0) as [m' [m'' H']]; subst m.
assert (H2 : list_permut (a :: m' ++ m'') l).
setoid_rewrite (permut_cons a0);
setoid_rewrite <- H1; pattern a at 1; do 2 rewrite app_comm_cons;
setoid_rewrite <- permut_cons_inside; auto.
exists (la ++ m' ++ m''); split.
setoid_rewrite H0; setoid_rewrite <- H2; apply list_permut_sym;
setoid_rewrite <- permut_cons_inside; auto.
apply (rmv_case (l1:=la ++ m' ++ m'') (l2:= m' ++ a0 :: m'') (l:= m' ++ m'') R la H); auto; 
apply list_permut_sym; setoid_rewrite <- permut_cons_inside; auto.

absurd (In a0 (a :: m)).
intro H2; elim H2; clear H2; intro H2; subst;
[ absurd (a0 = a0) | absurd (In a0 m) ]; trivial.
setoid_rewrite H1; left; trivial.
Qed.


(** *** Accessibility lemmata. *)
Lemma list_permut_acc :
  forall R l1 l2, list_permut l2 l1 -> 
  Acc (multiset_extension_step R) l1 -> Acc (multiset_extension_step R) l2.
Proof.
intros R l1 l2 Meq A1; apply Acc_intro; intros l M2;
inversion A1; apply H; subst;
setoid_rewrite <- Meq; trivial.
Qed.

Add Morphism (fun R => Acc (multiset_extension_step R)) : acc_morph.
Proof.
intros R1 R2 H l1 l2 P; split; [intro A1 | intro A2].

apply Acc_incl with (multiset_extension_step R1).
intros k1 k2 K; inversion K as [_k1 _k2 lc la a Ha Q1 Q2 H1 H2]; subst.
revert Q1 Q2; apply rmv_case; intros; apply H; apply Ha; assumption.
apply list_permut_acc with l1; trivial; setoid_rewrite <- P; auto.

apply Acc_incl with (multiset_extension_step R2).
intros k1 k2 K; inversion K as [_k1 _k2 lc la a Ha Q1 Q2 H1 H2]; subst.
revert Q1 Q2; apply rmv_case; intros; apply H; apply Ha; assumption.
apply list_permut_acc with l2; trivial.
Qed.

Lemma dickson_aux1 :
forall (R : relation elt) a,
 (forall b, R b a -> 
  forall m, Acc (multiset_extension_step R) m -> 
            Acc (multiset_extension_step R) (b :: m)) ->
 forall m, Acc (multiset_extension_step R) m -> 
 (forall m', (multiset_extension_step R) m' m -> 
             Acc (multiset_extension_step R) (a :: m')) ->
 Acc (multiset_extension_step R) (a :: m).
Proof. 
intros R a IH2_a m Acc_m IHa_M; apply Acc_intro;
intros n H; elim (two_cases H); clear H.
intros [n' [P M]]; refine (list_permut_acc P _); apply IHa_M; trivial.
intros [k [M P]]; refine (list_permut_acc P _); clear P; induction k; trivial; simpl;
apply IH2_a.
apply M; left; trivial.
apply IHk; intros; apply M; right; trivial.
Qed.

Lemma dickson_aux2 :
forall R m,
  Acc (multiset_extension_step R) m ->
  forall a, (forall b, R b a -> 
             forall m, Acc (multiset_extension_step R) m -> 
                       Acc (multiset_extension_step R) (b :: m)) ->
   Acc (multiset_extension_step R) (a :: m). 
Proof.
intros R m Acc_m a IH2_a;
apply (Acc_iter  (R:= multiset_extension_step R)
(fun m => Acc (multiset_extension_step R) m -> 
Acc (multiset_extension_step R) (a :: m))); trivial;
clear m Acc_m;
intros m H Acc_m; apply dickson_aux1; trivial;
intros; apply H; trivial;
apply Acc_inv with m; trivial.
Qed.

Lemma dickson_aux3 :
forall R a, Acc R a -> forall m, Acc (multiset_extension_step R) m ->
Acc (multiset_extension_step R) (a :: m).
Proof.
intros R a Acc_a;
apply (Acc_iter  (R:= R)
(fun a => Acc R a -> forall m, Acc (multiset_extension_step R) m -> 
Acc (multiset_extension_step R) (a :: m))); trivial;
clear a Acc_a;
intros a H Acc_a m Acc_m; apply dickson_aux2; trivial;
intros; apply H; trivial;
apply Acc_inv with a; trivial.
Qed.

(** Main lemma. *)
Lemma dickson : 
  forall R, well_founded R -> well_founded (multiset_extension_step R).
Proof.
intros R Wf_R; unfold well_founded in *;
intros m; induction m as [ | a m].
apply Acc_intro; intros m H; inversion_clear H;
absurd (a :: l = nil);
[ discriminate | apply list_permut_nil; setoid_rewrite H2; auto].
apply dickson_aux3; trivial.
Qed.

(** ** More results on transitive closure of mult_step *)

Lemma list_permut_trans_clos_multiset_extension_step_1 :
  forall R l1 l2 l, list_permut l1 l2 -> 
  (trans_clos (multiset_extension_step R)) l1 l -> 
  (trans_clos (multiset_extension_step R)) l2 l.
Proof.
intros R l1 l1' l P H; induction H as [ l1 l2 H | l1 l2 l3 H1 H2 H3].
apply t_step; apply list_permut_multiset_extension_step_1 with l1; trivial.
apply t_trans with l2; trivial; apply list_permut_multiset_extension_step_1 with l1; trivial.
Qed.

Lemma list_permut_trans_clos_multiset_extension_step_2 :
  forall R l1 l2 l, list_permut l1 l2 -> 
  (trans_clos (multiset_extension_step R)) l l1 -> 
  (trans_clos (multiset_extension_step R)) l l2.
Proof.
intros R l1 l3 l P H; induction H as [ l1 l2 H | l1 l2 l3' H1 H2 H3].
apply t_step; apply list_permut_multiset_extension_step_2 with l2; trivial.
apply t_trans with l2; trivial; apply H3; trivial.
Qed.

Lemma context_multiset_extension_step_app1 :
  forall R l1 l2 l, multiset_extension_step R l1 l2 ->
                         multiset_extension_step R (l ++ l1) (l ++ l2).
Proof.
intros R l1 l2 l H; destruct H as [l1 l2 l12 la a H P1 P2].
apply (@rmv_case R (l++l1) (l++l2) (l++l12) la a); trivial.
apply list_permut_trans with (l ++ la ++ l12).
setoid_rewrite <- permut_app1; trivial.
do 2 rewrite <- app_ass; setoid_rewrite <- permut_app2; trivial.
apply list_permut_app_app.
apply list_permut_trans with (l ++ a :: l12).
setoid_rewrite <- permut_app1; trivial.
apply list_permut_sym; setoid_rewrite <- permut_cons_inside; 
apply list_permut_refl.
Qed.

Lemma context_trans_clos_multiset_extension_step_app1 :
  forall R l1 l2 l, trans_clos (multiset_extension_step R) l1 l2 ->
                         trans_clos (multiset_extension_step R) (l ++ l1) (l ++ l2).
Proof.
intros R l1 l2 l H; induction H.
apply t_step; apply context_multiset_extension_step_app1; trivial.
apply t_trans with (l ++ y); trivial.
apply context_multiset_extension_step_app1; trivial.
Qed.

Lemma context_multiset_extension_step_cons :
  forall R, (forall a, ~R a a) -> 
  forall a l1 l2, multiset_extension_step R (a :: l1) (a :: l2) ->
                         multiset_extension_step R l1 l2.
Proof.
intros R irrefl_R a l1 l2 H;
inversion H as [a_l1 a_l2 lc lb b H' P1 P2 H2 H3]; subst.
destruct ((proj1 (in_permut_in a P2)) (or_introl _ (refl_equal _))) as [a_eq_b | a_in_lc].
subst b; assert (a_in_lb_lc := (proj1 (in_permut_in a P1)) (or_introl _ (refl_equal _))).
destruct (in_app_or _ _ _ a_in_lb_lc) as [a_in_lb | a_in_lc].
absurd (R a a); [apply (irrefl_R a) | apply H'; trivial].
destruct (In_split _ _ a_in_lc) as [lc' [lc'' H'']]; subst.
apply (rmv_case R (l1:=l1) (l2:= l2) (l:=lc' ++ lc'') lb (a:=a)); trivial.
setoid_rewrite (permut_cons a); setoid_rewrite P1; 
rewrite ass_app; apply list_permut_sym; 
setoid_rewrite <- permut_cons_inside; rewrite ass_app; auto.
setoid_rewrite (permut_cons a); setoid_rewrite P2;
rewrite app_comm_cons; apply list_permut_sym; 
setoid_rewrite <- permut_cons_inside;
rewrite app_comm_cons; auto.

destruct (In_split _ _ a_in_lc) as [lc' [lc'' H'']]; subst.
apply (rmv_case R (l1:=l1) (l2:= l2) (l:=lc' ++ lc'') lb (a:=b)); trivial.
setoid_rewrite (permut_cons a); setoid_rewrite P1; 
rewrite ass_app; apply list_permut_sym; 
setoid_rewrite <- permut_cons_inside; rewrite ass_app; auto.
setoid_rewrite (permut_cons a); setoid_rewrite P2;
rewrite app_comm_cons; apply list_permut_sym; 
setoid_rewrite <- permut_cons_inside;
rewrite app_comm_cons; auto.
Qed.

Lemma remove_context_multiset_extension_step_app1 :
  forall R,  (forall a, ~R a a) -> 
  forall l1 l2 l, multiset_extension_step R (l ++ l1) (l ++ l2) ->
                         multiset_extension_step R l1 l2.
Proof.
intros R irrefl_R l1 l2 l; generalize l1 l2; clear l1 l2; 
induction l as [ | a l]; trivial.
intros l1 l2 H; 
assert (H' : multiset_extension_step R (a :: l1) (a :: l2)).
apply IHl.
apply list_permut_multiset_extension_step_2 with ((a :: l) ++ l2).
simpl; setoid_rewrite <- permut_cons_inside; auto.
apply list_permut_multiset_extension_step_1 with ((a :: l) ++ l1); trivial.
simpl; setoid_rewrite <- permut_cons_inside; auto.
apply context_multiset_extension_step_cons with a; trivial.
Qed.

Lemma context_multiset_extension_step_app2 :
  forall R l1 l2 l, multiset_extension_step R l1 l2 ->
                         multiset_extension_step R (l1 ++ l) (l2 ++ l).
Proof.
intros R l1 l2 l H; destruct H as [l1 l2 l12 la a H P1 P2].
apply (@rmv_case R (l1++l) (l2++l) (l12++l) la a); trivial.
rewrite <- app_ass; setoid_rewrite <- permut_app2; trivial.
rewrite app_comm_cons; setoid_rewrite <- permut_app2; trivial.
Qed.

Function consn (ll : list (elt * list elt)) : list elt :=
  match ll with 
  | nil => nil
  | (e,_) :: ll => e:: (consn ll)
  end.

Lemma in_consn : 
  forall a ll, In a (consn ll) <-> exists la, In (a,la) ll.
Proof.
intros a ll; split; intro H;
functional induction (consn ll) 
   as [ | H1 b lb ll H2 IH].
contradiction.
destruct H as [a_eq_b | a_in_cnsl].
exists lb; subst; left; trivial.
destruct (IH a_in_cnsl) as [la H].
exists la; right; trivial.
destruct H; contradiction.
destruct H as [la [ala_eq_blb | ala_in_ll]].
injection ala_eq_blb; intros; subst; left; trivial.
right; apply IH; exists la; trivial.
Qed.

Lemma consn_app :
 forall ll1 ll2, consn (ll1 ++ ll2) = consn ll1 ++ consn ll2.
Proof.
intros ll1; induction ll1 as [ | [a la] ll1]; simpl; trivial; simpl; 
intros; rewrite IHll1; trivial.
Qed.

Function appendn (ll : list (elt * list elt)) : list elt :=
  match ll with 
  | nil => nil
  | (_,l) :: ll => l ++ (appendn ll)
  end.

Lemma appendn_app :
 forall ll1 ll2, appendn (ll1 ++ ll2) = appendn ll1 ++ appendn ll2.
Proof.
intros ll1; induction ll1 as [ | [a la] ll1]; simpl; trivial; simpl; 
intros; rewrite IHll1; rewrite ass_app; trivial.
Qed.

Lemma in_appendn : 
  forall a ll, In a (appendn ll) -> exists b, exists lb, In (b,lb) ll /\ In a lb.
Proof.
intros a ll H; 
functional induction (appendn ll) 
   as [ | H1 b lb ll H2 IH].
contradiction.
destruct (in_app_or _ _ _ H) as [a_in_lb | a_in_appl].
exists b; exists lb; split; trivial; left; trivial.
destruct (IH a_in_appl) as [c [lc [H1 H2]]]; 
exists c; exists lc; split; trivial; right; trivial.
Qed.

Lemma multiset_closure_aux :
  forall (R : relation elt) p q l pq, 
  list_permut p ((appendn l) ++ pq) ->
  list_permut q ((consn l) ++ pq) ->
  l <> nil ->
  (forall a, forall la, In (a,la) l -> forall b, In b la -> R b a) ->
  trans_clos (multiset_extension_step R) p q.
Proof.
intros R p q l; generalize p q; clear p q; induction l as [ | [x lx] l].
simpl; intros p q pq Pp Pq H; absurd (@nil (elt * list elt) = nil); trivial.
simpl; intros p q pq Pp Pq _ H; destruct l as [ | [y ly] l]; simpl in *.
rewrite <- (app_nil_end lx) in Pp; 
simpl in Pq; apply t_step; exact (rmv_case R lx (H x lx (or_introl _ (refl_equal _))) Pp Pq).
apply t_trans with (x :: ly ++ (appendn l) ++ pq).
do 2 rewrite app_ass in Pp;
refine (rmv_case R lx (H x lx (or_introl _ (refl_equal _))) Pp (list_permut_refl _)).
refine (list_permut_trans_clos_multiset_extension_step_2  
                      (list_permut_sym Pq) _).
apply (@context_trans_clos_multiset_extension_step_app1 R
            (ly ++ appendn l ++ pq) (y :: consn l ++ pq) (x :: nil)); 
apply (@IHl (ly ++ appendn l ++ pq) (y :: consn l ++ pq) pq); auto.
rewrite ass_app; auto.
discriminate.
intros a la H1 b b_in_la; apply (H a la); trivial; right; trivial.
Qed.

Lemma multiset_closure_aux2 :
  forall (R : relation elt) p q le l pq, 
  list_permut p ((appendn l) ++ pq) ->
  list_permut q (le ++ (consn l) ++ pq) ->
  l <> nil \/ le <> nil ->
  (forall a, forall la, In (a,la) l -> forall b, In b la -> R b a) ->
  trans_clos (multiset_extension_step R) p q.
Proof.
intros R p q le l pq Pp Pq H H';
apply (@multiset_closure_aux R p q ((map (fun x => (x, @nil elt)) le) ++ l) pq).
setoid_rewrite Pp; setoid_rewrite <- permut_app2;
rewrite appendn_app; clear Pq H; induction le as [ | e le]; simpl; auto.
setoid_rewrite Pq; setoid_rewrite ass_app; setoid_rewrite <- permut_app2;
rewrite consn_app; clear Pq H; induction le as [ | e le]; simpl; auto;
setoid_rewrite IHle; auto.
intro H''; destruct (app_eq_nil _ _ H'') as [ l_eq_nil le_eq_nil];
destruct H as [le_diff_nil | l_diff_nil].
absurd (l = nil); trivial.
destruct le as [ | e le].
absurd (@nil elt = nil); trivial.
discriminate.
clear Pq H; induction le as [ | e le]; simpl; trivial.
intros a la [H | H] b b_in_la; 
[injection H; intros; subst; contradiction | apply (IHle a la); trivial].
Qed.


Module LDS.
Definition A := (elt * (list elt))%type.
Lemma eq_A_dec : forall a1 a2 : A, {a1 = a2} + {a1 <> a2}.
Proof.
intros [e1 l1] [e2 l2]; destruct (eq_elt_dec e1 e2) as [e1_eq_e2 | e1_diff_e2].
destruct (list_eq_dec eq_elt_dec l1 l2) as [l1_eq_l2 | l1_diff_l2].
left; subst; trivial.
right; intro H; apply l1_diff_l2; injection H; trivial.
right; intro H; apply e1_diff_e2; injection H; trivial.
Qed.
End LDS.

Module LLP := list_permut.Make (LDS).

Lemma permut_consn :
 forall ll1 ll2, LLP.list_permut ll1 ll2 -> list_permut (consn ll1) (consn ll2).
Proof.
intros ll1; induction ll1 as [ | [a la] ll1]; simpl; intros ll2 P.
rewrite (LLP.list_permut_nil (LLP.list_permut_sym P)); simpl; auto.
assert (ala_in_ll2 : In (a,la) ll2).
apply (proj1 (LLP.in_permut_in (a,la) P)); left; trivial.
destruct (In_split _ _ ala_in_ll2) as [ll2' [ll2'' H]]; subst ll2.
setoid_rewrite <- LLP.permut_cons_inside in P.
rewrite consn_app; simpl.
setoid_rewrite <- permut_cons_inside.
rewrite <- consn_app; apply IHll1; trivial.
Qed.

Lemma permut_appendn :
 forall ll1 ll2, LLP.list_permut ll1 ll2 -> list_permut (appendn ll1) (appendn ll2).
Proof.
intros ll1; induction ll1 as [ | [a la] ll1]; simpl; intros ll2 P.
rewrite (LLP.list_permut_nil (LLP.list_permut_sym P)); simpl; auto.
assert (ala_in_ll2 : In (a,la) ll2).
apply (proj1 (LLP.in_permut_in (a,la) P)); left; trivial.
destruct (In_split _ _ ala_in_ll2) as [ll2' [ll2'' H]]; subst ll2.
setoid_rewrite <- LLP.permut_cons_inside in P.
rewrite appendn_app; simpl.
refine (list_permut_trans _ (list_permut_app_app _ _)).
rewrite <- ass_app.
setoid_rewrite <- permut_app1.
refine (list_permut_trans _ (list_permut_app_app _ _)).
rewrite <- appendn_app; apply IHll1; trivial.
Qed.

Lemma multiset_closure_aux3 :
  forall l lc cns, list_permut (consn l) (lc ++ cns) ->
               exists ll, exists ll', LLP.list_permut l (ll ++ ll') /\ 
                                           list_permut (consn ll)  lc /\ 
                                           list_permut (consn ll') cns.
Proof.
intro l; induction l as [ | [e le] l]; intros lc cns P.
exists (@nil (elt * list elt)); exists (@nil (elt * list elt)).
simpl in P; destruct (app_eq_nil _ _ (list_permut_nil (list_permut_sym P))); subst;
simpl; repeat split; auto.
assert (e_in_lc_cns : In e (lc ++ cns)).
setoid_rewrite <- P; simpl; left; trivial.
destruct (in_app_or _ _ _ e_in_lc_cns) as [e_in_lc | e_in_cns].
destruct (In_split _ _ e_in_lc) as [lc' [lc'' H]]; subst lc;
rewrite <- ass_app in P; simpl in P; setoid_rewrite <- permut_cons_inside in P;
rewrite ass_app in P;
destruct (IHl (lc' ++ lc'') cns P) as [ll [ll' [P' [H1 H2]]]].
exists ((e,le) :: ll); exists ll'; repeat split; auto.
simpl; setoid_rewrite <- LLP.permut_cons; trivial.
simpl; setoid_rewrite H1; setoid_rewrite <- permut_cons_inside; auto.
destruct (In_split _ _ e_in_cns) as [cns' [cns'' H]]; subst cns; simpl in P;
rewrite ass_app in P; simpl in P; setoid_rewrite <- permut_cons_inside in P;
rewrite <- ass_app in P.
destruct (IHl lc (cns' ++ cns'') P) as [ll [ll' [P' [H1 H2]]]].
exists ll; exists ((e,le) :: ll'); repeat split; auto.
setoid_rewrite <- LLP.permut_cons_inside; trivial.
simpl; setoid_rewrite H2; setoid_rewrite <- permut_cons_inside; auto.
Qed.

Lemma multiset_closure :
  forall R, (forall x y, {R x y}+{~R x y}) -> transitive _ R ->
  forall p q, trans_clos (multiset_extension_step R) p q ->
  exists l, exists pq,
  list_permut p ((appendn l) ++ pq) /\
  list_permut q ((consn l) ++ pq) /\
  l <> nil /\
  (forall a, forall la, In (a,la) l -> forall b, In b la -> R b a) /\
  ((forall a, ~R a a) -> forall a, In a (consn l) -> In a (appendn l) -> False).
Proof.
intros R R_dec trans_R p q p_lt_q; induction p_lt_q as [p q p_lt_q | p q r p_lt_q q_lt_r].
(* R_step *)
destruct p_lt_q as [p q pq la a la_lt_a Pp Pq].
exists ((a,la) :: nil); exists pq; simpl; repeat split; auto.
rewrite <- app_nil_end; auto.
discriminate.
intros x lx [H | H] b b_in_lx; 
[injection H; intros; subst; apply la_lt_a; trivial | contradiction].
rewrite <- app_nil_end;
intros irrefl_R b [a_eq_b | Abs] b_in_la;
 [subst b; apply (irrefl_R a); apply la_lt_a; trivial | contradiction].

(* Transitive step *)
destruct p_lt_q as [p q pq la a la_lt_a Pp Pq].
destruct IHq_lt_r as [l [qr [Pq' [Pr [l_diff_nil [app_lt_cns app_disj_cns]]]]]].
assert (a_in_appl_qr : In a ((appendn l) ++ qr)).
setoid_rewrite  <- Pq'; setoid_rewrite Pq; left; trivial.
destruct (In_dec eq_elt_dec a (appendn l)) as [a_in_appl | a_not_in_app].

destruct (in_appendn _ _ a_in_appl) as [x [lx [xlx_in_l a_in_lx]]].
destruct (In_split _ _ a_in_lx) as [lx' [lx'' H]]; subst lx;
destruct (In_split _ _  xlx_in_l) as [l' [l'' H]]; subst l.
rewrite appendn_app in Pq'; simpl in Pq'; do 3 rewrite <-  ass_app in Pq'; 
simpl in Pq'; rewrite ass_app in Pq'; setoid_rewrite Pq in Pq';
setoid_rewrite <- permut_cons_inside in Pq'; rewrite <- ass_app in Pq'.
generalize (remove_equiv_is_sound (consn (l' ++ l'')) la);
destruct (remove_equiv (consn (l' ++ l'')) la) as [cns la'];
intros [lc [Pcns [Pla cns_disj_la']]];
destruct (multiset_closure_aux3 (l' ++ l'') lc cns Pcns) as [ll [ll' [P' [H1 H2]]]];
exists ((x, lx' ++ lx'' ++ la' ++ (appendn ll)) :: ll'); exists (lc ++ qr); split.
setoid_rewrite Pp; setoid_rewrite Pla; setoid_rewrite Pq'; simpl;
do 6 rewrite ass_app; setoid_rewrite <- permut_app2;
refine (list_permut_trans _ (list_permut_app_app _ _)); do 7 rewrite <- ass_app;
setoid_rewrite <- permut_app1; rewrite ass_app;
refine (list_permut_trans (list_permut_app_app _ _) _); do 2 rewrite <- ass_app;
do 2 setoid_rewrite <- permut_app1; rewrite ass_app;
refine (list_permut_trans (list_permut_app_app _ _) _); rewrite ass_app;
refine (list_permut_trans _ (list_permut_app_app _ _)); 
setoid_rewrite <- permut_app2; do 2 rewrite <- appendn_app;
apply permut_appendn; trivial.
split.
rewrite ass_app; setoid_rewrite Pr; setoid_rewrite <- permut_app2;
rewrite consn_app; simpl; apply list_permut_sym; 
setoid_rewrite <- permut_cons_inside; setoid_rewrite H2;
refine (list_permut_trans (list_permut_app_app _ _) _);
setoid_rewrite <- Pcns; rewrite consn_app; auto.
split.
discriminate.
split.
simpl; intros y ly [yly_eq_xly | yly_in_ll''] b b_in_ly.
injection yly_eq_xly; intros; subst y ly; clear yly_eq_xly.
rewrite ass_app in b_in_ly;
destruct (in_app_or _ _ _ b_in_ly) as [b_in_lx | b_in_la_app]; clear b_in_ly.
apply (app_lt_cns x (lx' ++ a :: lx'')); trivial; apply in_insert; trivial.
destruct (in_app_or _ _ _ b_in_la_app) as [b_in_la | b_in_app].
apply trans_R with a.
apply la_lt_a; setoid_rewrite Pla; apply in_or_app; right; trivial.
apply (app_lt_cns x (lx' ++ a :: lx'')); trivial.
destruct (in_appendn _ _ b_in_app) as [y [ly [yly_in_ll b_in_ly]]].
apply trans_R with y.
apply (app_lt_cns y ly); trivial; apply in_insert;
apply (proj2 (LLP.in_permut_in (y,ly) P')); apply in_or_app; left; trivial.
assert (y_in_lc : In y lc).
setoid_rewrite <- H1; setoid_rewrite in_consn; exists ly; trivial.
apply trans_R with a.
apply la_lt_a; setoid_rewrite Pla; apply in_or_app; left; trivial.
apply (app_lt_cns x (lx' ++ a :: lx'')); trivial.
apply (app_lt_cns y ly); trivial; apply in_insert;
apply (proj2 (LLP.in_permut_in (y,ly) P')); apply in_or_app; right; trivial.
simpl; intros irrefl_R b [x_eq_b | b_in_cns] b_in_lx_la_app;
do 3 rewrite <- ass_app in b_in_lx_la_app; do 2 rewrite ass_app in b_in_lx_la_app.
subst b; destruct (in_app_or _ _ _ b_in_lx_la_app) as [b_in_lx_la | b_in_app].
destruct (in_app_or _ _ _ b_in_lx_la) as [b_in_lx | b_in_la].
apply (irrefl_R x); apply (app_lt_cns x (lx' ++ a :: lx'')); trivial; apply in_insert; trivial.
apply (irrefl_R x); apply trans_R with a.
apply la_lt_a; setoid_rewrite Pla; apply in_or_app; right; trivial.
apply (app_lt_cns x (lx' ++ a :: lx'')); trivial; apply in_or_app; right; left; trivial.
apply (app_disj_cns irrefl_R x).
rewrite consn_app; apply in_or_app; right; left; trivial.
rewrite <- appendn_app in b_in_app.
setoid_rewrite <- (permut_appendn P') in b_in_app.
rewrite appendn_app; apply in_or_app;
rewrite appendn_app in b_in_app; 
destruct (in_app_or _ _ _ b_in_app); [left | simpl; right; apply in_or_app; right]; trivial.
destruct (in_app_or _ _ _ b_in_lx_la_app) as [b_in_lx_la | b_in_app]; clear b_in_lx_la_app.
destruct (in_app_or _ _ _ b_in_lx_la) as [b_in_lx | b_in_la]; clear b_in_lx_la.
apply (app_disj_cns irrefl_R b).
rewrite consn_app; simpl; apply in_insert;
rewrite <- consn_app; setoid_rewrite (permut_consn P');
rewrite consn_app; apply in_or_app; right; trivial.
rewrite appendn_app; apply in_or_app; right;
simpl; apply in_or_app; left; apply in_insert; trivial.
apply (cns_disj_la' b); trivial; setoid_rewrite <- H2; trivial.
apply (app_disj_cns irrefl_R b).
rewrite consn_app; simpl; apply in_insert;
rewrite <- consn_app; setoid_rewrite (permut_consn P');
rewrite consn_app; apply in_or_app; right; trivial.
rewrite <- appendn_app in b_in_app;
setoid_rewrite <- (permut_appendn P') in b_in_app.
rewrite appendn_app; apply in_or_app;
rewrite appendn_app in b_in_app; 
destruct (in_app_or _ _ _ b_in_app); [left | simpl; right; apply in_or_app; right]; trivial.

destruct (in_app_or _ _ _ a_in_appl_qr) as [a_in_appl | a_in_qr]; clear a_in_appl_qr.
absurd (In a (appendn l)); trivial.
destruct (In_split _ _ a_in_qr) as [qr' [qr'' H]]; subst qr;
setoid_rewrite Pq in Pq'; rewrite ass_app in Pq';
setoid_rewrite <- permut_cons_inside in Pq'; rewrite <- ass_app in Pq'.

generalize (remove_equiv_is_sound (consn l) la);
destruct (remove_equiv (consn l) la) as [cns la'];
intros [lc [Pcns [Pla cns_disj_la']]];
destruct (multiset_closure_aux3 l lc cns Pcns) as [ll [ll' [P' [H1 H2]]]].
exists ((a, la' ++ (appendn ll)) :: ll'); exists (lc ++ (qr' ++ qr'')); split.
setoid_rewrite Pp; setoid_rewrite Pq'; simpl; do 4 rewrite ass_app;
do 2 setoid_rewrite <- permut_app2; setoid_rewrite Pla; rewrite <- ass_app;
refine (list_permut_trans (list_permut_app_app _ _) _);
setoid_rewrite <- permut_app2; rewrite <- ass_app;
rewrite <- appendn_app; setoid_rewrite <- (permut_appendn P'); auto.
split.
setoid_rewrite Pr; simpl; rewrite ass_app; apply list_permut_sym; 
setoid_rewrite <- permut_cons_inside; do 2 rewrite ass_app;
do 2 setoid_rewrite <- permut_app2; setoid_rewrite <- H1;
setoid_rewrite (permut_consn P'); rewrite consn_app;
apply list_permut_app_app.
split.
discriminate.
split.
intros y ly [yly_eq_aly | yly_in_ll'] b b_in_ly.
injection yly_eq_aly; intros; subst y ly; clear yly_eq_aly.
destruct (in_app_or _ _ _ b_in_ly) as [b_in_la' | b_in_app].
apply la_lt_a; setoid_rewrite Pla; apply in_or_app; right; trivial.
destruct (in_appendn _ _ b_in_app) as [z [lz [zlz_in_ll b_in_lz]]].
apply trans_R with z.
apply (app_lt_cns z lz); trivial;
apply (proj2 (LLP.in_permut_in (z,lz) P')); apply in_or_app; left; trivial.
apply la_lt_a; setoid_rewrite Pla; apply in_or_app; left; setoid_rewrite <- H1;
setoid_rewrite in_consn; exists lz; trivial.
apply (app_lt_cns y ly); trivial; 
apply (proj2 (LLP.in_permut_in (y,ly) P')); apply in_or_app; right; trivial.
intros irrefl_R; simpl; rewrite <- ass_app; rewrite ass_app;
intros b b_in_a_cns b_in_la_app;
destruct (in_app_or _ _ _ b_in_la_app) as [b_in_la | b_in_app].
destruct b_in_a_cns as [b_eq_a | b_in_cns].
destruct (in_app_or _ _ _ b_in_la) as [b_in_la' | b_in_app].
apply (irrefl_R a); apply la_lt_a; setoid_rewrite Pla; apply in_or_app; right; subst; trivial.
destruct (in_appendn _ _ b_in_app) as [z [lz [zlz_in_ll b_in_lz]]].
apply (irrefl_R a); apply trans_R with z.
apply (app_lt_cns z lz); subst; trivial;
apply (proj2 (LLP.in_permut_in (z,lz) P')); apply in_or_app; left; trivial.
apply la_lt_a; setoid_rewrite Pla; apply in_or_app; left; setoid_rewrite <- H1;
setoid_rewrite in_consn; exists lz; trivial.
destruct (in_app_or _ _ _ b_in_la) as [b_in_la' | b_in_app].
apply (cns_disj_la' b); trivial; setoid_rewrite <- H2; trivial.
apply (app_disj_cns irrefl_R b).
setoid_rewrite (permut_consn P'); rewrite consn_app; apply in_or_app; right; trivial.
setoid_rewrite (permut_appendn P'); rewrite appendn_app; apply in_or_app; left; trivial.
destruct b_in_a_cns as [b_eq_a | b_in_cns].
subst b; apply (a_not_in_app); setoid_rewrite (permut_appendn P');
rewrite appendn_app; apply in_or_app; right; trivial.
apply (app_disj_cns irrefl_R b).
setoid_rewrite (permut_consn P'); rewrite consn_app; apply in_or_app; right; trivial.
setoid_rewrite (permut_appendn P'); rewrite appendn_app; apply in_or_app; right; trivial.
Qed.

Lemma context_trans_clos_multiset_extension_step_cons :
  forall R, (forall x y, {R x y}+{~R x y}) -> transitive _ R -> (forall a, ~R a a) -> 
  forall a l1 l2, trans_clos (multiset_extension_step R) (a :: l1) (a :: l2) ->
                         trans_clos (multiset_extension_step R) l1 l2.
Proof.
intros R dec_R trans_R irrefl_R a l1 l2 H.
destruct (multiset_closure dec_R trans_R H)
    as [ll [q [P1 [P2 [p_diff_nil [app_lt_cns cns_disj_app]]]]]].
destruct (In_dec eq_elt_dec a q) as [a_in_q | a_not_in_q].
destruct (In_split _ _ a_in_q) as [q' [q'' H']]; subst q.
rewrite ass_app in P1; setoid_rewrite <- permut_cons_inside in P1; rewrite <- ass_app in P1.
rewrite ass_app in P2; setoid_rewrite <- permut_cons_inside in P2; rewrite <- ass_app in P2.
exact (multiset_closure_aux R (q' ++ q'') P1 P2 p_diff_nil app_lt_cns).
assert False; [idtac | contradiction].
apply (cns_disj_app irrefl_R a).
assert (a_in_cns_ll_q : In a (consn ll ++ q)).
setoid_rewrite <- (in_permut_in a P2); left; trivial.
destruct (in_app_or _ _ _ a_in_cns_ll_q) as [a_in_cns_ll | a_in_q]; 
[trivial | absurd (In a q); trivial].
assert (a_in_app_ll_q : In a (appendn ll ++ q)).
setoid_rewrite <- (in_permut_in a P1); left; trivial.
destruct (in_app_or _ _ _ a_in_app_ll_q) as [a_in_app_ll | a_in_q]; 
[trivial | absurd (In a q); trivial].
Qed.

Lemma remove_context_trans_clos_multiset_extension_step_app1 :
  forall R, (forall x y, {R x y}+{~R x y}) -> transitive _ R -> (forall a, ~R a a) -> 
  forall l1 l2 l, trans_clos (multiset_extension_step R) (l ++ l1) (l ++ l2) ->
                         trans_clos (multiset_extension_step R) l1 l2.
Proof.
intros R dec_R trans_R irrefl_R l1 l2 l; generalize l1 l2; clear l1 l2; 
induction l as [ | a l]; trivial.
intros l1 l2 H; apply IHl;
apply context_trans_clos_multiset_extension_step_cons with a; trivial.
Qed.

Definition R_or_eq_dec (R : relation elt) (R_dec :  forall e1 e2, {R e1 e2}+{~R e1 e2}) e1 e2 : comp :=
  if DS.eq_A_dec e1 e2 
  then Equivalent
  else 
     if R_dec e1 e2 
      then Less_than
      else 
        if R_dec e2 e1
        then Greater_than
        else Uncomparable.

Lemma R_or_eq_dec_is_sound :
  forall R R_dec e1 e2,
  match R_or_eq_dec R R_dec e1 e2 with
  | Equivalent => e1 = e2
  | Less_than => R e1 e2
  | Greater_than => R e2 e1
  | Uncomparable => e1 <> e2 /\ ~R e1 e2 /\ ~R e2 e1
  end.
Proof.
intros R R_dec e1 e2; unfold R_or_eq_dec;
destruct (DS.eq_A_dec e1 e2) as [e1_eq_e2 | e1_diff_e2]; trivial.
destruct (R_dec e1 e2) as [e1Re2 | not_e1Re2]; trivial.
destruct (R_dec e2 e1) as [e2Re1 | not_e2Re1]; intuition.
Qed.

Definition mult (R : relation elt) (R_dec : forall e1 e2, {R e1 e2}+{~R e1 e2}) 
  (l1 l2 : list elt) : comp :=
  match remove_equiv l1 l2 with
    | (nil, nil) => Equivalent
    | (l1, l2) => 
	match (list_forall 
	      (fun t2 => 
		 list_exists 
		   (fun t1 => 
		      match R_or_eq_dec R R_dec t1 t2 with
			| Greater_than => true
			| _ => false
                      end) l1) l2) with
        | true => Greater_than
	| false =>
	  match (list_forall 
		(fun t1 =>
		   list_exists
		     (fun t2 =>
			match R_or_eq_dec R R_dec t2 t1 with
			  | Greater_than => true
			  | _ => false
                         end) l2) l1) with
	  | true => Less_than
	  | false => Uncomparable
          end
     end
end.

Lemma nil_is_the_smallest :
  forall R e l, trans_clos (multiset_extension_step R) nil (e :: l).
Proof.
intros R e' l; generalize e'; clear e'; induction l as [ | e l].
intros e; apply t_step; refine (@rmv_case _ _ _ nil nil e _ _ _); auto; contradiction.
intros e'; apply trans_clos_is_trans with (e' :: l); trivial.
apply t_step; refine (@rmv_case _ _ _ (e' :: l) nil e _ _ _); auto.
contradiction.
setoid_rewrite <- (permut_cons_inside e' (e :: l) (e :: nil) l); auto.
Qed.

Lemma R_dec_sym : 
  forall R R_dec, transitive _ R -> (forall a, ~R a a) -> 
  forall e1 e2,   
   R_or_eq_dec R R_dec e2 e1 = match R_or_eq_dec R R_dec e1 e2 with
  | Equivalent => Equivalent
  | Less_than => Greater_than
  | Greater_than => Less_than
  | Uncomparable => Uncomparable
  end.
intros R R_dec trans_R irrefl_R e1 e2; 
generalize (R_or_eq_dec_is_sound R R_dec e1 e2);
generalize (R_or_eq_dec_is_sound R R_dec e2 e1);
destruct (R_or_eq_dec R R_dec e1 e2);
destruct (R_or_eq_dec R R_dec e2 e1); trivial.
intros; subst e2; absurd (R e1 e1); trivial; apply irrefl_R.
intros; subst e2; absurd (R e1 e1); trivial; apply irrefl_R.
intros [e2_diff_e1 _] e1_eq_e2; subst e2; absurd (e1 = e1); trivial.
intros; subst e2; absurd (R e1 e1); trivial; apply irrefl_R.
intros; absurd (R e1 e1); [apply irrefl_R | apply trans_R with e2; trivial].
intros [_ [_ not_e1_R_e2]] e1_R_e2; absurd (R e1 e2); trivial.
intros; subst e2; absurd (R e1 e1); trivial; apply irrefl_R.
intros; absurd (R e1 e1); [apply irrefl_R | apply trans_R with e2; trivial].
intros [_ [not_e2_R_e1 _]] e2_R_e1; absurd (R e2 e1); trivial.
intros  e2_eq_e1 [e1_diff_e2 _]; subst e2; absurd (e1 = e1); trivial.
intros e2_R_e1 [_ [_ not_e2_R_e1]]; absurd (R e2 e1); trivial.
intros e1_R_e2 [_ [not_e1_R_e2 _]]; absurd (R e1 e2); trivial.
Qed.

Lemma greater_case :
   forall R R_dec l1 l2,
      list_forall 
	      (fun t2 => 
		 list_exists 
		   (fun t1 => 
		      match R_or_eq_dec R R_dec t1 t2 with
			| Greater_than => true
			| _ => false
                      end) l1) l2 = true ->
   exists le, exists ll, list_permut l1 (consn ll ++ le) /\
                              list_permut l2 (appendn ll) /\
                              (forall a la, In (a,la) ll -> forall b, In b la -> R b a).
Proof. 
intros R R_dec l1 l2; generalize l1; clear l1; 
induction l2 as [ | e2 l2].
intros l1 _; exists l1; exists (@nil (elt * list elt)); simpl; intuition.
intros l1; simpl.
generalize (list_exists_is_sound (fun t1 : DS.A =>
      match R_or_eq_dec R R_dec t1 e2 with
      | Equivalent => false
      | Less_than => false
      | Greater_than => true
      | Uncomparable => false
      end) l1);
destruct (list_exists (fun t1 : DS.A =>
      match R_or_eq_dec R R_dec t1 e2 with
      | Equivalent => false
      | Less_than => false
      | Greater_than => true
      | Uncomparable => false
      end) l1).
intro H; generalize (proj1 H (refl_equal _)); clear H; 
intros [a1 [a1_in_l1 e2_R_a1]]; generalize (IHl2 l1); simpl.
destruct (list_forall
   (fun t2 : DS.A =>
    list_exists
      (fun t1 : DS.A =>
       match R_or_eq_dec R R_dec t1 t2 with
       | Equivalent => false
       | Less_than => false
       | Greater_than => true
       | Uncomparable => false
       end) l1) l2).
intro H; generalize (H (refl_equal _)); clear H;
intros [le [ll [P1 [P2 app_lt_cns]]]] _.
assert (a1_in_cns_ll_le : In a1 (consn ll ++ le)).
setoid_rewrite <- (in_permut_in a1 P1); trivial.
assert (e2_R_a1' : R e2 a1).
generalize (R_or_eq_dec_is_sound R R_dec a1 e2);
destruct (R_or_eq_dec R R_dec a1 e2); trivial;
absurd (false = true); trivial; discriminate.
destruct (in_app_or _ _ _ a1_in_cns_ll_le) as [a1_in_cns_ll | a1_in_le].
destruct (proj1 (in_consn _ _) a1_in_cns_ll) as [la ala_in_ll].
destruct (In_split _ _ ala_in_ll) as [ll' [ll'' H]]; subst ll.
exists le; exists (ll' ++ (a1, e2 :: la) :: ll''); split.
rewrite consn_app; simpl; rewrite consn_app in P1; simpl in P1; trivial.
split.
rewrite appendn_app; simpl; setoid_rewrite <- permut_cons_inside;
rewrite appendn_app in P2; simpl in P2; trivial.
intros x lx xlx_in_ll b b_in_lx;
destruct (in_app_or _ _ _ xlx_in_ll) as [xlx_ll' | [xlx_eq_a1la | xlx_in_ll'']].
apply (app_lt_cns x lx); trivial; apply in_or_app; left; trivial.
injection xlx_eq_a1la; intros; subst x lx.
destruct b_in_lx as [b_eq_e2 | b_in_la]; [subst b | apply (app_lt_cns a1 la)]; trivial.
apply (app_lt_cns x lx); trivial; apply in_or_app; do 2 right; trivial.

destruct (In_split _ _ a1_in_le) as [le' [le'' H]]; subst le.
exists (le' ++ le''); exists ((a1, e2 :: nil) :: ll); split.
simpl; setoid_rewrite P1; rewrite ass_app; apply list_permut_sym;
setoid_rewrite <- permut_cons_inside; rewrite ass_app; auto.
split.
simpl; setoid_rewrite <- permut_cons; trivial.
intros x lx [xlx_eq_a1_e2 | xlx_in_ll] b b_in_lx.
injection xlx_eq_a1_e2; intros; subst x lx;
destruct b_in_lx as [b_eq_e2 | Abs]; [subst b; trivial | contradiction].
apply (app_lt_cns x lx); trivial; right; trivial.
intros; absurd (false = true); trivial; discriminate.
intros; absurd (false = true); trivial; discriminate.
Qed.

Lemma mult_is_sound :
 forall R R_dec l1 l2,
  match mult R R_dec l1 l2 with
  | Equivalent => list_permut l1 l2
  | Less_than => trans_clos (multiset_extension_step R) l1 l2
  | Greater_than => trans_clos (multiset_extension_step R) l2 l1
  | _ => True
  end.
intros R R_dec l1 l2; unfold mult; simpl;
generalize (remove_equiv_is_sound l1 l2); 
destruct (remove_equiv l1 l2) as [l1' l2'];
intros [l [P1 [P2 H]]];
destruct l1' as [ | e1' l1'];
destruct l2' as [ | e2' l2'].
setoid_rewrite P1; setoid_rewrite P2; auto.
simpl; rewrite <- app_nil_end in P1;
apply (multiset_closure_aux2 R (p := l1) (q := l2) 
                      (le := e2' :: l2') (l := nil) l); simpl; auto.
setoid_rewrite P2; rewrite app_comm_cons; apply list_permut_app_app.
right; discriminate.
intros; contradiction.
simpl; rewrite <- app_nil_end in P2;
apply (multiset_closure_aux2 R (p := l2) (q := l1) 
                      (le := e1' :: l1') (l := nil) l); simpl; auto.
setoid_rewrite P1; rewrite app_comm_cons; apply list_permut_app_app.
right; discriminate.
intros; contradiction.
generalize (greater_case R R_dec (e1' :: l1') (e2' :: l2'));
destruct (list_forall
        (fun t2 : DS.A =>
         list_exists
           (fun t1 : DS.A =>
            match R_or_eq_dec R R_dec t1 t2 with
            | Equivalent => false
            | Less_than => false
            | Greater_than => true
            | Uncomparable => false
            end) (e1' :: l1')) (e2' :: l2')).
intro H'; generalize (H' (refl_equal _)); clear H'; intros [le [ll [P1' [P2' app_lt_cns]]]].
apply (multiset_closure_aux2 R (p := l2) (q := l1) (le := le)  (l := ll) l); simpl; auto.
setoid_rewrite P2; setoid_rewrite P2'; apply list_permut_app_app.
setoid_rewrite P1; setoid_rewrite P1';
apply list_permut_trans with (l ++ le ++ consn ll).
setoid_rewrite <- permut_app1; apply list_permut_app_app.
rewrite (ass_app le); apply list_permut_app_app.
destruct ll as [ | [x lx] ll].
destruct le as [ | e le].
generalize (list_permut_length P1'); simpl; intro; 
absurd (S (length l1') = 0); trivial; discriminate.
right; discriminate.
left; discriminate.
intros _; generalize (greater_case R R_dec (e2' :: l2') (e1' :: l1'));
destruct (list_forall
        (fun t1 : DS.A =>
         list_exists
           (fun t2 : DS.A =>
            match R_or_eq_dec R R_dec t2 t1 with
            | Equivalent => false
            | Less_than => false
            | Greater_than => true
            | Uncomparable => false
            end) (e2' :: l2')) (e1' :: l1')).
intro H'; generalize (H' (refl_equal _)); clear H'; intros [le [ll [P2' [P1' app_lt_cns]]]].
apply (multiset_closure_aux2 R (p := l1) (q := l2) (le := le)  (l := ll) l); simpl; auto.
setoid_rewrite P1; setoid_rewrite P1'; apply list_permut_app_app.
setoid_rewrite P2; setoid_rewrite P2';
apply list_permut_trans with (l ++ le ++ consn ll).
setoid_rewrite <- permut_app1; apply list_permut_app_app.
rewrite (ass_app le); apply list_permut_app_app.
destruct ll as [ | [x lx] ll].
destruct le as [ | e le].
generalize (list_permut_length P2'); simpl; intro; 
absurd (S (length l2') = 0); trivial; discriminate.
right; discriminate.
left; discriminate.
trivial.
Qed.


Lemma mult_is_complete_equiv :
 forall R R_dec l1 l2, list_permut l1 l2 -> mult R R_dec l1 l2 = Equivalent.
Proof.
intros R R_dec l1 l2 P; unfold mult;
generalize (remove_equiv_is_sound l1 l2);
destruct (remove_equiv l1 l2) as [l1' l2'];
intros [l [P1 [P2 H]]].
destruct l1' as [ | e1' l1'].
destruct l2' as [ | e2' l2'].
trivial.
setoid_rewrite P in P1; rewrite <- app_nil_end in P1; setoid_rewrite P1 in P2;
generalize (list_permut_length P2); rewrite length_app;
rewrite <- (plus_0_r (length l)); rewrite <- plus_assoc;
intro H'; generalize (plus_reg_l _ _ _ H'); clear H'; 
intro H'; simpl in H'; absurd (0 = S (length l2')); trivial; discriminate.
setoid_rewrite P in P1; setoid_rewrite P1 in P2;
setoid_rewrite <- permut_app1 in P2;
assert False; [apply (H e1') | contradiction].
left; trivial.
setoid_rewrite <- P2; left; trivial.
Qed.

Lemma mult_is_complete_greater :
 forall (R : relation elt) R_dec, transitive _ R -> (forall a, ~ R a a) ->
 forall l l1 l2, (forall a, In a l1 -> In a l2 -> False) ->
   trans_clos (multiset_extension_step R) (l ++ l2) (l ++ l1) -> 
	(list_forall 
	      (fun t2 => 
		 list_exists 
		   (fun t1 => 
		      match R_or_eq_dec R R_dec t1 t2 with
			| Greater_than => true
			| _ => false
                      end) l1) l2) = true.
Proof.
intros R R_dec trans_R irrefl_R l l1 l2 l2_disj_l1 ll2_lt_ll1.
assert (l2_lt_l1 := remove_context_trans_clos_multiset_extension_step_app1 
                               R_dec trans_R irrefl_R l2 l1 l ll2_lt_ll1).
clear ll2_lt_ll1.
destruct (multiset_closure R_dec trans_R l2_lt_l1) 
   as [ll [lc [P2 [P1 [ll_diff_nil [app_lt_cns cns_disj_app]]]]]].
assert (lc_eq_nil : lc = nil).
destruct lc as [ | c lc].
trivial.
assert False; [idtac | contradiction].
apply (l2_disj_l1 c);
[ setoid_rewrite P1 | setoid_rewrite P2]; apply in_or_app; right; left; trivial.
subst lc; rewrite <- app_nil_end in P1; rewrite <- app_nil_end in P2.
assert (Dummy := permut_list_forall_exists
 (fun t2 t1 : DS.A =>  
      match R_or_eq_dec R R_dec t1 t2 with
      | Equivalent => false
      | Less_than => false
      | Greater_than => true
      | Uncomparable => false
      end) P2 P1); simpl in Dummy; unfold elt in Dummy; rewrite Dummy; clear Dummy.
clear l1 P1 l2 P2 l2_disj_l1 l2_lt_l1;
induction ll as [ | [a la] ll].
absurd (@nil (elt * list elt) = nil); trivial.
simpl; rewrite list_forall_app.
assert (H : list_forall
         (fun t1 : DS.A =>
          list_exists
            (fun t2 : DS.A =>
             match R_or_eq_dec R R_dec t2 t1 with
             | Equivalent => false
             | Less_than => false
             | Greater_than => true
             | Uncomparable => false
             end) (consn ll)) (appendn ll) = true).
destruct ll as [ | [b lb] ll]; trivial; rewrite IHll; trivial.
discriminate.
intros c lc clc_in_ll x x_in_lc ; apply (app_lt_cns c lc); trivial; right; trivial.
intros _ c c_in_cns c_in_app; apply (cns_disj_app irrefl_R c); trivial.
right; trivial.
simpl; apply in_or_app; right; trivial.
assert (Dummy := list_forall_impl
(fun t1 : DS.A =>
       list_exists
         (fun t2 : DS.A =>
          match R_or_eq_dec R R_dec t2 t1 with
          | Equivalent => false
          | Less_than => false
          | Greater_than => true
          | Uncomparable => false
          end) (consn ll))
(fun t1 : DS.A =>
      Bool.ifb
        match R_or_eq_dec R R_dec a t1 with
        | Equivalent => false
        | Less_than => false
        | Greater_than => true
        | Uncomparable => false
        end true
        (list_exists
           (fun t2 : DS.A =>
            match R_or_eq_dec R R_dec t2 t1 with
            | Equivalent => false
            | Less_than => false
            | Greater_than => true
            | Uncomparable => false
            end) (consn ll))) (appendn ll)); unfold elt in *;
rewrite Dummy; clear Dummy; trivial.
generalize la (app_lt_cns a la (or_introl _ (refl_equal _)));
clear la ll_diff_nil app_lt_cns cns_disj_app H;
intros la la_lt_a; induction la as [ | b la]; trivial.
simpl; destruct (list_forall
            (fun t1 : DS.A =>
             Bool.ifb
               match R_or_eq_dec R R_dec a t1 with
               | Equivalent => false
               | Less_than => false
               | Greater_than => true
               | Uncomparable => false
               end true
               (list_exists
                  (fun t2 : DS.A =>
                   match R_or_eq_dec R R_dec t2 t1 with
                   | Equivalent => false
                   | Less_than => false
                   | Greater_than => true
                   | Uncomparable => false
                   end) (consn ll))) la); simpl.
generalize (R_or_eq_dec_is_sound _ R_dec a b);
destruct (R_or_eq_dec R R_dec a b).
intro a_eq_b; subst b; absurd (R a a); [apply (irrefl_R a) | apply la_lt_a; left; trivial].
intro; absurd (R a a); 
[apply (irrefl_R a) | apply trans_R with b; trivial; apply la_lt_a; left; trivial].
trivial.
intros [_ [_ not_b_R_a]]; absurd (R b a); trivial; apply la_lt_a; left; trivial.
absurd (false = true); [discriminate | apply IHla; intros; apply la_lt_a; right; trivial].
intros b b_in_app; destruct (R_or_eq_dec R R_dec a b); trivial.
Qed.


Lemma mult_is_complete :
 forall (R : relation elt) R_dec, transitive _ R -> (forall a, ~ R a a) ->
 forall l1 l2,
  match mult R R_dec l1 l2 with
  | Uncomparable => ~ list_permut l1 l2 /\
                                      ~trans_clos (multiset_extension_step R) l1 l2 /\ 
                                      ~trans_clos (multiset_extension_step R) l2 l1
  | _ => True
  end.
Proof.
intros R R_dec trans_R irrefl_R l1 l2.
unfold mult.
generalize (remove_equiv_is_sound l1 l2); destruct (remove_equiv l1 l2) as [l1' l2'].
intros [l [P1 [P2 l1'_disj_l2']]].
assert (l2'_disj_l1' : forall x : elt, In x l2' -> In x l1' -> False).
intros x H1 H2; apply (l1'_disj_l2' x); trivial.
generalize (@mult_is_complete_equiv R R_dec l1 l2)
(@mult_is_complete_greater R R_dec trans_R irrefl_R l l1' l2' l1'_disj_l2')
(@mult_is_complete_greater R R_dec trans_R irrefl_R l l2' l1' l2'_disj_l1').
destruct l1' as [ | e1' l1'].
destruct l2' as [ | e2' l2']; simpl; trivial.
destruct (list_forall
        (fun t2 : DS.A =>
         list_exists
           (fun t1 : DS.A =>
            match R_or_eq_dec R R_dec t1 t2 with
            | Equivalent => false
            | Less_than => false
            | Greater_than => true
            | Uncomparable => false
            end) (e1' :: l1')) l2'); trivial.
destruct (list_forall
        (fun t1 : DS.A =>
         list_exists
           (fun t2 : DS.A =>
            match R_or_eq_dec R R_dec t2 t1 with
            | Equivalent => false
            | Less_than => false
            | Greater_than => true
            | Uncomparable => false
            end) l2') (e1' :: l1')); trivial.
intros H1 H2 H3; repeat split; intro H.
setoid_rewrite H in P1; setoid_rewrite P1 in P2; 
setoid_rewrite <- permut_app1 in P2;
apply (l1'_disj_l2' e1'); [idtac | setoid_rewrite <- P2]; left; trivial.
absurd (false = true); [discriminate | apply H3];
apply list_permut_trans_clos_multiset_extension_step_2 with l2; trivial;
apply list_permut_trans_clos_multiset_extension_step_1 with l1; trivial.
absurd (false = true); [discriminate | apply H2];
apply list_permut_trans_clos_multiset_extension_step_2 with l1; trivial;
apply list_permut_trans_clos_multiset_extension_step_1 with l2; trivial.
Qed.

End Make.

