(******************************************************************************)
(*       _     *                                                              *)
(*      <o>    *   The Coccinelle Library / Evelyne Contejean                 *)
(*    -/@|@\-  *          CNRS-LRI-Universite Paris Sud                       *)
(*    -@ | @-  *                                                              *)
(*    -\@|@/-  *   This file is distributed under the terms of the            *)
(*      -v-    *   CeCILL-C licence                                           *)
(*             *                                                              *)
(******************************************************************************)

Add LoadPath "list_extensions". 

(** * Equational theory on a term algebra *)

Require Import List.
Require Import Relations.
Require Import Arith.
Require Import more_list.
Require Import term.

(** ** Module Type EqTh, an equational theory over a term algebra. *)


(** ** Module Type EqTh, an equational theory over a term algebra. *)

Module Type EqTh.

  Declare Module T : term.Term.

  Import T.

  (** *** One step at top. *)
  Inductive axiom (R : relation term) : term -> term -> Prop :=
    | instance : forall t1 t2 sigma,
      R t1 t2 -> axiom R (apply_subst sigma t1) (apply_subst sigma t2).

  (** *** One step at any position. *)
  Inductive one_step (R : relation term) : term -> term -> Prop :=
  | at_top : forall t1 t2, axiom R t1 t2 -> one_step R t1 t2
  | in_context : forall f l1 l2, 
      one_step_list R l1 l2 -> one_step R (Term f l1) (Term f l2)

      with one_step_list (R : relation term) : list term -> list term -> Prop :=
    | head_step : forall t1 t2 l,
      one_step R t1 t2 -> one_step_list R (t1 :: l) (t2 :: l)
    | tail_step : forall t l1 l2,
      one_step_list R l1 l2 -> one_step_list R (t :: l1) (t :: l2).

(*
  Parameter axiom:  forall (R : relation term), term -> term -> Prop.
  Parameter one_step: forall (R : relation term), term -> term -> Prop.
  Parameter one_step_list: forall (R : relation term), list term -> list term -> Prop.

  Axiom one_step_ind
     : forall (R : relation term)
         (P : forall t t0 : term, one_step R t t0 -> Prop)
         (P0 : forall l l0 : list term, one_step_list R l l0 -> Prop),
       (forall (t1 t2 : term) (a : axiom R t1 t2), P t1 t2 (at_top R t1 t2 a)) ->
       (forall (f0 : symbol) (l1 l2 : list term) (o : one_step_list R l1 l2),
        P0 l1 l2 o -> P (Term f0 l1) (Term f0 l2) (in_context R f0 l1 l2 o)) ->
       (forall (t1 t2 : term) (l : list term) (o : one_step R t1 t2),
        P t1 t2 o -> P0 (t1 :: l) (t2 :: l) (head_step R t1 t2 l o)) ->
       (forall (t : term) (l1 l2 : list term) (o : one_step_list R l1 l2),
        P0 l1 l2 o -> P0 (t :: l1) (t :: l2) (tail_step R t l1 l2 o)) ->
       forall (t t0 : term) (o : one_step R t t0), P t t0 o.

  Axiom one_step_list_ind
     : forall (R : relation term)
         (P : forall t t0 : term, one_step' R t t0 -> Prop)
         (P0 : forall l l0 : list term, one_step_list' R l l0 -> Prop),
       (forall (t1 t2 : term) (a : axiom R t1 t2), P t1 t2 (at_top R t1 t2 a)) ->
       (forall (f0 : symbol) (l1 l2 : list term) (o : one_step_list' R l1 l2),
        P0 l1 l2 o -> P (Term f0 l1) (Term f0 l2) (in_context R f0 l1 l2 o)) ->
       (forall (t1 t2 : term) (l : list term) (o : one_step' R t1 t2),
        P t1 t2 o -> P0 (t1 :: l) (t2 :: l) (head_step R t1 t2 l o)) ->
       (forall (t : term) (l1 l2 : list term) (o : one_step_list' R l1 l2),
        P0 l1 l2 o -> P0 (t :: l1) (t :: l2) (tail_step R t l1 l2 o)) ->
       forall (l l0 : list term) (o : one_step_list' R l l0), P0 l l0 o.
*)


  (** *** Equational Theory as reflexive, symetric 
      and transitive closure of an equational step. *)
  
(*   Parameter rwr: forall (R : relation term), term -> term -> Prop. *)

  (** *** Rewriting Theory as transitive closure of an equational step. *)
  Inductive rwr (R : relation term) : term -> term -> Prop :=
    | th_step : forall t1 t2, one_step R t1 t2 -> rwr R t1 t2
    | th_trans : forall t1 t2 t3, rwr R t1 t2 -> rwr R t2 t3 -> rwr R t1 t3.

  Fixpoint rwr_list (R : relation term) (l1 l2 : list term) {struct l1} : Prop :=
    match l1, l2 with
      | nil, nil => False
      | nil, _ :: _ => False
      | _ :: _, nil => False
      | h1::tl1, h2::tl2 => (rwr R h1 h2 /\ tl1 = tl2) \/
        (h1 = h2 /\ rwr_list R tl1 tl2) \/
        (rwr R h1 h2 /\ rwr_list R tl1 tl2) 
    end.

  Axiom context_cons :
    forall R f t1 t2 l, rwr R t1 t2 -> rwr R (Term f (t1 :: l)) (Term f (t2 :: l)).

  Axiom one_step_context_in :
    forall R t1 t2 f l l', one_step R t1 t2 -> 
      one_step R (Term f (l ++ t1 :: l')) (Term f (l ++ t2 :: l')).

  Axiom context_in :
    forall R t1 t2, rwr R t1 t2 -> 
      forall f l l', rwr R (Term f (l ++ t1 :: l')) (Term f (l ++ t2 :: l')).

  Axiom rwr_list_expand : 
    forall R l1 l2, rwr_list R l1 l2 -> 
      exists t1, exists l1', exists l1'', exists t2, exists l2', exists l2'',
	l1 = l1' ++ t1 :: l1'' /\ l2 = l2' ++ t2 :: l2'' /\
	(l1' = l2' \/ rwr_list R l1' l2') /\
	rwr R t1 t2 /\
	(l1'' = l2'' \/ rwr_list R l1'' l2'').

  Axiom general_context :
    forall R f l1 l2, rwr_list R l1 l2 -> rwr R (Term f l1) (Term f l2).

  Axiom rwr_list_length_eq :
    forall R l1 l2, rwr_list R l1 l2 -> length l1 = length l2.

  Axiom  rwr_apply_subst :
    forall R t1 t2 sigma, rwr R t1 t2 -> 
      rwr R (apply_subst sigma t1) (apply_subst sigma t2).

  Axiom rwr_inv : 
    forall A R (eq : A -> A -> Prop) (f : term -> A), 
      (forall a1 a2 a3, eq a1 a2 -> eq a2 a3 -> eq a1 a3) ->
      (forall t1 t2, one_step R t1 t2 -> eq (f t1) (f t2)) ->
      forall t1 t2, rwr R t1 t2 -> eq (f t1) (f t2).

  Axiom R1_included_in_R2_one_step :
    forall (R1 R2 : term -> term -> Prop) , 
      (forall t1 t2, R1 t1 t2 -> R2 t1 t2) ->
      forall t1 t2, one_step R1 t1 t2 -> one_step R2 t1 t2.

  Axiom R1_included_in_R2 :
    forall (R1 R2 : term -> term -> Prop), 
      (forall t1 t2, R1 t1 t2 -> R2 t1 t2) ->
      forall t1 t2, rwr R1 t1 t2 -> rwr R2 t1 t2.

  Axiom rwr_R1_included_in_R2 :
    forall (R1 R2 : term -> term -> Prop), 
      (forall t1 t2, rwr R1 t1 t2 -> R2 t1 t2) ->
      forall t1 t2, rwr R1 t1 t2 -> rwr R2 t1 t2.


  Axiom th_eq_closure_one_step :
    forall (R : term -> term -> Prop), 
      forall t1 t2, one_step (rwr R) t1 t2 -> rwr R t1 t2.

  Axiom th_eq_closure :
    forall t1 t2 (R : term -> term -> Prop),
      rwr (rwr R) t1 t2 -> rwr R t1 t2.

  Axiom  no_need_of_instance_is_modular :
    forall (R1 R2 : term -> term -> Prop), 
      (forall t1 t2, axiom R1 t1 t2 -> R1 t1 t2) ->
      (forall t1 t2, axiom R2 t1 t2 -> R2 t1 t2) ->
      (forall t1 t2, axiom (union term R1 R2) t1 t2 -> (union term R1 R2) t1 t2).

  Definition sym_refl R (t1 t2 : term) := R t1 t2 \/ R t2 t1 \/ t1 = t2.
  Definition th_eq R := rwr (sym_refl R).

  Axiom th_refl : forall R t, th_eq R t t.

  Axiom th_sym : forall R t1 t2, th_eq R t1 t2 -> th_eq R t2 t1.

End EqTh.


(** ** A functor building a Module EqTh from a term algebra. *)
Module Make (T1 : term.Term)  <: EqTh with Module T:=T1.

Module T := T1.
Import T.

(** *** One step at top. *)
Inductive axiom (R : relation term) : term -> term -> Prop :=
  | instance : forall t1 t2 sigma,
       R t1 t2 -> axiom R (apply_subst sigma t1) (apply_subst sigma t2).

(** *** One step at any position. *)
Inductive one_step (R : relation term) : term -> term -> Prop :=
  | at_top : forall t1 t2, axiom R t1 t2 -> one_step R t1 t2
  | in_context : forall f l1 l2, 
      one_step_list R l1 l2 -> one_step R (Term f l1) (Term f l2)

with one_step_list (R : relation term) : list term -> list term -> Prop :=
  | head_step : forall t1 t2 l,
       one_step R t1 t2 -> one_step_list R (t1 :: l) (t2 :: l)
 | tail_step : forall t l1 l2,
     one_step_list R l1 l2 -> one_step_list R (t :: l1) (t :: l2).

(** *** Rewriting Theory as transitive closure of an equational step. *)
Inductive rwr (R : relation term) : term -> term -> Prop :=
  | th_step : forall t1 t2, one_step R t1 t2 -> rwr R t1 t2
  | th_trans : forall t1 t2 t3, rwr R t1 t2 -> rwr R t2 t3 -> rwr R t1 t3.

Hint Constructors axiom one_step one_step_list rwr.

(** *** Generalization over lists of terms. *) 
Fixpoint rwr_list (R : relation term) (l1 l2 : list term) {struct l1} : Prop :=
   match l1, l2 with
   | nil, nil => False
   | nil, _ :: _ => False
   | _ :: _, nil => False
   | h1::tl1, h2::tl2 => (rwr R h1 h2 /\ tl1 = tl2) \/
                                      (h1 = h2 /\ rwr_list R tl1 tl2) \/
                                      (rwr R h1 h2 /\ rwr_list R tl1 tl2) 
   end.

(** *** Rewriting is a congruence. *)
Lemma context_cons :
  forall R f t1 t2 l, rwr R t1 t2 -> rwr R (Term f (t1 :: l)) (Term f (t2 :: l)).
Proof.
intros R f t1 t2 l H;
induction H as [ | t1 t2 t3 _ Trans1 _ Trans2]; auto;
apply th_trans with (Term f (t2 :: l)); trivial.
Qed.

Lemma one_step_context_in :
  forall R t1 t2 f l l', one_step R t1 t2 -> 
   one_step R (Term f (l ++ t1 :: l')) (Term f (l ++ t2 :: l')).
Proof.
intros R t1 t2 f l l' H; apply in_context; 
induction l as [ | t l ]; simpl; [ apply head_step | apply tail_step ]; 
trivial.
Qed.

Lemma context_in :
  forall R t1 t2, rwr R t1 t2 -> 
  forall f l l', rwr R (Term f (l ++ t1 :: l')) (Term f (l ++ t2 :: l')).
Proof.
intros R t1 t2 H; 
induction H as [ t1 t2 Step 
                          | t1 t2 t3 _ Trans1 _ Trans2].
intros f l l'; apply th_step; apply one_step_context_in; trivial.
intros f l l'; apply th_trans with (Term f (l ++ t2 :: l')); trivial.
Qed.

Lemma rwr_list_expand : 
  forall R l1 l2, rwr_list R l1 l2 -> 
  exists t1, exists l1', exists l1'', exists t2, exists l2', exists l2'',
  l1 = l1' ++ t1 :: l1'' /\ l2 = l2' ++ t2 :: l2'' /\
  (l1' = l2' \/ rwr_list R l1' l2') /\
  rwr R t1 t2 /\
  (l1'' = l2'' \/ rwr_list R l1'' l2'').
Proof.
intros R l1; induction l1 as [ | a1 l1]; 
intros [ | a2 l2] l1_R_l2; simpl in l1_R_l2; try contradiction.
destruct l1_R_l2 as [[t1_R_t2 l1_eq_l2] | [[t1_eq_t2 l1_R_l2] | [t1_R_t2 l1_R_l2]]].
exists a1; exists (nil (A:= term)); exists l1; exists a2; exists (nil (A:= term)); exists l2;
intuition.
subst a2; destruct (IHl1 _ l1_R_l2) as [t1 [l1' [l1'' [t2 [l2' [l2'' [H1 [H2 [H3 H4]]]]]]]]].
exists t1; exists (a1 :: l1'); exists l1''; exists t2; exists (a1 :: l2'); exists l2'';
intuition.
rewrite <- app_comm_cons; subst; trivial.
rewrite <- app_comm_cons; subst; trivial.
left; subst; trivial.
rewrite <- app_comm_cons; subst; trivial.
rewrite <- app_comm_cons; subst; trivial.
left; subst; trivial.
rewrite <- app_comm_cons; subst; trivial.
rewrite <- app_comm_cons; subst; trivial.
right; simpl; right; left; split; trivial.
rewrite <- app_comm_cons; subst; trivial.
rewrite <- app_comm_cons; subst; trivial.
right; simpl; right; left; split; trivial.
exists a1; exists (nil (A:= term)); exists l1; exists a2; exists (nil (A:= term)); exists l2;
intuition.
Qed.

Lemma general_context_aux :
  forall R f l1 l2 l' l'', rwr_list R l1 l2 -> 
  rwr R (Term f (l' ++ l1 ++ l'')) (Term f (l' ++ l2 ++ l'')).
Proof.
intros R f l1; induction l1 as [ | t1 l1]; intros [ | t2 l2] l' l'' l1_R_l2;
simpl in l1_R_l2; try contradiction.
destruct l1_R_l2 as [[t1_R_t2 l1_eq_l2] | [[t1_eq_t2 l1_R_l2] | [t1_R_t2 l1_R_l2]]].
subst; do 2 rewrite <- app_comm_cons; apply context_in; trivial.
assert (H:= IHl1 l2 (l' ++ t2 :: nil) l'' l1_R_l2).
do 2 rewrite app_ass in H; simpl in H; subst; do 2 rewrite <- app_comm_cons; trivial.
apply th_trans with (Term f (l' ++ (t2 :: l1) ++ l'')).
do 2 rewrite <- app_comm_cons; apply context_in; trivial.
assert (H:= IHl1 l2 (l' ++ t2 :: nil) l'' l1_R_l2).
do 2 rewrite app_ass in H; simpl in H; subst; do 2 rewrite <- app_comm_cons; trivial.
Qed.

Lemma general_context :
  forall R f l1 l2, rwr_list R l1 l2 -> rwr R (Term f l1) (Term f l2).
Proof.
intros R f l1 l2 l1_R_l2; 
assert (H:= general_context_aux R f l1 l2 nil nil l1_R_l2); 
simpl in H; do 2 rewrite <- app_nil_end in H; exact H.
Qed.

Lemma rwr_list_length_eq :
  forall R l1 l2, rwr_list R l1 l2 -> length l1 = length l2.
Proof.
intros R l1; induction l1 as [ | t1 l1]; intros [ | t2 l2] H; 
simpl in H; try contradiction; intuition.
subst; simpl; trivial.
simpl; rewrite (IHl1 l2); trivial.
simpl; rewrite (IHl1 l2); trivial.
Qed.

(** *** Compatibility with substitutions. *)
Lemma rwr_apply_subst :
  forall R t1 t2 sigma, rwr R t1 t2 -> 
   rwr R (apply_subst sigma t1) (apply_subst sigma t2).
Proof.
intros R; 
assert (one_step_apply_subst :
forall t1 t2 sigma , one_step R t1 t2 ->
one_step R (apply_subst sigma t1) (apply_subst sigma t2)).
intros t; pattern t; apply term_rec3; clear t.
intros x1 t2 sigma H; 
inversion_clear H as [ t1' t2' H'|];
inversion_clear H' as [ t1' t2' tau H''];
do 2 (rewrite <- subst_comp_is_subst_comp); auto.
intros f1 l1 Hrec t2 sigma H; 
inversion_clear H as [ t1' t2' H'| f l1' l2 H''].
inversion_clear H' as [ t1' t2' tau H'']; 
do 2 (rewrite <- subst_comp_is_subst_comp); auto.
simpl; apply in_context; generalize l2 H''; clear l2 H''; 
induction l1 as [ | a1 l1].
intros l2 H; inversion_clear H.
intros l2' H; inversion_clear H as [ t1' a2 l H' | t' l1' l2 H'].
simpl; apply head_step; apply Hrec; trivial; left; trivial.
simpl; apply tail_step;  apply IHl1; trivial; intros; 
apply Hrec; trivial; right; trivial.

intros t1 t2 sigma H; 
induction H as [ | t1 t2 t3 H1 Hrec1 H2 Hrec2]; auto;
apply th_trans with (apply_subst sigma t2); auto.
Qed.

(** *** Inclusion of relations and associated equational theories. *)
Lemma rwr_inv : 
  forall A R (eq : A -> A -> Prop) (f : term -> A), 
  (forall a1 a2 a3, eq a1 a2 -> eq a2 a3 -> eq a1 a3) ->
  (forall t1 t2, one_step R t1 t2 -> eq (f t1) (f t2)) ->
  forall t1 t2, rwr R t1 t2 -> eq (f t1) (f t2).
Proof.
intros A R eq f eq_trans Step_inv t1 t2 H; induction H; auto.
apply eq_trans with (f t2); auto.
Qed.

Lemma R1_included_in_R2_axiom :
  forall (R1 R2 : term -> term -> Prop) , 
  (forall t1 t2, R1 t1 t2 -> R2 t1 t2) ->
  forall t1 t2, axiom R1 t1 t2 -> axiom R2 t1 t2.
Proof.
intros R1 R2 Incl t1 t2 H; inversion H; auto.
Qed.

Lemma R1_included_in_R2_one_step :
  forall (R1 R2 : term -> term -> Prop) , 
  (forall t1 t2, R1 t1 t2 -> R2 t1 t2) ->
  forall t1 t2, one_step R1 t1 t2 -> one_step R2 t1 t2.
Proof.
intros R1 R2 Incl t1; pattern t1; apply term_rec3; clear t1.
intros x1 t2 H; inversion H; 
apply at_top; apply R1_included_in_R2_axiom with R1; trivial.
intros f1 l1 Hrec t2 H; inversion_clear H as [| f l1' l2 H'].
apply at_top; apply R1_included_in_R2_axiom with R1; trivial.
apply in_context; generalize l2 H'; clear l2 H'; 
induction l1 as [ | a1 l1 ]; intros l2 H; inversion_clear H.
apply head_step; apply Hrec; trivial; left; trivial.
apply tail_step; apply IHl1; trivial; intros; apply Hrec; trivial; right; trivial.
Qed.

Lemma R1_included_in_R2 :
  forall (R1 R2 : term -> term -> Prop), 
  (forall t1 t2, R1 t1 t2 -> R2 t1 t2) ->
  forall t1 t2, rwr R1 t1 t2 -> rwr R2 t1 t2.
Proof.
intros R1 R2 Incl t1 t2 H; induction H; auto.
apply th_step; apply R1_included_in_R2_one_step with R1; trivial.
apply th_trans with t2; auto.
Qed.

Lemma rwr_R1_included_in_R2 :
  forall (R1 R2 : term -> term -> Prop), 
  (forall t1 t2, rwr R1 t1 t2 -> R2 t1 t2) ->
  forall t1 t2, rwr R1 t1 t2 -> rwr R2 t1 t2.
Proof.
intros R1 R2 Incl t1 t2 H; 
apply R1_included_in_R2 with (rwr R1); auto.
apply th_step; apply at_top;
generalize (instance (rwr R1) t1 t2 nil H); 
do 2 rewrite empty_subst_is_id; trivial.
Qed.

Lemma th_eq_closure_one_step :
  forall (R : term -> term -> Prop), 
  forall t1 t2, one_step (rwr R) t1 t2 -> rwr R t1 t2.
Proof.
intros R t1; pattern t1; apply term_rec3; clear t1.
intros x1 t2 H; inversion_clear H as [ t1' t2' H' | ];
inversion_clear H'; apply rwr_apply_subst; trivial.
intros f1 l1 Hrec t2 H; inversion_clear H as [ t1' t2' H' | f' l1' l2 H'].
inversion_clear H'; apply rwr_apply_subst; trivial.
apply general_context; generalize l2 H'; clear l2 H'.
induction l1 as [ | a1 l1 ]; intros l2 H;
inversion H as [ H1 a2 H3 a1_R_a2 H5 H6 | H1 H2 l2' l1_R_l2' H5 H6]; subst.
left; split; [ apply Hrec; [left | idtac] | idtac]; trivial.
right; left; split; [idtac | apply IHl1; [ intros; apply Hrec | idtac]]; trivial.
right; trivial.
Qed.

Lemma th_eq_closure :
  forall t1 t2 (R : term -> term -> Prop),
  rwr (rwr R) t1 t2 -> rwr R t1 t2.
Proof.
intros t1 t2 R; generalize  (th_trans R) 
(th_eq_closure_one_step R); 
generalize (rwr R); clear R; 
intros rwr_R rwr_R_trans 
rwr_R_step H; induction H as [ | ]; auto.
apply rwr_R_trans with t2; auto.
Qed.

Lemma R_in_rwr_R : 
  forall t1 t2 (R : term -> term -> Prop),
  R t1 t2 -> rwr R t1 t2.
Proof.
intros t1 t2 R H; apply th_step; apply at_top; 
generalize (instance R t1 t2 nil H); do 2 rewrite empty_subst_is_id; trivial.
Qed.

Lemma no_need_of_instance_is_modular :
  forall (R1 R2 : term -> term -> Prop), 
  (forall t1 t2, axiom R1 t1 t2 -> R1 t1 t2) ->
  (forall t1 t2, axiom R2 t1 t2 -> R2 t1 t2) ->
  (forall t1 t2, axiom (union term R1 R2) t1 t2 -> (union term R1 R2) t1 t2).
Proof.
intros R1 R2 H1 H2 t1 t2 H;
inversion H as [t1' t2' sigma H']; subst;
inversion H'; [left; apply H1 | right; apply H2];
apply instance; trivial.
Qed.

Definition sym_refl R (t1 t2 : term) := R t1 t2 \/ R t2 t1 \/ t1 = t2.
Definition th_eq R := rwr (sym_refl R).

Lemma th_refl : forall R t, th_eq R t t.
Proof.
unfold th_eq, sym_refl.
intros R t; apply th_step; apply at_top; rewrite <- (empty_subst_is_id t);
apply instance; right; right; trivial.
Qed.

Lemma th_sym : forall R t1 t2, th_eq R t1 t2 -> th_eq R t2 t1.
Proof.
intro R.
intros t1 t2 H; refine (rwr_inv term (sym_refl R) 
(fun u1 u2 => th_eq R u2 u1) (fun t => t) _ _ _ _ _); trivial.
unfold th_eq; intros; apply th_trans with a2; trivial.
unfold th_eq; clear t1 t2 H; intros t1 t2 H; apply th_step.
assert (axiom_sym : 
forall t1 t2, axiom (sym_refl R) t1 t2 -> axiom (sym_refl R) t2 t1).
clear t1 t2 H; intros t1 t2 H; inversion_clear H as [u1 u2 sigma H'].
destruct H' as [H' | [H' | H']].
apply instance; right; left; trivial.
apply instance; left; trivial.
apply instance; right; right; subst; trivial.
generalize t2 H; clear t2 H; pattern t1; apply term_rec3; clear t1.
intros v1 t2 H; inversion_clear H as [ H1 H2 H' | f l1 l2 H' ].
apply at_top; apply axiom_sym; trivial.
intros f l1 IH t2 H; inversion_clear H as [H1 H2 H' | H1 H2 l2 H'].
apply at_top; apply axiom_sym; trivial.
apply in_context; generalize l2 H'; clear f t2 l2 H'; 
induction l1 as [ | t1 l1].
intros l2 H'; inversion H'.
intros [ | t2 l2] H'; 
inversion_clear H' as [ H1 H2 H'' | f l1' l2' H'' ].
apply head_step; apply IH; trivial; left; trivial.
apply tail_step; apply IHl1; trivial; intros; apply IH; trivial; right; trivial.
Qed.

End Make.


(* 
*** Local Variables: *** 
*** coq-prog-name: "coqtop -emacs -I . -I ../basis -I ../term_orderings -I ../list_extensions -I ../term_algebra" *** 
*** coq-prog-args: nil *** 
*** End: *** 
*) 
