(******************************************************************************)
(*       _     *                                                              *)
(*      <o>    *   The Coccinelle Library / Evelyne Contejean                 *)
(*    -/@|@\-  *          CNRS-LRI-Universite Paris Sud                       *)
(*    -@ | @-  *                                                              *)
(*    -\@|@/-  *   This file is distributed under the terms of the            *)
(*      -v-    *   CeCILL-C licence                                           *)
(*             *                                                              *)
(******************************************************************************)

Add LoadPath "basis". 
Add LoadPath "list_extensions". 
Add LoadPath "term_algebra". 

Require Import Bool.
Require Import List.
Require Import closure.
Require Import more_list.
Require Import equiv_list.
Require Import list_permut.
Require Import dickson.
Require Import Relations.
Require Import Wellfounded.
Require Import Arith.
Require Import Wf_nat.
Require Import term.
Require Import ordered_set.
Require Import Recdef.

(** A non-dependant version of lexicographic extension. *)
Definition lex (A B : Set) (eq_A_dec : forall a1 a2, {a1=a2}+{a1<>a2}) 
(o1 : relation A) (o2 : relation B) (s t : _ * _) :=
  match s with
  | (s1,s2) =>
    match t with
    | (t1,t2) => 
       if eq_A_dec s1 t1
       then o2 s2 t2
       else o1 s1 t1
     end
   end.

(** Transitivity of  lexicographic extension. *)
Lemma lex_trans :
 forall (A B : Set) eq_A_dec o1 o2, 
 antisymmetric A o1 -> transitive A o1 -> transitive B o2 ->
 transitive _ (lex _ _ eq_A_dec o1 o2).
Proof.
unfold transitive, lex; 
intros A B eq_A_dec o1 o2 A1 T1 T2 p1 p2 p3; 
destruct p1 as [a1 b1]; destruct p2 as [a2 b2]; destruct p3 as [a3 b3];
elim (eq_A_dec a1 a2); intro eq1.
subst a2; elim (eq_A_dec a1 a3); intro eq2; trivial; apply T2.
intro lt1; elim (eq_A_dec a1 a3); intro eq3.
subst a3; elim (eq_A_dec a2 a1); intro eq2.
subst a2; absurd (a1=a1); trivial.
intro lt2; generalize (A1 _ _ lt1 lt2); contradiction.
elim (eq_A_dec a2 a3); intro eq4.
subst a3; trivial.
intro lt2; apply T1 with a2; trivial.
Qed.

(** Well-foundedness of  lexicographic extension. *)
Lemma wf_lex :
  forall A B eq_A_dec o1 o2, well_founded o1 -> well_founded o2 ->
  well_founded (lex A B eq_A_dec o1 o2).
Proof.
intros A B eq_A_dec o1 o2 W1 W2; unfold well_founded in *; destruct a.
generalize b; clear b; pattern a; refine (well_founded_ind W1 _ _ a);
clear a; intros a IH1 b; pattern b; refine (well_founded_ind W2 _ _ b).
clear b; intros b IH2; apply Acc_intro.
destruct y; simpl; elim (eq_A_dec a0 a); intro a0_eq_a.
subst a0; apply IH2.
intro; apply IH1; trivial.
Qed.

(** ** Module Type Precedence, 
** Definition of a precedence. *)
Module Type Precedence.
Parameter A : Set.
Parameter prec : relation A.

Inductive status_type : Set :=
  | Lex : status_type
  | Mul : status_type.

Parameter status : A -> status_type.

Axiom prec_dec : forall a1 a2 : A, {prec a1 a2} + {~ prec a1 a2}.
Axiom prec_antisym : forall s, prec s s -> False.
Axiom prec_transitive : transitive A prec.

End Precedence.

(** ** Module Type RPO, 
** Definition of RPO from a precedence on symbols. *)

Module Type RPO.

Declare Module T : term.Term.
Declare Module P : Precedence with Definition A:= T.symbol.

Import T.
Import P.
Declare Module LP : list_permut.Permut with Definition DS.A:=term.
Import LP.

(** ** Definition of rpo.*)
Inductive rpo : term -> term -> Prop :=
  | Subterm : forall f l t s, In s l -> rpo_eq t s -> rpo t (Term f l)
  | Top_gt : 
       forall f g l l', prec g f -> 
       (forall s', In s' l' -> rpo s' (Term f l)) -> 
       rpo (Term g l') (Term f l)
  | Top_eq_lex : 
        forall f l l', status f = Lex -> rpo_lex l' l -> 
        (forall s', In s' l' -> rpo s' (Term f l)) ->
        rpo (Term f l') (Term f l)
  | Top_eq_mul : 
        forall f l l', status f = Mul -> rpo_mul l' l -> 
        rpo (Term f l') (Term f l)

with rpo_eq : term -> term -> Prop :=
  | Equiv : forall t t', equiv t t' -> rpo_eq t t'
  | Lt : forall s t, rpo s t -> rpo_eq s t

with equiv : term -> term -> Prop :=
  | Eq : forall t, equiv t t
  | Eq_lex : 
     forall f ll l1 l2, status f = Lex -> (forall t1 t2, In (t1,t2) ll -> equiv t1 t2) -> 
     l1 = map (fun st => fst st) ll ->
     l2 = map (fun st => snd st) ll ->
     equiv (Term f  l1) (Term f l2) 
  | Eq_mul :
     forall f ll l1 l2,  status f = Mul -> (forall t1 t2, In (t1,t2) ll -> equiv t1 t2) -> 
     list_permut l1 (map (fun st => fst st) ll) -> 
     list_permut l2 (map (fun st => snd st) ll) -> 
     equiv (Term f l1) (Term f l2)
  
with rpo_lex : list term -> list term -> Prop :=
  | List_gt : 
      forall s t l l', rpo s t -> length l = length l' -> 
      rpo_lex (s :: l) (t :: l')
  | List_eq : forall s s' l l', equiv s s' -> rpo_lex l l' -> rpo_lex (s :: l) (s' :: l')

with rpo_mul : list term -> list term -> Prop :=
  | List_mul : 
       forall a lg ls llc l l', 
       (forall t1 t2, In (t1,t2) llc -> equiv t1 t2) -> 
       list_permut l' (ls ++ (map (fun st => fst st) llc)) ->
       list_permut l (a :: lg ++ (map (fun st => snd st) llc)) ->
       (forall b, In b ls -> exists a', In a' (a :: lg) /\ rpo b a') ->
       rpo_mul l' l.

(* equiv is actually an equivalence *)
Axiom equiv_equiv  : equivalence term equiv.

(* Two equivalent terms have the same behaviour for rpo *)
Axiom equiv_rpo_equiv_1 :
  forall t t', equiv t t' -> (forall s, rpo s t <-> rpo s t').

Axiom equiv_rpo_equiv_2 :
  forall t t', equiv t t' -> (forall s, rpo t s <-> rpo t' s).

Axiom equiv_rpo_equiv_3 :
  forall t t', equiv t t' -> (forall s, rpo_eq s t <-> rpo_eq s t').

Axiom equiv_rpo_equiv_4 :
  forall t t', equiv t t' -> (forall s, rpo_eq t s <-> rpo_eq t' s).

(** ** rpo is a preorder, and its reflexive closure is an ordering. *)
Axiom rpo_closure :
  forall s t u, 
  (rpo t s -> rpo u t -> rpo u s) /\
  (rpo s t -> rpo t s -> False) /\
  (rpo s s -> False) /\
  (rpo_eq s t -> rpo_eq t s -> equiv s  t).

Axiom rpo_trans : forall s t u, rpo s t -> rpo t u -> rpo s u.

(** ** Main theorem: when the precedence is well-founded, so is the rpo. *)
Axiom wf_rpo : well_founded prec -> well_founded rpo.

(** ** RPO is compatible with the instanciation by a substitution. *)
Axiom rpo_subst :
  forall t s, rpo s t -> 
  forall sigma, rpo (apply_subst sigma s) (apply_subst sigma t).

(** ** RPO is compatible with adding context. *)
Axiom rpo_add_context :
 forall p ctx s t, rpo s t -> is_a_pos ctx p = true -> 
  rpo (replace_at_pos ctx s p) (replace_at_pos ctx t p).

End RPO.

Module Make (T1: term.Term) 
                    (P1 : Precedence with Definition A := T1.symbol)
<: RPO with Module T := T1 with Module P:=P1. 

Module T := T1.
Module P := P1.

Import T.
Import P.

Module LP := list_permut.Make (Term_eq_dec).
Import LP.

(*
Module LDS.
Definition A := (term * term)%type.
Lemma eq_A_dec : forall a1 a2 : A, {a1 = a2} + {a1 <> a2}.
Proof.
intros [e1 e2] [e1' e2']; destruct (eq_term_dec e1 e1') as [e1_eq_e1' | e1_diff_e1'].
destruct (eq_term_dec e2 e2') as [e2_eq_e2' | e2_diff_e2'].
left; subst; trivial.
right; intro H; apply e2_diff_e2'; injection H; trivial.
right; intro H; apply e1_diff_e1'; injection H; trivial.
Qed.
End LDS.

Module LLP := list_permut.Make (LDS).
*)

(** ** Definition of size-based well-founded orderings for induction.*)
Definition o_size s t := size s < size t.

Lemma wf_size :  well_founded o_size.
Proof.
generalize (well_founded_ltof _ size); unfold ltof; trivial.
Qed.

Definition size2 s := match s with (s1,s2) => (size s1, size s2) end.
Definition o_size2 s t := lex _ _ eq_nat_dec lt lt (size2 s) (size2 t).
Lemma wf_size2 : well_founded o_size2.
Proof.
refine (wf_inverse_image _ _ (lex _ _ eq_nat_dec lt lt) size2 _);
apply wf_lex; apply lt_wf.
Qed.

Definition size3 s := match s with (s1,s2) => (size s1, size2 s2) end.
Definition o_size3 s t := 
  lex _ _ eq_nat_dec lt (lex _ _ eq_nat_dec lt lt) (size3 s) (size3 t).
Lemma wf_size3 : well_founded o_size3.
Proof.
refine (wf_inverse_image _ _ 
  (lex _ _ eq_nat_dec lt (lex _ _ eq_nat_dec lt lt)) size3 _);
apply wf_lex; [ idtac | apply wf_lex ]; apply lt_wf.
Qed.

Lemma lex1 : 
 forall s f l t1 u1 t2 u2, In s l -> o_size3 (s,(t1,u1)) (Term f l,(t2,u2)).
Proof.
intros s f l t1 u1 t2 u2 In_s; unfold o_size3, size3, size2, lex;
elim (eq_nat_dec (size s) (size (Term f l))).
intro eq1; absurd (size (Term f l) < size (Term f l)); auto with arith;
generalize (size_direct_subterm s (Term f l) In_s); rewrite eq1; trivial.
intros; apply (size_direct_subterm s (Term f l) In_s).
Qed.

Lemma lex1_bis : 
 forall a f l t1 u1 t2 u2, o_size3 (Term f l,(t1,u1)) (Term f (a::l),(t2,u2)).
Proof.
intros a f l t1 u1 t2 u2; unfold o_size3, size3, size2, lex;
elim (eq_nat_dec (size (Term f l)) (size (Term f (a :: l)))); intro eq1.
do 2 rewrite size_unfold in eq1; injection eq1; clear eq1; intro eq1;
absurd (list_size size l < list_size size l); auto with arith;
generalize (plus_le_compat_r _ _ (list_size size l) (size_ge_one a));
rewrite <- eq1; trivial.
do 2 rewrite size_unfold;
simpl; apply lt_n_S; apply lt_le_trans with (1 + list_size size l);
auto with arith; 
apply plus_le_compat_r; apply size_ge_one.
Qed.

Lemma lex2 :
  forall t f l s u1 u2, In t l -> o_size3 (s,(t,u1)) (s,(Term f l, u2)).
Proof.
intros t f l s u1 u2 In_t;
unfold o_size3, size3, size2, lex;
elim (eq_nat_dec (size s) (size s)); intro eq1.
elim (eq_nat_dec (size t) (size (Term f l))); intro eq2.
absurd (size (Term f l) < size (Term f l)); auto with arith;
generalize (size_direct_subterm t (Term f l) In_t); rewrite eq2; trivial.
apply (size_direct_subterm t (Term f l) In_t).
absurd (size s = size s); trivial.
Qed.

Lemma lex3 :
  forall u f l s t, In u l -> o_size3 (s,(t,u)) (s,(t,Term f l)).
Proof.
intros u f l s t In_u;
unfold o_size3, size3, size2, lex;
elim (eq_nat_dec (size s) (size s)); intro eq1.
elim (eq_nat_dec (size t) (size t)); intro eq2.
apply (size_direct_subterm u (Term f l) In_u).
absurd (size t = size t); trivial.
absurd (size s = size s); trivial.
Qed.

Lemma o_size3_trans : transitive _ o_size3.
Proof.
generalize (lex_trans _ _ eq_nat_dec lt (lex _ _ eq_nat_dec lt lt));
unfold transitive, o_size3, size3, o_size2; intros H x y z lt1 lt2;
destruct x; destruct p; rename t into x1; rename t0 into x2; rename t1 into x3;
destruct y; destruct p; rename t into y1; rename t0 into y2; rename t1 into y3;
destruct z; destruct p; rename t into z1; rename t0 into z2; rename z1 into z3.
apply H with (size y1, size2 (y2,y3)); trivial.
unfold antisymmetric; 
intros n m lt' lt''; generalize (lt_asym n m lt' lt''); contradiction.
apply lt_trans.
fold transitive; apply lex_trans.
unfold antisymmetric; 
intros n m lt' lt''; generalize (lt_asym n m lt' lt''); contradiction.
unfold transitive; apply lt_trans.
unfold transitive; apply lt_trans.
Qed.

(** ** Definition of rpo.*)

Inductive rpo : term -> term -> Prop :=
  | Subterm : forall f l t s, In s l -> rpo_eq t s -> rpo t (Term f l)
  | Top_gt : 
       forall f g l l', prec g f -> 
       (forall s', In s' l' -> rpo s' (Term f l)) -> 
       rpo (Term g l') (Term f l)
  | Top_eq_lex : 
        forall f l l', status f = Lex -> rpo_lex l' l -> 
        (forall s', In s' l' -> rpo s' (Term f l)) ->
        rpo (Term f l') (Term f l)
  | Top_eq_mul : 
        forall f l l', status f = Mul -> rpo_mul l' l -> 
        rpo (Term f l') (Term f l)

with rpo_eq : term -> term -> Prop :=
  | Equiv : forall t t', equiv t t' -> rpo_eq t t'
  | Lt : forall s t, rpo s t -> rpo_eq s t

with equiv : term -> term -> Prop :=
  | Eq : forall t, equiv t t
  | Eq_lex : 
     forall f ll l1 l2, status f = Lex -> (forall t1 t2, In (t1,t2) ll -> equiv t1 t2) -> 
     l1 = map (fun st => fst st) ll ->
     l2 = map (fun st => snd st) ll ->
     equiv (Term f  l1) (Term f l2) 
  | Eq_mul :
     forall f ll l1 l2,  status f = Mul -> (forall t1 t2, In (t1,t2) ll -> equiv t1 t2) -> 
     list_permut l1 (map (fun st => fst st) ll) -> 
     list_permut l2 (map (fun st => snd st) ll) -> 
     equiv (Term f l1) (Term f l2)
  
with rpo_lex : list term -> list term -> Prop :=
  | List_gt : 
      forall s t l l', rpo s t -> length l = length l' -> 
      rpo_lex (s :: l) (t :: l')
  | List_eq : forall s s' l l', equiv s s' -> rpo_lex l l' -> rpo_lex (s :: l) (s' :: l')

with rpo_mul : list term -> list term -> Prop :=
  | List_mul : 
       forall a lg ls llc l l', 
       (forall t1 t2, In (t1,t2) llc -> equiv t1 t2) -> 
       list_permut l' (ls ++ (map (fun st => fst st) llc)) ->
       list_permut l (a :: lg ++ (map (fun st => snd st) llc)) ->
       (forall b, In b ls -> exists a', In a' (a :: lg) /\ rpo b a') ->
       rpo_mul l' l.

(* equiv is actually an equivalence *)
Lemma equiv_equiv  : equivalence term equiv.
Proof.
split.
(* Reflexivity *)
intro t; apply Eq.
(* Transitivity *)
intros t1; pattern t1; apply term_rec3.
intros v t2 t3 t1_eq_t2 t2_eq_t3; inversion t1_eq_t2; subst; trivial.
intros f l1 E_l1 t2 t3 t1_eq_t2 t2_eq_t3; 
inversion t1_eq_t2 as [ | f' ll l1' l2 Sf IH_ll H1 H2 | f' ll l1' l2 Sf H P1 P2]; subst; trivial.
inversion t2_eq_t3 as [ | f' ll' l2' l3 Sf' IH_ll' | f' ll' l2' l3 Sf' H P1 P2 ]; clear t2_eq_t3;
subst; trivial.
apply Eq_lex with (combine (A:= term) (B:= term) 
                                               (map (fun st : term * term => fst st) ll)
                                               (map (fun st : term * term => snd st) ll')); trivial.
generalize ll IH_ll E_l1 ll' IH_ll' H1; clear t1 ll IH_ll E_l1 ll' IH_ll' H1 t1_eq_t2 f Sf Sf';
induction ll as [ | [t1 t2] ll]; intros IH_ll E_l1 ll' IH_ll' H1.
intros; contradiction.
destruct ll' as [ | [t2' t3] ll']; simpl in H1.
discriminate.
injection H1; clear H1; intros H1 t2_eq_t2'; subst t2'.
simpl; intros u1 u2 [u1u2_eq_t1t2 | u1u2_in_ll''].
injection u1u2_eq_t1t2; intros; subst u1 u2; apply E_l1 with t2.
left; trivial.
apply IH_ll; left; trivial.
apply IH_ll'; left; trivial.
refine (IHll _ _ ll' _ H1 u1 u2 u1u2_in_ll'').
intros; apply IH_ll; right; trivial.
intros t t_in_ll1; apply E_l1; right; trivial.
intros; apply IH_ll'; right; trivial.
assert (L:= length_map (fun st => snd st) ll); rewrite H1 in L; rewrite length_map in L.
generalize ll ll' L; clear ll ll' L IH_ll E_l1 t1_eq_t2 IH_ll' H1 t1;
induction ll as [ | [t1 t2] ll]; intros [ | [t2' t3] ll']; simpl; intro L; trivial; try discriminate.
rewrite <- (IHll ll'); trivial; injection L; trivial.
assert (L:= length_map (fun st => snd st) ll); rewrite H1 in L; rewrite length_map in L.
generalize ll' ll L; clear ll' ll L IH_ll E_l1 t1_eq_t2 IH_ll' H1 t1;
induction ll' as [ | [t2 t3] ll']; intros [ | [t1 t2'] ll]; simpl; intro L; trivial; try discriminate.
rewrite <- (IHll' ll); trivial; injection L; trivial.
rewrite Sf in Sf'; discriminate.
subst; inversion t2_eq_t3 as [ | f' ll' l1' l2' Sf' IH_ll' | f' ll' l2' l3 Sf' H P2' P3]; subst; trivial.
rewrite Sf in Sf'; discriminate.
assert (IH_ll : forall t : term,  In t  (map (fun st : term * term => fst st) ll) -> 
                        forall y z : term, equiv t y -> equiv y z -> equiv t z).
intros t t_in_l1; apply E_l1; setoid_rewrite (in_permut_in t P1); trivial.
clear E_l1; 
destruct (trans_equiv_list equiv ll ll') as [ll'' [E_ll'' [P1'' P2'']]]; trivial.
setoid_rewrite <- P2; trivial.
apply Eq_mul with ll''; trivial.
setoid_rewrite <- P1 in P1''; trivial.
setoid_rewrite <- P3 in P2''; trivial.
(* Symmetry *)
intros t1; pattern t1; apply term_rec3.
intros v t2 t1_eq_t2; inversion t1_eq_t2; subst; trivial.
intros f l1 IHl t2 t1_eq_t2; 
inversion t1_eq_t2 as 
  [ 
  | f' ll l1' l2 Sf Hll H1 H2 
  | f' ll l1' l2 Sf H P1 P2 H1 H2].
subst; trivial.
subst; destruct (equiv_swap _ equiv ll Hll) as [ll' [E_ll' [H1 [H2 H3]]]].
rewrite H1; rewrite H2; apply Eq_lex with ll'; trivial;
intros t2 t3 t2t3_in_ll'; setoid_rewrite <- H3 in t2t3_in_ll'; apply IHl. 
setoid_rewrite in_map_iff; exists (t3,t2); simpl; split; trivial.
apply Hll; trivial.
subst; destruct (equiv_swap _ equiv ll H0) as [ll' [E_ll' [H1 [H2 H3]]]].
apply Eq_mul with ll'; trivial.
intros t2 t3 t2t3_in_ll'; setoid_rewrite <- H3 in t2t3_in_ll'; apply IHl. 
setoid_rewrite (in_permut_in t3 P1);
setoid_rewrite in_map_iff; exists (t3,t2); simpl; split; trivial.
apply H0; trivial.
rewrite <- H2; trivial.
rewrite <- H1; trivial.
Qed.

Lemma equiv_same_top :
  forall f g l l', equiv (Term f l) (Term g l') -> f = g.
Proof.
intros f g l l' H; inversion H; subst; trivial.
Qed.

Lemma equiv_same_length :
  forall f1 f2 l1 l2, equiv (Term f1 l1) (Term f2 l2) -> length l1 = length l2.
Proof.
intros f1 f2 l1 l2 t1_eq_t2; inversion t1_eq_t2; subst.
trivial.
do 2 rewrite length_map; trivial.
assert (Dummy := list_permut_length H5); 
unfold elt, DS.A, Term_eq_dec.A in *; rewrite Dummy; clear Dummy;
assert (Dummy := list_permut_length H6); 
unfold elt, DS.A, Term_eq_dec.A in *; rewrite Dummy; clear Dummy;
do 2 rewrite length_map; trivial.
Qed.

Lemma equiv_same_size :
  forall t t', equiv t t' -> size t = size t'.
Proof.
intros t; pattern t; apply term_rec3; clear t.
intros v t' t_eq_t'; inversion t_eq_t'; subst; trivial.
intros f l IHl t' t_eq_t'; inversion t_eq_t'; trivial.
(* Lex case *)
subst; do 2 rewrite size_unfold; apply (f_equal (fun n => 1 + n)).
clear t_eq_t'; induction ll as [ | [s t] ll]; trivial.
simpl; rewrite (IHl s (or_introl _ (refl_equal _)) t).
rewrite IHll; trivial.
intros; apply H2; right; trivial.
intros; apply IHl; trivial; right; trivial.
apply H2; left; trivial.
(* Mul case *)
subst; do 2 rewrite size_unfold; apply (f_equal (fun n => 1 + n)).
assert (H := list_permut_size size H3); unfold elt, DS.A, Term_eq_dec.A in *;
setoid_rewrite H; clear H.
assert (H := list_permut_size size H5); unfold elt, DS.A, Term_eq_dec.A in *;
setoid_rewrite H; clear H.
assert (H : forall s t, In (s, t) ll -> size s = size t).
intros s t st_in_ll; apply IHl.
setoid_rewrite (in_permut_in s H3).
setoid_rewrite in_map_iff; exists (s,t); split; trivial.
apply H2; trivial.
clear IHl H2 H3 H5; induction ll as [ | [s t] ll]; trivial.
simpl; rewrite (H s t (or_introl _ (refl_equal _))); rewrite IHll; trivial.
intros; apply H; right; trivial.
Qed.

Lemma rpo_subterm_equiv :
 forall s t, equiv t s -> forall tj, direct_subterm tj t -> rpo tj s.
Proof.
intro s; pattern s; apply term_rec2; clear s; 
induction n as [ | n]; intros t St.
absurd (1 <= 0); auto with arith; 
apply le_trans with (size t); trivial; apply size_ge_one.
intros s s_eq_t; inversion s_eq_t; subst.
(* Equality *)
intros tj H; destruct t as [ | f l]; simpl in H; [contradiction | idtac ].
apply (Subterm f l tj tj); trivial; apply Equiv; apply Eq.
(* Lex case *)
simpl; intros tj tj_in_ll1; 
destruct (equiv_in_list _ equiv tj ll H0 tj_in_ll1) as [ tj' [ll1 [ll2 [tj_eq_tj' H']]]].
subst ll; refine (Subterm f _ tj tj' _ _).
rewrite map_app; apply in_or_app; right; left; trivial.
apply Equiv; apply H0; apply in_or_app; right; left; trivial.
(* Mul case *)
simpl; intros tj tj_in_l1; setoid_rewrite (in_permut_in tj H1) in tj_in_l1.
destruct (equiv_in_list _ equiv tj ll H0 tj_in_l1) as [ tj' [ll1 [ll2 [tj_eq_tj' H']]]].
subst ll; refine (Subterm f _ tj tj' _ _).
setoid_rewrite (in_permut_in tj' H2);
rewrite map_app; apply in_or_app; right; left; trivial.
apply Equiv; apply H0; apply in_or_app; right; left; trivial.
Qed.

Lemma rpo_subterm :
 forall s t, rpo t s -> forall tj, direct_subterm tj t -> rpo tj s.
Proof.
intros s t;
assert (H : forall p : term * term,
             match p with
              | (s,t) => rpo t s -> forall tj, direct_subterm tj t -> rpo tj s
             end).
clear s t; intro p; pattern p; refine (well_founded_ind wf_size2 _ _ _); clear p.
intros [s [ v | f l]] IH t_lt_s tj tj_in_l; simpl in tj_in_l; [ contradiction | idtac].
inversion t_lt_s as [ f' l' t' s' s'_in_l' t'_le_s'
                               | f' g' k k' g'_prec_f' H
                               | f' k k' Sf' k'_lt_k H
                               | f' k k' Sf' k'_lt_k ].
(* Subterm *)
subst; inversion t'_le_s' as [ t'' s'' t'_eq_s' | t'' s'' t'_lt_s']; clear t'_le_s'; subst.
apply (Subterm f' l' tj s'); trivial;
apply Lt; apply rpo_subterm_equiv with (Term f l); trivial.
apply (Subterm f' l' tj s'); trivial.
apply Lt; apply (IH (s', Term f l)); trivial.
unfold o_size2, size2, lex.
destruct (eq_nat_dec (size s') (size (Term f' l'))) as [H | H].
absurd (size (Term f' l') < size (Term f' l')); auto with arith;
generalize (size_direct_subterm s' (Term f' l') s'_in_l'); rewrite H; trivial.
apply size_direct_subterm; trivial.
(* Top_gt *)
subst; apply H2; trivial.
(* Top_eq_lex *)
subst; apply H2; trivial.
(* Top_eq_mul *)
inversion k'_lt_k as [a lg ls llc l1 l2 H P1 P2 H']; subst.
setoid_rewrite (in_permut_in tj P1) in tj_in_l;
destruct (in_app_or _ _ _ tj_in_l) as [tj_in_ls | tj_in_lc]; clear tj_in_l.
destruct (H' _ tj_in_ls) as [a' [a'_in_a_lg tj_lt_a']];
apply (Subterm f k tj a').
setoid_rewrite (in_permut_in a' P2); rewrite app_comm_cons; 
apply in_or_app; left; trivial.
apply Lt; trivial.
destruct (equiv_in_list _ equiv tj llc) as [tj' [ll1 [ll2 [tj_eq_tj' H'']]]]; trivial; subst llc;
apply (Subterm f k tj tj').
setoid_rewrite (in_permut_in tj' P2).
right; apply in_or_app; right;
rewrite map_app; apply in_or_app; right; left; trivial.
apply Equiv; apply H; apply in_or_app; right; left; trivial.
apply (H (s,t)).
Qed.

Lemma trans_clos_subterm_rpo:
  forall s t,  trans_clos direct_subterm s t -> rpo s t.
Proof.
intros s t H; induction H as [ s [ v | f l ] H | s t u H1 H2].
inversion H.
apply (Subterm f l s s); trivial; apply Equiv; apply Eq.
apply (rpo_subterm u t); trivial.
Qed.

Lemma var_case :
  forall t v, rpo (Var v) t -> trans_clos direct_subterm (Var v) t.
Proof.
assert (No_subterm_for_var : 
               forall t t', trans_clos direct_subterm t' t -> (forall v, t = Var v -> False)).
intros t t' H; induction H as [ t' t H | t'' t' t H1 H2].
intros v t_eq_v; subst t; inversion H.
apply IHH2.
intros t; pattern t; apply term_rec3; clear t.
intros t v v_lt_t; inversion v_lt_t as [ f l s' u u_in_l v_le_u | | | ]; subst.
intros f l IHl v v_lt_t; inversion v_lt_t as [ f' l' s' u u_in_l v_le_u | | | ]; subst.
inversion v_le_u as [v' u' v_eq_u | v' u' v_lt_u]; subst.
inversion v_eq_u; subst.
apply closure.t_step; trivial.
apply trans_clos_is_trans with u.
apply IHl; trivial.
apply closure.t_step; trivial.
Qed.

Lemma subterm_equiv :
  forall s t t', trans_clos direct_subterm s t -> equiv t t' ->
                    exists s', trans_clos direct_subterm s' t' /\ equiv s s'.
Proof.
assert (No_subterm_for_var : 
               forall t t', trans_clos direct_subterm t' t -> (forall v, t = Var v -> False)).
intros t t' H; induction H as [ t' t H | t'' t' t H1 H2].
intros v t_eq_v; subst t; inversion H.
apply IHH2.
assert (Direct_case : forall s t t', direct_subterm s t -> equiv t t' ->
exists s' : term, trans_clos direct_subterm s' t' /\ equiv s s').
intros s t t' s_sub_t t_eq_t'; 
inversion t_eq_t' as [ | f' ll l1 l2 Sf E_ll | f' ll l1 l2 Sf E_ll P1 P2]; subst.
exists s; split; [apply closure.t_step | apply Eq]; trivial.
destruct (equiv_in_list _ equiv s ll E_ll s_sub_t) as [s' [ll1' [ll2' [s_eq_s' H]]]]; subst ll;
exists s'; split.
simpl; rewrite map_app; apply closure.t_step; simpl; apply in_or_app; right; left; trivial.
apply E_ll; apply in_or_app; right; left; trivial.
simpl in s_sub_t;
assert (H := in_permut_in s P1); unfold elt, DS.A, Term_eq_dec.A in *;
setoid_rewrite H in s_sub_t; clear H;
destruct (equiv_in_list _ equiv s ll E_ll s_sub_t) as [s' [ll1' [ll2' [s_eq_s' H]]]]; subst ll;
exists s'; split.
apply closure.t_step; simpl; setoid_rewrite (in_permut_in s' P2); 
rewrite map_app; apply in_or_app; right; left; trivial.
apply E_ll; apply in_or_app; right; left; trivial.
intros s t t' s_sub_t; generalize t'; clear t'; 
induction s_sub_t as [ s t s_sub_t | s s' t s_sub_s' s'_sub_t ].
intro t'; apply Direct_case; trivial.
intros t' t_eq_t'; destruct (IHs'_sub_t t' t_eq_t') as [s'' [s''_sub_t' s'_eq_s'']].
destruct (Direct_case s s' s'' s_sub_s' s'_eq_s'') as [u [u_sub_s'' s_equiv_u]].
exists u; split; trivial.
apply trans_clos_is_trans with s''; trivial.
Qed.

Lemma equiv_rpo_equiv_1_aux : 
forall l k k' ll,
   (forall t : term,
      In t l -> forall t0 t' : term, equiv t0 t' -> rpo t t0 -> rpo t t') ->
   (forall t1 t2 : term, In (t1, t2) ll -> equiv t1 t2) ->
   list_permut k (map (fun st : term * term => fst st) ll) ->
   list_permut k' (map (fun st : term * term => snd st) ll) ->
   rpo_mul l k -> rpo_mul l k'.
Proof.
intros l k k' ll IHl E_ll P1 P2 l_lt_k;
inversion l_lt_k as [a lg ls llc k'' k''' E_llc P1' P2' H]; subst.
setoid_rewrite P1 in P2'.
assert (a_in_ll1 : In a (map (fun st : term * term => fst st) ll)).
setoid_rewrite (in_permut_in a P2'); left; trivial.
destruct (equiv_in_list _ equiv a ll E_ll a_in_ll1) as [a' [ll1 [ll2 [a_eq_a' H']]]]; subst ll.
rewrite map_app in P2'; simpl in P2'; assert (P2'' := list_permut_sym P2');
clear P2'; setoid_rewrite <- permut_cons_inside in P2'';
rewrite <- map_app in P2''.
assert (E_ll' : forall t1 t2, In (t1,t2) (ll1 ++ ll2) -> equiv t1 t2).
intros; apply E_ll; apply in_insert; trivial.
destruct (split_equiv equiv_equiv (ll1 ++ ll2) _ _ E_ll' (list_permut_sym P2''))
as [llg [llc' [E_llg [E_llc' [Q1 [Q2 Q]]]]]].
destruct (trans_equiv_list equiv llc llc') as [llc'' [E_llc'' [H1 H2]]]; trivial.
intros e1 _; apply (equiv_trans _ _ equiv_equiv).
apply (List_mul a' (map (fun st => snd st) llg) ls llc''); trivial.
setoid_rewrite P1'; setoid_rewrite <- permut_app1; trivial.
setoid_rewrite P2; apply list_permut_sym; rewrite map_app; simpl;
setoid_rewrite <- permut_cons_inside; do 2 rewrite <- map_app;
setoid_rewrite Q; do 2 rewrite map_app;
setoid_rewrite <- permut_app1; apply list_permut_sym; trivial.
intros b b_in_ls; destruct (H b b_in_ls) as [a'' [[a''_eq_a | a''_in_lg] b_lt_a'']].
subst a''; exists a'; split; 
[left; trivial 
| apply IHl with a; trivial; 
  setoid_rewrite (in_permut_in b P1'); apply in_or_app; left; trivial].
setoid_rewrite (in_permut_in a'' Q1) in a''_in_lg;
destruct (equiv_in_list _ equiv a'' llg E_llg a''_in_lg) as [a3 [llg1 [llg2 [a''_eq_a3 H3]]]]; 
subst llg; exists a3; split.
right; rewrite map_app; apply in_or_app; right; left; trivial.
apply IHl with a''; trivial;
setoid_rewrite (in_permut_in b P1'); apply in_or_app; left; trivial.
Qed.

Lemma equiv_rpo_equiv_1 :
  forall t t', equiv t t' -> (forall s, rpo s t <-> rpo s t').
Proof.
assert (H : forall t t', equiv t t' -> (forall s, rpo s t -> rpo s t')).
intros t t' t_eq_t' s; generalize t t' t_eq_t'; clear t t' t_eq_t'; 
pattern s; apply term_rec3; clear s.
(* Variable case *)
intros v t t' t_eq_t' v_lt_t; assert (v_sub_t := var_case t v v_lt_t).
destruct (subterm_equiv (Var v) t t' v_sub_t t_eq_t') as [s' [s'_subt_t' v_eq_s']].
inversion v_eq_s' as [v' | | ]; subst;
apply trans_clos_subterm_rpo; trivial.
(* Composed case *)
intros f l IHl t t' t_eq_t' s_lt_t;
generalize t t' t_eq_t' l IHl s_lt_t; clear t t' t_eq_t' l IHl s_lt_t;
intro t; pattern t; apply term_rec3; clear t.
intros v t' v_eq_t' l IHl s_lt_v; inversion s_lt_v.
intros g k IHk t' gk_eq_t' l IHl fl_lt_gk;
inversion fl_lt_gk as [ f'' l' s' t'' t''_in_l' s_le_t''
                               | f'' g' k' k'' g'_prec_f'' H'
                               | f'' k' k'' Sf'' k''_lt_k' H H1 H2 
                               | f'' k' k'' Sf'' k''_lt_k' ]; clear fl_lt_gk; subst. 
(* Subterm *)
inversion gk_eq_t' as [ | f' ll l1 l2 Sf E_ll | f' ll l1 k' Sf E_ll P1 P2 ]; subst; trivial; clear gk_eq_t'.
(* Subterm, equivalence is syntactic identity *)
apply Subterm with t''; trivial.
(* Subterm, equivalence with Lex top symbol *)
destruct (equiv_in_list _ equiv t'' ll E_ll t''_in_l') as [t''' [ll1' [ll2' [t'_eq_t'' H]]]]; subst ll;
refine (Subterm g _ (Term f l) t''' _ _).
rewrite map_app; apply in_or_app; right; left; trivial.
inversion s_le_t'' as [s t4 s_eq_t'' | s t4 s_lt_t'']; clear s_le_t''; subst t4 s.
apply Equiv;  apply (equiv_trans _ _ equiv_equiv) with t''; trivial.
apply Lt; apply IHk with t''; trivial.
(* Subterm, equivalence with Mul top symbol *)
setoid_rewrite (in_permut_in t'' P1) in t''_in_l';
destruct (equiv_in_list _ equiv t'' ll E_ll t''_in_l') as [t''' [ll1' [ll2' [t'_eq_t'' H]]]]; subst ll;
refine (Subterm g _ (Term f l) t''' _ _); trivial.
setoid_rewrite (in_permut_in t''' P2);
rewrite map_app; apply in_or_app; right; left; trivial.
inversion s_le_t'' as [s t4 s_eq_t'' | s t4 s_lt_t'']; clear s_le_t''; subst t4 s.
apply Equiv;  apply (equiv_trans _ _ equiv_equiv) with t''; trivial.
apply Lt; apply IHk with t''; trivial; setoid_rewrite (in_permut_in t'' P1);
rewrite map_app; apply in_or_app; right; left; trivial.
(* Top_gt  *)
inversion gk_eq_t' as [ | f' ll l1 l2 Sf E_ll | f' ll l1 k' Sf E_ll P1 P2 ]; subst.
(* Top_gt, equivalence is syntactic identity *)
apply Top_gt; trivial.
(* Top_gt, equivalence with Lex top symbol *)
apply Top_gt; trivial.
intros s' s'_in_l; apply (IHl s' s'_in_l _ _ gk_eq_t'); apply H'; trivial.
(* Top_gt, equivalence with Mul top symbol *)
apply Top_gt; trivial.
intros s' s'_in_l; apply (IHl s' s'_in_l _ _ gk_eq_t'); apply H'; trivial.
(* Top_eq_lex *)
inversion gk_eq_t' as [ | f' ll l1 l2 Sf E_ll | f' ll l1 k' Sf E_ll P1 P2 ]; subst. 
(* Top_eq_lex, equivalence is syntactic identity *)
apply Top_eq_lex; trivial.
(* Top_eq_lex, equivalence with Lex top symbol *)
apply Top_eq_lex; trivial.
generalize l IHl ll E_ll k''_lt_k'; clear l IHl k''_lt_k' H3 ll E_ll IHk gk_eq_t'.
induction l as [ | s l]; intros IHl' ll E_ll k''_lt_k'; 
inversion k''_lt_k' as [ H1 t H2 l1 s_lt_t L H4 H' | H1 t H2 l1 s_eq_t l_lt_l1 H3 H']; subst; 
destruct ll as [ | [t' t''] ll]; try discriminate;
simpl in H'; injection H'; intros; subst; clear H'; simpl.
apply List_gt.
apply IHl' with t'; [left | apply E_ll; left | idtac ]; trivial.
rewrite L; do 2 rewrite length_map; trivial.
apply List_eq.
apply (equiv_trans _ _ equiv_equiv) with t'; trivial; apply E_ll; left; trivial.
apply IHl; trivial.
intros t t_in_l; apply IHl'; right; trivial.
intros; apply E_ll; right; trivial.
intros s' s'_in_l; apply (IHl s' s'_in_l _ _ gk_eq_t'); apply H3; trivial.
rewrite Sf in Sf''; discriminate.
(* Top_eq_mul *)
inversion gk_eq_t' as [ | f' ll l1 l2 Sf E_ll | f' ll l1 k' Sf E_ll P1 P2 ]; subst. 
(* Top_eq_mul, equivalence is syntactic identity *)
apply Top_eq_mul; trivial.
rewrite Sf in Sf''; discriminate.
(* Top_eq_mul, equivalence with Mul top symbol *)
apply Top_eq_mul; trivial.
apply (equiv_rpo_equiv_1_aux l k k') with ll; trivial.

intros t t' t_eq_t' s; split; apply H; trivial.
apply (equiv_sym _ _ equiv_equiv); trivial.
Qed.


Lemma equiv_rpo_equiv_2_aux : 
forall l k k' ll,
   (forall t : term,
      In t l -> forall t0 t' : term, equiv t0 t' -> rpo t0 t -> rpo t' t) ->
   (forall t1 t2 : term, In (t1, t2) ll -> equiv t1 t2) ->
   list_permut k (map (fun st : term * term => fst st) ll) ->
   list_permut k' (map (fun st : term * term => snd st) ll) ->
   rpo_mul k l -> rpo_mul k' l.
Proof.
intros l k k' ll IHl E_ll P1 P2 k_lt_l;
inversion k_lt_l as [a lg ls llc k'' k''' E_llc P1' Pl H]; subst.
setoid_rewrite P1 in P1'.
destruct (split_equiv equiv_equiv ll _ _ E_ll P1') 
  as [lls [llc' [E_lls [E_llc' [Q1 [Q2 Q]]]]]].
destruct (equiv_swap _ equiv llc' E_llc') as [llc'' [E_llc'' [H1 [H2 H3]]]].
destruct (trans_equiv_list equiv llc'' llc) as [llc3 [E_llc3 [R1 R2]]]; trivial.
intros e1 _; apply (equiv_trans _ _ equiv_equiv).
intros; apply (equiv_sym _ _ equiv_equiv); apply E_llc''; trivial.
unfold elt, DS.A, Term_eq_dec.A in *; rewrite <- H1; 
apply list_permut_sym; trivial.
apply (List_mul a lg (map (fun st : elt * elt => snd st) lls) llc3); trivial.
setoid_rewrite P2; setoid_rewrite Q; rewrite map_app;
setoid_rewrite <- permut_app1;
unfold elt, DS.A, Term_eq_dec.A in *; rewrite H2; auto.
setoid_rewrite Pl; setoid_rewrite <- permut_cons;
setoid_rewrite <- permut_app1;
unfold elt, DS.A, Term_eq_dec.A in *; trivial.
intros b b_in_lls2;
destruct (equiv_in_list_2 _ equiv _ _ E_lls b_in_lls2) 
  as [b' [lls1 [lls2 [b'_eq_b H']]]].
destruct (H b') as [a' [a'_in_alg b'_lt_a']].
setoid_rewrite (in_permut_in b' Q1);
subst; rewrite map_app; apply in_or_app; right; left; trivial.
exists a'; split; trivial; apply IHl with b'; trivial;
setoid_rewrite (in_permut_in a' Pl); rewrite app_comm_cons;
apply in_or_app; left; trivial.
Qed.

Lemma equiv_rpo_equiv_2 :
  forall t t', equiv t t' -> (forall s, rpo t s <-> rpo t' s).
Proof.
assert (H : forall t t', equiv t t' -> (forall s, rpo t s -> rpo t' s)).
intros t t' t_eq_t' s; generalize t t' t_eq_t'; clear t t' t_eq_t'; 
pattern s; apply term_rec3; clear s.
intros v t t' t_eq_t' t_lt_s; inversion t_lt_s.
intros f l IHl t t' t_eq_t' t_lt_s.
generalize l IHl t' t_eq_t' t_lt_s; clear l IHl t' t_eq_t' t_lt_s; 
pattern t; apply term_rec3; clear t.
intros v l IHl t' v_eq_t' v_lt_fl; inversion v_eq_t'; subst; trivial.
intros g k IHk l IHl t' gk_eq_t' gk_lt_fl;
inversion gk_lt_fl as [ g'' k' s' t'' t''_in_l t_le_t''
                               | g'' f' k' k'' g_prec_f H'
                               | g'' k' k'' Sf kk1_lt_l H H1 H2 
                               | g'' k' k'' Sf k''_lt_k' ]; clear gk_lt_fl; subst. 
(* Subterm *)
inversion t_le_t''; subst; apply Subterm with t''; trivial.
apply Equiv;  apply (equiv_trans _ _ equiv_equiv) with (Term g k); trivial;
apply (equiv_sym _ _ equiv_equiv); trivial.
apply Lt; apply IHl with (Term g k); trivial.
(* Top_gt *)
subst; inversion gk_eq_t' as [ | g' kk k1 k2 Sg E_kk | g' kk k1 k2 Sg E_kk P1 P2 ]; subst; trivial;
apply Top_gt; trivial.
(* Top_gt, g is Lex *)
intros s' s'_in_kk2; destruct (equiv_in_list_2 _ equiv s' kk) as [s'' [kk1 [kk2 [s''_eq_s' H'']]]];
trivial.
apply IHk with s''; trivial.
subst kk; rewrite map_app; apply in_or_app; right; left; trivial.
apply H'; subst kk; rewrite map_app; apply in_or_app; right; left; trivial.
(* Top_gt, g is Mul *)
intros s' s'_in_kk2; setoid_rewrite (in_permut_in s' P2) in s'_in_kk2;
destruct (equiv_in_list_2 _ equiv s' kk) as [s'' [kk1 [kk2 [s''_eq_s' H'']]]];
trivial.
apply IHk with s''; trivial.
setoid_rewrite (in_permut_in s'' P1);
subst kk; rewrite map_app; apply in_or_app; right; left; trivial.
apply H'; setoid_rewrite (in_permut_in s'' P1);
subst kk; rewrite map_app; apply in_or_app; right; left; trivial.
(* Top_eq_lex *)
inversion gk_eq_t' as [ | g' kk k1 k2 Sg E_kk | g' kk k1 k2 Sg E_kk P1 P2 ]; subst.
apply Top_eq_lex; trivial.
apply Top_eq_lex; trivial.
generalize kk E_kk l IHl kk1_lt_l; clear kk E_kk l IHl kk1_lt_l IHk H3 gk_eq_t';
intro kk; induction kk as [ | [u1 u2] kk]; simpl.
intros; trivial.
intros E_u1u2_kk l IHl kk1_lt_l; 
inversion kk1_lt_l as [ s t l'' l' u1_lt_t L | s s' l'' l' u1_eq_s' kk1_lt_l']; subst.
apply List_gt.
apply IHl with u1; [ left | apply E_u1u2_kk; left | idtac]; trivial.
rewrite length_map in L; rewrite length_map; trivial.
apply List_eq.
apply (equiv_trans _ _ equiv_equiv) with u1; trivial;
apply (equiv_sym _ _ equiv_equiv); apply E_u1u2_kk; left; trivial.
apply IHkk; trivial.
intros; apply E_u1u2_kk; right; trivial.
intros t t_in_l'; apply IHl; right; trivial.
intros s' s'_in_kk2; 
destruct (equiv_in_list_2 _ equiv s' kk E_kk s'_in_kk2)  as [s'' [kk1 [kk2 [s''_eq_s' H'']]]];
apply IHk with s''; trivial.
subst kk; rewrite map_app; apply in_or_app; right; left; trivial.
apply H3; subst kk; rewrite map_app; apply in_or_app; right; left; trivial.
rewrite Sg in Sf; discriminate.
(* Top_eq_mul *)
inversion gk_eq_t' as [ | g' kk k1 k2 Sg E_kk | g' kk k1 k2 Sg E_kk P1 P2 ]; subst.
apply Top_eq_mul; trivial.
rewrite Sg in Sf; discriminate.
apply Top_eq_mul; trivial;
apply equiv_rpo_equiv_2_aux with k kk; trivial.

intros t t' t_eq_t' s; split; apply H; trivial.
apply (equiv_sym _ _ equiv_equiv); trivial.
Qed.

Lemma equiv_rpo_equiv_3 :
  forall t t', equiv t t' -> (forall s, rpo_eq s t <-> rpo_eq s t').
Proof.
assert (H: forall t t', equiv t t' -> (forall s, rpo_eq s t -> rpo_eq s t')).
intros t t' t_eq_t' s s_le_t; inversion s_le_t; subst.
apply Equiv; apply (equiv_trans _ _ equiv_equiv) with t; trivial.
apply Lt; setoid_rewrite <- (equiv_rpo_equiv_1 _ _ t_eq_t'); trivial.
intros t t' t_eq_t' s; split; apply H; trivial.
apply (equiv_sym _ _ equiv_equiv); trivial.
Qed.

Lemma equiv_rpo_equiv_4 :
  forall t t', equiv t t' -> (forall s, rpo_eq t s <-> rpo_eq t' s).
assert (H: forall t t', equiv t t' -> (forall s, rpo_eq t s -> rpo_eq t' s)).
intros t t' t_eq_t' s t_le_s; inversion t_le_s; subst.
apply Equiv; apply (equiv_trans _ _ equiv_equiv) with t; trivial;
apply (equiv_sym _ _ equiv_equiv); trivial.
apply Lt; setoid_rewrite <- (equiv_rpo_equiv_2 _ _ t_eq_t'); trivial.
intros t t' t_eq_t' s; split; apply H; trivial.
apply (equiv_sym _ _ equiv_equiv); trivial.
Qed.

Lemma rpo_lex_same_length :
  forall l l', rpo_lex l l' -> length l = length l'.
Proof.
induction l; intros l' rpo_lex_l; inversion rpo_lex_l.
simpl; rewrite H3; trivial.
simpl; rewrite (IHl l'0); trivial.
Qed.

(** ** rpo is a preorder, and its reflexive closure is an ordering. *)
Lemma rpo_trans : forall u t s, rpo u t -> rpo t s ->  rpo u s.
Proof.
intros u t s;
cut (forall triple : term * (term * term),
       match triple with
       | (s,(t,u)) => rpo t s -> rpo u t -> rpo u s
       end).
intros H u_lt_t t_lt_s; apply (H (s,(t,u))); trivial.
clear s t u; intro triple; pattern triple; 
refine (well_founded_ind wf_size3 _ _ triple); clear triple.
intros [[v | f l] [t u]] IH.
(* Variable case *)
intros t_lt_v; inversion t_lt_v.
(* Compound case *)
intros t_lt_fl u_lt_t;
inversion t_lt_fl as [ f'' l' s' t' t'_in_l' t_le_t'
                               | f'' f' k'' l' f'_prec_f H''
                               | f'' k'' l' Sf l'_lt_l H H1 H2 
                               | f'' k'' l' Sf l'_lt_l ]; subst.
(* Subterm *)
apply Subterm with t'; trivial; apply Lt;
inversion t_le_t' as [t'' t''' t_eq_t' | t'' t''' t_lt_t']; subst.
setoid_rewrite <- (equiv_rpo_equiv_1 _ _ t_eq_t'); trivial.
apply (IH (t',(t,u))); trivial; apply lex1; trivial.
(* Top_gt *)
inversion u_lt_t as [ f'' l'' s' t' t'_in_l' u_le_t'
                               | g'' f'' k'' l'' f''_prec_f' H'''
                               | g'' k'' l'' Sf' l''_lt_l' H H1 H2 
                               | g'' k'' l'' Sf' l''_lt_l' ]; subst.
(* Top_gt, Subterm *)
inversion u_le_t' as [t'' t''' u_eq_t' | t'' t''' u_lt_t']; subst.
setoid_rewrite (equiv_rpo_equiv_2 _ _ u_eq_t'); trivial; apply H''; trivial.
apply (IH (Term f l,(t',u))); trivial.
apply lex2; trivial.
apply H''; trivial.
(* Top_gt, Top_gt *)
apply Top_gt.
apply prec_transitive with f'; trivial.
intros u u_in_l''; apply (IH (Term f l, (Term f' l', u))); trivial.
apply lex3; trivial.
apply H'''; trivial.
(* Top_gt, Top_eq_lex *)
apply Top_gt; trivial.
intros u u_in_l''; apply (IH (Term f l, (Term f' l', u))); trivial.
apply lex3; trivial.
apply H0; trivial.
(* Top_gt, Top_eq_mul *)
apply Top_gt; trivial.
intros u u_in_l''; apply (IH (Term f l, (Term f' l', u))); trivial.
apply lex3; trivial.
apply rpo_subterm with (Term f' l''); trivial.
(* Top_eq_lex *)
inversion u_lt_t as [ f'' l'' s' t' t'_in_l' u_le_t'
                               | g'' f'' k'' l'' f''_prec_f' H'''
                               | g'' k'' l'' Sf' l''_lt_l' H H1 H2 
                               | g'' k'' l'' Sf' l''_lt_l' ]; subst.
(* Top_eq_lex, Subterm *)
inversion u_le_t' as [t'' t''' u_eq_t' | t'' t''' u_lt_t']; subst.
setoid_rewrite (equiv_rpo_equiv_2 _ _ u_eq_t');
apply rpo_subterm with (Term f l'); trivial.
apply (IH (Term f l, (t', u))); trivial.
apply lex2; trivial.
apply H0; trivial.
(* Top_eq_lex, Top_gt *)
apply Top_gt; trivial.
intros u u_in_l''; apply (IH (Term f l, (Term f l', u))); trivial.
apply lex3; trivial.
apply H'''; trivial.
(* Top_eq_lex, Top_eq_lex *)
apply Top_eq_lex; trivial.
generalize l' l'' l'_lt_l l''_lt_l' IH; clear l' l'' l'_lt_l l''_lt_l' IH H0 H3 t_lt_fl u_lt_t;
induction l as [ | s l]; intros l' l'' l'_lt_l l''_lt_l' IH.
inversion l'_lt_l.
inversion l'_lt_l as [ t s' k' k t_lt_s L | t s' k' k t_eq_s k'_lt_k];
inversion l''_lt_l' as [ u t' k'' h' u_lt_t L' | u t' k'' h' u_eq_t k''_lt_k'].
subst; injection H3; intros; subst; apply List_gt.
apply (IH (s,(t,u))); trivial.
apply lex1; left; trivial.
rewrite L'; trivial.
subst; injection H3; intros; subst; apply List_gt.
setoid_rewrite (equiv_rpo_equiv_2 _ _ u_eq_t); trivial.
rewrite (rpo_lex_same_length _ _ k''_lt_k'); trivial.
subst; injection H3; intros; subst; apply List_gt.
setoid_rewrite <- (equiv_rpo_equiv_1 _ _ t_eq_s); trivial.
rewrite <- (rpo_lex_same_length _ _ k'_lt_k); trivial.
subst; injection H3; intros; subst; apply List_eq.
apply (equiv_trans _ _ equiv_equiv) with t; trivial.
apply IHl with k'; trivial.
intros; apply IH;
apply o_size3_trans with (Term f l, (Term f k', Term f k'')); trivial; apply lex1_bis.
intros u u_in_l''; apply (IH (Term f l, (Term f l', u))); trivial.
apply lex3; trivial.
apply H3; trivial.
(* Top_eq_lex, Top_eq_mul *)
rewrite Sf in Sf'; discriminate.
(* Top_eq_mul *)
inversion u_lt_t as [ f'' l'' s' t' t'_in_l' u_le_t'
                               | g'' f'' k'' l'' f''_prec_f' H'''
                               | g'' k'' l'' Sf' l''_lt_l' H H1 H2 
                               | g'' k'' l'' Sf' l''_lt_l' ]; subst.
(* Top_mul_lex, Subterm *)
inversion u_le_t' as [t'' t''' u_eq_t' | t'' t''' u_lt_t']; subst.
setoid_rewrite (equiv_rpo_equiv_2 _ _ u_eq_t');
apply rpo_subterm with (Term f l'); trivial.
apply (IH (Term f l, (t', u))); trivial.
apply lex2; trivial.
apply rpo_subterm with (Term f l'); trivial.
(* Top_eq_mul, Top_gt *)
apply Top_gt; trivial.
intros u u_in_l''; apply (IH (Term f l, (Term f l', u))); trivial.
apply lex3; trivial.
apply H'''; trivial.
(* Top_eq_mul, Top_eq_lex *)
rewrite Sf in Sf'; discriminate.
(* Top_eq_mul, Top_eq_mul *)
apply Top_eq_mul; trivial.
destruct l'_lt_l as [a lg ls llc l l' E_llc P' P ls_lt_alg].
destruct l''_lt_l' as [a' lg' ls' llc' l' l'' E_llc' Q' Q ls'_lt_alg'].
setoid_rewrite P' in Q; rewrite app_comm_cons in Q.
destruct (ac_syntactic _ _ _ _ Q) as [k1 [k2 [k3 [k4 [P1 [P2 [P3 P4]]]]]]].
destruct (split_equiv equiv_equiv llc k1 k2 E_llc P1) 
  as [kk1 [kk2 [E_kk1 [E_kk2 [Pk1 [Pk2 Pk1k2]]]]]].
destruct (split_equiv_2 equiv_equiv llc' k1 k3 E_llc' P3) 
  as [kk1' [kk3 [E_kk1' [E_kk3 [Pk1' [Pk3 Pk1k3]]]]]].
destruct (trans_equiv_list equiv kk1' kk1) as [kk1'' [E_kk1'' [R R']]]; trivial.
intros e1 _; apply (equiv_trans _ _ equiv_equiv).
setoid_rewrite <- Pk1'; trivial.
apply (List_mul a (lg ++ (map (fun st => snd st) kk2)) 
                        (ls' ++ map (fun st => fst st) kk3) kk1''); trivial.
setoid_rewrite Q'; rewrite <- ass_app; setoid_rewrite <- permut_app1;
apply list_permut_trans with (map (fun st : elt * elt => fst st) (kk3 ++ kk1')).
setoid_rewrite Pk1k3; do 2 rewrite map_app; apply list_permut_app_app.
rewrite map_app; setoid_rewrite <- permut_app1; trivial.
setoid_rewrite P; setoid_rewrite <- permut_cons; rewrite <- ass_app;
setoid_rewrite <- permut_app1; 
apply list_permut_trans with (map (fun st : elt * elt => snd st) (kk2 ++ kk1)).
setoid_rewrite Pk1k2; do 2 rewrite map_app; apply list_permut_app_app.
rewrite map_app; setoid_rewrite <- permut_app1; trivial.
intros b b_in_ls'_fst_kk3; 
destruct (in_app_or _ _ _ b_in_ls'_fst_kk3) as [b_in_ls' | b_in_fst_kk3].
destruct (ls'_lt_alg'  _ b_in_ls') as [a'' [a''_in_a'lg' b_lt_a'']].
setoid_rewrite (in_permut_in a'' P4) in a''_in_a'lg';
destruct (in_app_or _ _ _ a''_in_a'lg') as [a''_in_k2 | a''_in_k4].
setoid_rewrite (in_permut_in a'' Pk2) in a''_in_k2;
destruct (equiv_in_list _ equiv a'' kk2) as [a3 [kk2' [kk2'' [a''_eq_a3 H]]]]; trivial.
exists a3; split.
right; apply in_or_app; right; subst; rewrite map_app;
apply in_or_app; right; left; trivial.
setoid_rewrite (equiv_rpo_equiv_1 a3 a''); trivial; 
apply (equiv_sym _ _ equiv_equiv); trivial.
destruct (ls_lt_alg a'') as [a3 [a3_in_alg a''_lt_a3]].
setoid_rewrite (in_permut_in a'' P2); apply in_or_app; right; trivial.
exists a3; split.
rewrite app_comm_cons; apply in_or_app; left; trivial.
apply (IH (a3,(a'',b))); trivial; apply lex1;
setoid_rewrite (in_permut_in a3 P);
rewrite app_comm_cons; apply in_or_app; left; trivial.
destruct (equiv_in_list _ equiv b kk3) as [b' [kk3' [kk3'' [b_eq_b' H]]]]; trivial.
destruct (ls_lt_alg b') as [a'' [a''_in_alg b'_lt_a'']].
setoid_rewrite (in_permut_in b' P2); apply in_or_app; left; 
setoid_rewrite Pk3; subst; rewrite map_app; apply in_or_app; right; left; trivial.
exists a''; split.
rewrite app_comm_cons; apply in_or_app; left; trivial.
setoid_rewrite (equiv_rpo_equiv_2 b b' b_eq_b'); trivial.
Qed.

Lemma rpo_mul_remove_equiv_aux :
  forall l l' s s', (forall t, In t (s :: l) -> rpo t t -> False) -> 
                      equiv s s' -> rpo_mul (s :: l) (s' :: l') -> rpo_mul l l'. 
Proof.
intros l l' s s' Antirefl s_eq_s' sl_lt_s'l';  
inversion sl_lt_s'l' as [a lg ls llc k' k E_llc P P' ls_lt_alg]; subst.
assert (s_in_ls_llc1 : In s (ls ++ map (fun st => fst st) llc)).
setoid_rewrite <- (in_permut_in s P); left; trivial.
assert (s'_in_alg_llc2 : In s' (a :: lg ++ map (fun st => snd st) llc)).
setoid_rewrite <- (in_permut_in s' P'); left; trivial.
rewrite app_comm_cons in s'_in_alg_llc2;
destruct (in_app_or _ _ _ s_in_ls_llc1) as [s_in_ls | s_in_llc1]; clear s_in_ls_llc1;
destruct (in_app_or _ _ _ s'_in_alg_llc2) as [s'_in_alg | s'_in_llc2]; clear s'_in_alg_llc2.
(* s in in ls, s' is in (a :: lg) *)
destruct lg as [ | a' lg].
destruct s'_in_alg as [s'_eq_a | s'_in_nil]; [subst a | contradiction].
destruct (ls_lt_alg s s_in_ls) as [a' [[a'_eq_s | a'_in_nil] s_lt_a']].
subst a'; assert False; [idtac | contradiction]; 
apply (Antirefl s);
[left | setoid_rewrite (equiv_rpo_equiv_1 _ _ s_eq_s')]; trivial.
contradiction.
destruct (In_split _ _ s'_in_alg) as [lg1 [lg2 H]].
assert (H' : exists g, exists lg', lg1 ++ lg2 = g :: lg').
generalize (f_equal (@length term) H); rewrite app_length;
rewrite plus_comm; simpl; rewrite plus_comm; 
rewrite <- app_length; destruct (lg1 ++ lg2) as [ | g lg'].
intro; discriminate.
intros _; exists g; exists lg'; trivial.
destruct H' as [g [lg' H']];
destruct (In_split _ _ s_in_ls) as [ls1 [ls2 H'']];
apply (List_mul g lg' (ls1 ++ ls2) llc); trivial.
subst ls; rewrite <- ass_app in P; simpl in P; 
setoid_rewrite <- permut_cons_inside in P; rewrite ass_app in P; trivial.
rewrite app_comm_cons; rewrite <- H'; rewrite <- ass_app.
setoid_rewrite (permut_cons_inside s'); rewrite app_comm_cons;
rewrite ass_app; rewrite <- H; simpl; simpl in P'; trivial.
intros b b_in_ls1ls2; destruct (ls_lt_alg b) as [a'' [a''_in_aa'lg b_lt_a'']].
subst; apply in_insert; trivial.
rewrite H in a''_in_aa'lg.
assert (a''_in_s'_lg1_lg2 : a'' = s' \/ In a'' (lg1 ++ lg2)). 
destruct (in_app_or _ _ _ a''_in_aa'lg) as [a''_in_lg1 | a''_in_s'_lg2].
right; apply in_or_app; left; trivial.
destruct a''_in_s'_lg2 as [a''_eq_s' | a''_in_lg2].
left; subst; trivial.
right; apply in_or_app; right; trivial.
destruct (a''_in_s'_lg1_lg2) as [a''_eq_s' | a''_in_lg1_lg2].
subst a''; destruct (ls_lt_alg s) as [s'' [s''_in_alg s_lt_s'']]; trivial;
exists s''; split.
rewrite <- H'; rewrite H in s''_in_alg;
destruct (in_app_or _ _ _ s''_in_alg) as [s''_in_lg1 | [s''_eq_s' | s''_in_lg2]].
apply in_or_app; left; trivial.
subst s''; assert False; [idtac | contradiction];
apply (Antirefl s).
left; trivial.
setoid_rewrite (equiv_rpo_equiv_1 s s' s_eq_s'); trivial.
apply in_or_app; right; trivial.
apply rpo_trans with s; trivial.
setoid_rewrite (equiv_rpo_equiv_1 s s' s_eq_s'); trivial.
exists a''; split; trivial; rewrite <- H'; trivial.
(* s is in ls, s' is in llc *)
destruct (equiv_in_list_2 _ equiv s' llc E_llc s'_in_llc2) as [s'' [llc1 [llc2 [s''_eq_s' H']]]].
destruct (In_split _ _ s_in_ls) as [ls1 [ls2 H]].
apply (List_mul a lg (s'' :: ls1 ++ ls2) (llc1 ++ llc2)).
intros; apply E_llc; subst llc; apply in_insert; trivial.
setoid_rewrite (permut_cons s); setoid_rewrite P;
subst ls; rewrite <- ass_app; simpl; apply list_permut_sym;
setoid_rewrite <- permut_cons_inside;
subst llc; do 2 rewrite map_app; simpl; do 3 rewrite ass_app;
setoid_rewrite <- permut_cons_inside; auto.
subst llc; rewrite map_app in P'; simpl in P';
rewrite app_comm_cons in P'; rewrite ass_app in P';
setoid_rewrite <- permut_cons_inside in P';
rewrite <- ass_app in P'; rewrite <- map_app in P'; trivial.
intros b [b_eq_s'' | b_in_ls1_ls2].
subst b; destruct (ls_lt_alg s s_in_ls) as [a' [a'_in_alg s_lt_a']];
exists a'; split; trivial;
setoid_rewrite (equiv_rpo_equiv_2 _ _ s''_eq_s');
setoid_rewrite <- (equiv_rpo_equiv_2 _ _ s_eq_s'); trivial.
apply ls_lt_alg; subst; apply in_insert; trivial.
(* s is in llc, s' is in (a :: lg) *)
destruct (equiv_in_list _ equiv s llc E_llc s_in_llc1) as [s'' [llc1 [llc2 [s''_eq_s H']]]].
destruct (In_split _ _ s'_in_alg) as [lg1 [lg2 H]].
apply (List_mul s'' (lg1 ++ lg2) ls (llc1 ++ llc2)).
intros; apply E_llc; subst; apply in_insert; trivial.
rewrite map_app; rewrite ass_app;
setoid_rewrite (permut_cons_inside s); setoid_rewrite P;
subst llc; rewrite map_app; simpl; rewrite ass_app; auto.
setoid_rewrite (permut_cons s'); setoid_rewrite P';
rewrite app_comm_cons; rewrite H; rewrite <- ass_app;
simpl; apply list_permut_sym; setoid_rewrite <-permut_cons_inside;
subst llc; do 2 rewrite map_app; simpl; do 3 rewrite ass_app;
setoid_rewrite <- permut_cons_inside; auto.
intros b b_in_ls; destruct (ls_lt_alg _ b_in_ls) as [a' [a'_in_alg b_lt_a']].
rewrite H in a'_in_alg; assert (a'_in_s'_lg1_lg2 : a' = s' \/ In a' (lg1 ++ lg2)). 
destruct (in_app_or _ _ _ a'_in_alg) as [a'_in_lg1 | a'_in_s'_lg2].
right; apply in_or_app; left; trivial.
destruct a'_in_s'_lg2 as [a'_eq_s' | a'_in_lg2].
left; subst; trivial.
right; apply in_or_app; right; trivial.
destruct (a'_in_s'_lg1_lg2) as [a'_eq_s' | a'_in_lg1_lg2].
subst a'; exists s''; split.
left; trivial.
setoid_rewrite <- (equiv_rpo_equiv_1 s s'' s''_eq_s); 
setoid_rewrite (equiv_rpo_equiv_1 s s' s_eq_s'); trivial.
exists a'; split; trivial; right; trivial.
(* s is in llc1, s' is in llc2 *)
destruct (split_equiv_aux equiv_equiv llc s s' E_llc s_in_llc1 s'_in_llc2 s_eq_s')
  as [llc' [E_llc' [Q Q']]].
apply (List_mul a lg ls llc'); trivial.
setoid_rewrite (permut_cons_inside s); setoid_rewrite P;
setoid_rewrite <- permut_app1; trivial.
rewrite app_comm_cons; 
setoid_rewrite (permut_cons_inside s'); setoid_rewrite P';
rewrite app_comm_cons; setoid_rewrite <- permut_app1; trivial.
Qed.

Lemma rpo_antirefl :
  forall s, rpo s s -> False.
Proof.
intro s; pattern s; apply term_rec3; clear s.
intros v v_lt_v; inversion v_lt_v.
intros f l IH fl_lt_fl;  inversion fl_lt_fl  as [ f' l' s'' s' s'_in_l s_le_s'
                               | f' f'' k' k'' f_prec_f H' H1 H2
                               | f' k' k'' Sf l_lt_l H H1 H2
                               | f' k' k'' Sf l_lt_l ]; clear fl_lt_fl; subst.
(* Antirefl, subterm *)
inversion s_le_s' as [s s'' s_eq_s' | s s'' s_lt_s']; subst.
assert (s'_lt_s := Subterm f l s' s' s'_in_l (Equiv s' s' (Eq _)));
setoid_rewrite (equiv_rpo_equiv_1 _ _ s_eq_s') in s'_lt_s;
apply (IH s'); trivial.
apply (IH s'); trivial; apply rpo_trans with (Term f l); trivial;
apply Subterm with s'; trivial; apply Equiv; apply Eq.
(* Antirefl, Top_gt *)
apply prec_antisym with f; trivial.
(* Antirefl, Top_eq_lex *)
clear H0 f Sf; induction l as [ | s l]; 
inversion l_lt_l as [ | ]; subst.
apply (IH s); trivial; left; trivial.
apply IHl; trivial; intros t t_in_l; apply (IH t); right; trivial.
(* Antirefl, Top_eq_mul *)
clear f Sf; induction l as [ | s l].
inversion  l_lt_l as [a lg ls llc k' k'' E_llc P1 P2 H]; clear l_lt_l; subst.
assert (L' := list_permut_length P2); simpl in L'; discriminate.
apply IHl.
intros t t_in_l; apply IH; right; trivial.
apply (rpo_mul_remove_equiv_aux l l s s); trivial; apply Eq.
Qed.


Lemma rpo_closure :
  forall s t u, 
  (rpo t s -> rpo u t -> rpo u s) /\
  (rpo s t -> rpo t s -> False) /\
  (rpo s s -> False) /\
  (rpo_eq s t -> rpo_eq t s -> equiv s t).
Proof.
intros s t u; repeat split.
intros; apply (rpo_trans u t s); trivial.
intros; apply (rpo_antirefl s); apply rpo_trans with t; trivial.
apply rpo_antirefl.
intros s_le_t t_le_s;
destruct s_le_t as [s t s_eq_t | s t s_lt_t]; trivial.
destruct t_le_s as [t s t_eq_s | t s t_lt_s].
apply (equiv_sym _ _ equiv_equiv); trivial.
assert False; [idtac | contradiction].
apply (rpo_antirefl s); apply rpo_trans with t; trivial.
Qed.

(** ** Well-foundedness of rpo. *)
Lemma equiv_acc_rpo :
  forall t t', equiv t t' -> Acc rpo t -> Acc rpo t'.
Proof.
intros t t' t_eq_t' Acc_t; apply Acc_intro.
intro s; setoid_rewrite <- (equiv_rpo_equiv_1 _ _ t_eq_t').
inversion Acc_t as [H]; apply (H s); trivial.
Qed.

Inductive rpo_lex_rest : list term -> list term -> Prop :=
| Rpo_lex_rest : 
     forall l l', (forall s, In s l -> Acc rpo s) -> (forall s, In s l' -> Acc rpo s) ->
     rpo_lex l' l -> rpo_lex_rest l' l.

Lemma wf_rpo_lex_rest : well_founded rpo_lex_rest.
Proof.
unfold well_founded; intro l; pattern l; apply list_rec2; clear l;
induction n; intros [ | t l ] L.
apply Acc_intro; intros l' H;
inversion H as [ k k' Acc_l Acc_l' H' H1 H2]; subst; inversion H'.
simpl in L; absurd (S (length l) <= 0); auto with arith.
apply Acc_intro; intros l' H;
inversion H as [ k k' Acc_l Acc_l' H' H1 H2]; subst; inversion H'.
simpl in L; generalize (le_S_n _ _ L); clear L; intro L.
assert (H : Acc rpo t -> Acc rpo_lex_rest (t :: l)).
intro Acc_t; generalize l L; clear l L; induction Acc_t as [ s  Acc_s IHs].
intros l L; assert (Acc_l := IHn _ L); 
generalize s Acc_s IHs; clear s Acc_s IHs;
induction Acc_l as [l' Acc_l' IHl'].
intros s Acc_s IHs; apply Acc_intro.
intros l'' H; inversion H as [k' k'' Acc_s_l Acc_l'' H']; subst.
inversion H' as [u s' k' k'' u_lt_s L' | s' s'' k' k'' s'_eq_s k'_lt_l H1 H2].
subst; apply IHs; trivial; rewrite L'; trivial.
subst; apply IHl'. 
apply Rpo_lex_rest; trivial.
intros; apply Acc_s_l; right; trivial.
intros; apply Acc_l''; right; trivial.
rewrite (rpo_lex_same_length k' l'); trivial.
intros u u_lt_s'; apply (Acc_s u);
setoid_rewrite <- (equiv_rpo_equiv_1 _ _ s'_eq_s); trivial.
intros u u_lt_s' l Ll; apply IHs; trivial;
setoid_rewrite <- (equiv_rpo_equiv_1 _ _ s'_eq_s); trivial.
apply Acc_intro; intros l' H'; inversion H' as [H1 H2 Acc_t_l]; subst.
assert (wf_t := Acc_t_l t (or_introl _ (refl_equal _))).
apply (Acc_inv (H wf_t)); trivial.
Qed.

(** Definition of a finer grain for multiset extension. *)
Inductive rpo_mul_step : list term -> list term -> Prop :=
  | List_mul_step : 
       forall a ls llc l l', (forall s t, In (s,t) llc -> equiv s t) ->
        list_permut l' (ls ++ (map (fun st => fst st) llc)) -> 
        list_permut l (a :: (map (fun st => snd st) llc)) ->
       (forall b, In b ls -> rpo b a) ->
       rpo_mul_step l' l.

Lemma equiv_rpo_mul_step_equiv_1 :
  forall l l' ll k, (forall s t, In (s,t) ll -> equiv s t) ->
               list_permut l (map (fun st => fst st) ll) ->
               list_permut l' (map (fun st => snd st) ll) ->
               rpo_mul_step k l -> rpo_mul_step k l'.
Proof.
intros l l' ll k E_ll P P' k_lt_l;
inversion k_lt_l as [a ls llc h h' E_llc Q' Q ls_lt_a]; subst h h'.
assert (a_in_l : In a l).
setoid_rewrite (in_permut_in a Q); left; trivial.
setoid_rewrite (in_permut_in a P) in a_in_l.
destruct (equiv_in_list _ equiv a ll E_ll a_in_l) as [a' [ll1 [ll2 [a_eq_a' H]]]].
setoid_rewrite P in Q; rewrite H in Q; rewrite map_app in Q; simpl in Q;
generalize (list_permut_sym Q); setoid_rewrite <- permut_cons_inside;
rewrite <- map_app; clear Q; intro Q.
destruct (trans_equiv_list equiv llc (ll1 ++ ll2)) as [llc' [E_llc' [R R']]]; trivial.
intros e1 _; apply (equiv_trans _ _ equiv_equiv).
intros; apply E_ll; subst ll; apply in_insert; trivial.
apply (List_mul_step a' ls llc'); trivial. 
setoid_rewrite Q'; setoid_rewrite <- permut_app1; trivial.
setoid_rewrite P'; subst ll; rewrite map_app; simpl; apply list_permut_sym; 
setoid_rewrite <- permut_cons_inside; rewrite <- map_app;
apply list_permut_sym; trivial.
intros b b_in_ls; setoid_rewrite <- (equiv_rpo_equiv_1 _ _ a_eq_a');
apply ls_lt_a; trivial.
Qed.

(** The plain multiset extension is in the transitive closure of
the finer grain extension. *)
Lemma rpo_mul_trans_clos :
  inclusion _ rpo_mul (clos_trans _ rpo_mul_step).
Proof.
unfold inclusion; intros l' l H; 
inversion H as [a lg ls llc k k' E_llc P' P ls_lt_alg]; subst.
generalize l' l a ls llc E_llc P P' ls_lt_alg;
clear l' l a ls llc E_llc P P' ls_lt_alg H;
induction lg as [ | g lg]; intros l' l a ls llc E_llc P P' ls_lt_alg.
apply t_step; apply (List_mul_step a ls llc); trivial.
intros b b_in_ls; destruct (ls_lt_alg b b_in_ls) as [a' [[a'_eq_a | a'_in_nil] b_lt_a']].
subst; trivial.
contradiction.
assert (H: exists ls1, exists ls2, 
 list_permut ls (ls1 ++ ls2) /\
 (forall b, In b ls1 -> rpo b g) /\
 (forall b, In b ls2 -> exists a', In a' (a :: lg) /\ rpo b a')).
clear P'; induction ls as [ | s ls].
exists (nil : list term); exists (nil : list term); intuition.
contradiction.
contradiction.
destruct IHls as [ls1 [ls2 [P' [ls1_lt_g ls2_lt_alg]]]].
intros b b_in_ls; apply ls_lt_alg; right; trivial.
destruct (ls_lt_alg _ (or_introl _ (refl_equal _))) as 
[a' [[a'_eq_a | [a'_eq_g | a'_in_lg]] b_lt_a']].
subst a'; exists ls1; exists (s :: ls2); repeat split; trivial.
setoid_rewrite <- permut_cons_inside; trivial.
intros b [b_eq_s | b_in_ls2].
subst b; exists a; split; trivial; left; trivial.
apply ls2_lt_alg; trivial.
exists (s :: ls1); exists ls2; repeat split; trivial.
simpl; setoid_rewrite <- permut_cons; trivial.
intros b [b_eq_s | b_in_ls1].
subst; trivial.
apply ls1_lt_g; trivial.
exists ls1; exists (s :: ls2); repeat split; trivial.
setoid_rewrite <- permut_cons_inside; trivial.
intros b [b_eq_s | b_in_ls2].
subst b; exists a'; split; trivial; right; trivial.
apply ls2_lt_alg; trivial.
destruct H as [ls1 [ls2 [Pls [ls1_lt_g ls2_lt_alg]]]].
apply t_trans with (g :: ls2 ++ (map (fun st => fst st) llc)).
assert (H : map (fun st => fst st) (map (fun s => (s,s)) ls2) = ls2).
generalize ls2; intro ls'; induction ls' as [ | s ls']; auto;
simpl; rewrite IHls'; trivial.
apply t_step; apply (List_mul_step g ls1 (map (fun s => (s,s)) 
            (ls2 ++ (map (fun st => fst st) llc)))); trivial.
generalize (ls2 ++ map (fun st : term * term => fst st) llc);
intro lc; induction lc as [ | c lc].
intros; contradiction.
intros s t [st_eq_cc | st_in_lc].
injection st_eq_cc; intros; subst s t; apply Eq; trivial.
apply IHlc; trivial.
do 2 rewrite map_app; unfold elt, DS.A, Term_eq_dec.A in *; rewrite H;
rewrite ass_app; do 2 rewrite map_map; simpl; setoid_rewrite P';
setoid_rewrite <- permut_app2; trivial.
do 2 rewrite map_app; do 3 rewrite map_map; simpl;
setoid_rewrite <- permut_cons; setoid_rewrite <- permut_app2;
generalize ls2; intro ls'; induction ls' as [ | s ls']; auto;
simpl; setoid_rewrite <- permut_cons; trivial.
apply (IHlg (g :: ls2 ++ map (fun st : term * term => fst st) llc) l 
                   a ls2 ((g,g) :: llc)); trivial.
intros t1 t2 [t1t2_eq_gg | t1t2_in_llc].
injection t1t2_eq_gg; intros; subst t1 t2; apply Eq.
apply E_llc; trivial.
setoid_rewrite P; setoid_rewrite <- permut_cons;
simpl; setoid_rewrite <- permut_cons_inside; auto.
simpl; setoid_rewrite <- permut_cons_inside; auto.
Qed.

Inductive rpo_mul_rest : list term -> list term -> Prop :=
| Rpo_mul_rest : 
     forall l l', (forall s, In s l -> Acc rpo s) -> (forall s, In s l' -> Acc rpo s) ->
     rpo_mul l' l -> rpo_mul_rest l' l.

Inductive rpo_mul_step_rest : list term -> list term -> Prop :=
| Rpo_mul_step_rest : 
     forall l l', (forall s, In s l -> Acc rpo s) -> (forall s, In s l' -> Acc rpo s) ->
     rpo_mul_step l' l -> rpo_mul_step_rest l' l.

Lemma rpo_mul_rest_trans_clos :
  inclusion _ rpo_mul_rest (clos_trans _ rpo_mul_step_rest).
Proof.
unfold inclusion; intros l' l H; 
inversion H as [k k' Acc_l Acc_l' H' H1 H2 ]; subst.
inversion H' as [a lg ls llc k k' E_llc P' P ls_lt_alg]; subst.
generalize l' l a ls llc E_llc P' P ls_lt_alg Acc_l Acc_l'; 
clear l' l a ls llc E_llc P' P ls_lt_alg H Acc_l Acc_l' H';
induction lg as [ | g lg]; intros l' l a ls llc E_llc P' P ls_lt_alg Acc_l Acc_l'.
apply t_step; apply Rpo_mul_step_rest; trivial;
apply (List_mul_step a ls llc); trivial.
intros b b_in_ls; destruct (ls_lt_alg b b_in_ls) as [a' [[a'_eq_a | a'_in_nil] b_lt_a']].
subst; trivial.
contradiction.
assert (H : exists ls1, exists ls2, 
 list_permut ls (ls1 ++ ls2) /\
 (forall b, In b ls1 -> rpo b g) /\
 (forall b, In b ls2 -> exists a', In a' (a :: lg) /\ rpo b a')).
clear P'; induction ls as [ | s ls].
exists (nil : list term); exists (nil : list term); intuition.
contradiction.
contradiction.
destruct IHls as [ls1 [ls2 [P' [ls1_lt_g ls2_lt_alg]]]].
intros b b_in_ls; apply ls_lt_alg; right; trivial.
destruct (ls_lt_alg _ (or_introl _ (refl_equal _))) as 
[a' [[a'_eq_a | [a'_eq_g | a'_in_lg]] b_lt_a']].
subst a'; exists ls1; exists (s :: ls2); repeat split; trivial.
setoid_rewrite <- permut_cons_inside; trivial.
intros b [b_eq_s | b_in_ls2].
subst b; exists a; split; trivial; left; trivial.
apply ls2_lt_alg; trivial.
exists (s :: ls1); exists ls2; repeat split; trivial.
simpl; setoid_rewrite <- permut_cons; trivial.
intros b [b_eq_s | b_in_ls1].
subst; trivial.
apply ls1_lt_g; trivial.
exists ls1; exists (s :: ls2); repeat split; trivial.
setoid_rewrite <- permut_cons_inside; trivial.
intros b [b_eq_s | b_in_ls2].
subst b; exists a'; split; trivial; right; trivial.
apply ls2_lt_alg; trivial.
destruct H as [ls1 [ls2 [Pls [ls1_lt_g ls2_lt_alg]]]].
apply t_trans with (g :: ls2 ++ (map (fun st => fst st) llc)).
assert (H' : map (fun st => fst st) (map (fun s => (s,s)) ls2) = ls2).
generalize ls2; intro ls'; induction ls' as [ | s ls']; auto;
simpl; rewrite IHls'; trivial.
apply t_step; apply Rpo_mul_step_rest; trivial.
intros s [s_eq_g | s_in_ls2_llc1].
subst s; apply Acc_l; setoid_rewrite (in_permut_in g P); right; left; trivial.
apply Acc_l'; setoid_rewrite (in_permut_in s P');
destruct (in_app_or _ _ _ s_in_ls2_llc1) as [s_in_ls2 | s_in_llc1].
apply in_or_app; left; setoid_rewrite (in_permut_in s Pls);
apply in_or_app; right; trivial.
apply in_or_app; right; trivial.
apply (List_mul_step g ls1 (map (fun s => (s,s)) 
            (ls2 ++ (map (fun st => fst st) llc)))); trivial.
generalize (ls2 ++ map (fun st : term * term => fst st) llc);
intro lc; induction lc as [ | c lc].
intros; contradiction.
intros s t [st_eq_cc | st_in_lc].
injection st_eq_cc; intros; subst s t; apply Eq; trivial.
apply IHlc; trivial.
do 2 rewrite map_app; unfold elt, DS.A, Term_eq_dec.A in *; rewrite H';
rewrite ass_app; do 2 rewrite map_map; simpl; setoid_rewrite P';
setoid_rewrite <- permut_app2; trivial.
do 2 rewrite map_app; do 3 rewrite map_map; simpl.
setoid_rewrite <- permut_cons; setoid_rewrite <- permut_app2.
generalize ls2; intro ls'; induction ls' as [ | s ls']; auto;
simpl; setoid_rewrite <- permut_cons; trivial.
apply (IHlg (g :: ls2 ++ map (fun st : term * term => fst st) llc) l 
                   a ls2 ((g,g) :: llc)); trivial.
intros t1 t2 [t1t2_eq_gg | t1t2_in_llc].
injection t1t2_eq_gg; intros; subst t1 t2; apply Eq.
apply E_llc; trivial.
simpl; setoid_rewrite <- permut_cons_inside; auto.
setoid_rewrite P; setoid_rewrite <- permut_cons;
simpl; setoid_rewrite <- permut_cons_inside; auto.
intros s [s_eq_g | s_in_ls2_llc1].
subst s; apply Acc_l; setoid_rewrite (in_permut_in g P); right; left; trivial.
apply Acc_l'; setoid_rewrite (in_permut_in s P');
destruct (in_app_or _ _ _ s_in_ls2_llc1) as [s_in_ls2 | s_in_llc1].
apply in_or_app; left; setoid_rewrite (in_permut_in s Pls);
apply in_or_app; right; trivial.
apply in_or_app; right; trivial.
Qed.

(** Splitting in two disjoint cases. *)
Lemma two_cases_rpo :
 forall a m n, 
 rpo_mul_step n (a :: m) ->
 (exists a', exists n', equiv a a' /\ list_permut n (a' :: n') /\ 
             rpo_mul_step n' m) \/
 (exists ls, exists mm, (forall s t, In (s,t) mm -> equiv s t) /\
                                   (forall b, In b ls -> rpo b a) /\ 
                                   list_permut n (ls ++ (map (fun st => fst st) mm)) /\
                                   list_permut m (map (fun st => snd st) mm)).
Proof.
intros b m n M; inversion M as [a ls llc l l' E_llc P' P ls_lt_a]; subst.
assert (b_in_a_llc2 : In b (a :: map (fun st : term * term => snd st) llc)).
setoid_rewrite <- (in_permut_in b P); left; trivial.
simpl in b_in_a_llc2; destruct b_in_a_llc2 as [b_eq_a | b_in_llc2].
subst a; right; exists ls; exists llc; repeat split; trivial;
setoid_rewrite (permut_cons b); trivial.
destruct (equiv_in_list_2 _ equiv b _ E_llc b_in_llc2) as [b' [llc1 [llc2 [b'_eq_b H]]]].
subst llc; clear b_in_llc2.
rewrite map_app in P; simpl in P; rewrite app_comm_cons in P;
setoid_rewrite <- permut_cons_inside in P; simpl in P; rewrite <- map_app in P.
left; exists b'; 
exists  (ls ++ map (fun st : term * term => fst st) (llc1 ++ llc2));
repeat split.
apply (equiv_sym _ _ equiv_equiv); trivial.
setoid_rewrite P'; apply list_permut_sym;
do 2 rewrite map_app; simpl; do 2 rewrite ass_app;
setoid_rewrite <- permut_cons_inside; auto.
apply (List_mul_step a ls (llc1 ++ llc2)); auto.
intros; apply E_llc; apply in_insert; trivial.
Qed.

Lemma list_permut_map_acc :
 forall l l', list_permut l l' ->
 Acc rpo_mul_step_rest l ->  Acc rpo_mul_step_rest l'.
Proof.
intros l l' P A1; apply Acc_intro; intros l'' M2.
inversion A1 as [H]; apply H; 
inversion M2 as [k' k'' Acc_l' Acc_l'' H']; subst.
inversion H' as [a ls llc k' k'' E_llc P'' P' ls_lt_a]; subst.
apply Rpo_mul_step_rest; trivial.
intros s s_in_l; apply Acc_l'; setoid_rewrite <- (in_permut_in s P); trivial.
apply (List_mul_step a ls llc); trivial;
apply list_permut_trans with l'; trivial.
Qed.

(** Multiset extension of rpo on accessible terms lists is well-founded. *)
Lemma wf_rpo_mul_rest : well_founded rpo_mul_rest.
Proof.
apply wf_incl with (clos_trans _ rpo_mul_step_rest).
unfold inclusion; apply rpo_mul_rest_trans_clos.
apply wf_clos_trans.
unfold well_founded; intro l; induction l as [ | s l]. 
apply Acc_intro; intros m H; inversion H as [l l' Acc_l Acc_l' H']; subst;
inversion H' as [a ls llc l l' _ _ P]; subst;
absurd (0 = S (length llc)); 
[ discriminate 
| assert (L := list_permut_length P); simpl in L; rewrite length_map in L; trivial ].
rename IHl into Acc_l;
assert (Acc rpo s -> Acc rpo_mul_step_rest (s :: l)).
intro Acc_s; generalize l Acc_l; clear l Acc_l; 
pattern s; apply Acc_ind with term rpo; trivial; clear s Acc_s.
intros s Acc_s IHs l Acc_l; pattern l; 
apply Acc_ind with (list term) rpo_mul_step_rest; trivial; clear l Acc_l.
intros l Acc_l IHl; apply Acc_intro.
intros l' H; inversion H as [s_k k' Acc_s_l Acc_l' H']; subst.
destruct (two_cases_rpo s l l' H') 
              as [[s' [n' [s_eq_s' [P H'']]]] | [ls [llc [E_llc [ls_lt_s [P P']]]]]].
(* First case *)
apply (list_permut_map_acc (s' :: n')).
apply list_permut_sym; trivial.
apply Acc_intro; intros l'' l''_lt_s'_n; apply Acc_inv with (s :: n').
apply IHl; apply Rpo_mul_step_rest; trivial.
intros; apply Acc_s_l; right; trivial.
intros s'' s''_in_n'; apply Acc_l'; setoid_rewrite (in_permut_in s'' P); right; trivial.
apply Rpo_mul_step_rest; trivial.
simpl; intros u [u_eq_s | u_in_n'].
subst u; apply Acc_s_l; left; trivial.
apply Acc_l'; setoid_rewrite (in_permut_in u P); right; trivial.
inversion l''_lt_s'_n; trivial.
apply (equiv_rpo_mul_step_equiv_1 (s' :: n') (s :: n') 
            ((s',s) :: map (fun s => (s,s)) n') l'').
simpl; intros u1 u2 [u1u2_eq_s's | u1u2_in_n'n'].
injection u1u2_eq_s's; intros; subst u1 u2; 
apply (equiv_sym _ _ equiv_equiv); trivial.
generalize n' u1 u2 u1u2_in_n'n'; intro n''; clear u1 u2 u1u2_in_n'n';
induction n'' as [ | s'' n''].
contradiction.
simpl; intros u1 u2 [u1u2_eq_s''s'' | u1u2_in_n''n''].
injection u1u2_eq_s''s''; intros; subst u1 u2; apply Eq.
apply IHn''; trivial.
simpl; rewrite map_map; simpl; setoid_rewrite <- permut_cons;
generalize n'; intro n''; induction n'' as [ | s'' n'']; auto.
simpl; setoid_rewrite <- permut_cons; trivial.
simpl; rewrite map_map; simpl; setoid_rewrite <- permut_cons;
generalize n'; intro n''; induction n'' as [ | s'' n'']; auto.
simpl; setoid_rewrite <- permut_cons; trivial.
inversion l''_lt_s'_n; trivial.
(* Second case *)
apply (list_permut_map_acc (ls ++ map (fun st : term * term => fst st) llc)).
apply list_permut_sym; trivial.
clear P; induction ls as [ | b ls].
simpl; apply Acc_intro.
intros l'' l''_lt_llc1; inversion l''_lt_llc1 as [kkc1 k'' Acc_llc1 Acc_l'' H'']; subst kkc1 k'';
assert (l''_lt_l := equiv_rpo_mul_step_equiv_1 (map (fun st : term * term => fst st) llc)
                   l llc l'' E_llc (list_permut_refl _) P' H'');
apply Acc_inv with l. 
apply Acc_intro; trivial.
apply Rpo_mul_step_rest; trivial; intros; apply Acc_s_l; right; trivial.
simpl; apply IHs.
apply ls_lt_s; left; trivial.
intros; apply IHls; intros; apply ls_lt_s; right; trivial.
apply Acc_intro.
intros l' H'.
apply Acc_inv with (s :: l); trivial.
inversion H' as [k k' Acc_s_l Acc_l']; subst;
apply H; apply Acc_s_l; left; trivial.
Qed.

Inductive rpo_rest : (symbol * list term) -> (symbol * list term) -> Prop :=
 | Top_gt_rest : 
       forall f g l l', prec g f -> 
       (forall s, In s l -> Acc rpo s) -> (forall s, In s l' -> Acc rpo s) ->
       rpo_rest (g, l') (f, l)
  | Top_eq_lex_rest : 
        forall f l l', status f = Lex -> rpo_lex l' l -> 
        (forall s, In s l -> Acc rpo s) -> (forall s, In s l' -> Acc rpo s) ->
        rpo_rest (f, l') (f, l)
  | Top_eq_mul_rest : 
        forall f l l', status f = Mul -> rpo_mul l' l -> 
        (forall s, In s l -> Acc rpo s) -> (forall s, In s l' -> Acc rpo s) ->
        rpo_rest (f, l') (f, l).

Lemma wf_rpo_rest : well_founded prec -> well_founded rpo_rest.
Proof.
intros wf_prec; unfold well_founded; intros [f l]; generalize l; clear l; 
pattern f; apply (well_founded_induction_type wf_prec); clear f.
intros f IHf l; assert (Sf : forall f', f' = f -> status f' = status f).
intros; subst; trivial.
destruct (status f); generalize (Sf _ (refl_equal _)); clear Sf; intro Sf.
pattern l; apply (well_founded_induction_type wf_rpo_lex_rest); clear l.
intros l IHl; apply Acc_intro; intros [g l'] H; 
inversion H as [ f' g' k k' g_prec_f Acc_l Acc_l' 
                         | f' k k' Sf' H' Acc_l Acc_l'
                         | f' k k' Sf' H' Acc_l Acc_l' ]; subst.
apply IHf; trivial.
apply IHl; apply Rpo_lex_rest; trivial.
absurd (Lex = Mul); [discriminate | apply trans_eq with (status f); auto].
pattern l; apply (well_founded_induction_type wf_rpo_mul_rest); clear l.
intros l IHl; apply Acc_intro; intros [g l'] H; 
inversion H as [ f' g' k k' g_prec_f Acc_l Acc_l' 
                         | f' k k' Sf' H' Acc_l Acc_l'
                         | f' k k' Sf' H' Acc_l Acc_l' ]; subst.
apply IHf; trivial.
absurd (Lex = Mul); [discriminate | apply trans_eq with (status f); auto].
apply IHl; apply Rpo_mul_rest; trivial.
Qed.

Lemma acc_build :
  well_founded prec -> forall f_l,
  match f_l with (f, l) => 
  (forall t, In t l -> Acc rpo t) ->  Acc rpo (Term f l)
  end.
Proof.
intros wf_prec f_l; pattern f_l;
apply (well_founded_induction_type (wf_rpo_rest wf_prec)); clear f_l.
intros [f l] IH Acc_l; apply Acc_intro;
intros s; pattern s; apply term_rec2; clear s;
induction n as [ | n].
intros t St; absurd (1 <= 0); auto with arith; 
apply le_trans with (size t); trivial; apply size_ge_one.
intros s Ss H; inversion H as [ f' l' s' t t_in_l H' 
                                                | f' g k l' g_prec_f 
                                                | f' k l' Sf H' H''
                                                | f' k l' Sf H'].
(* Subterm case *)
assert (Acc_t := Acc_l _ t_in_l);
subst; inversion H'; subst; [idtac | apply Acc_inv with t; trivial ].
apply Acc_intro; intro u; setoid_rewrite (equiv_rpo_equiv_1 s t); trivial;
intro; apply Acc_inv with t; trivial.
(* Top gt *)
subst; assert (Acc_l' : forall s, In s l' -> Acc rpo s).
intros; apply IHn.
apply lt_n_Sm_le; apply lt_le_trans with (size (Term g l')); trivial.
apply size_direct_subterm; trivial.
apply rpo_trans with (Term g l'); trivial; apply Subterm with s; trivial;
apply Equiv; apply Eq.
apply (IH (g,l')); trivial.
apply Top_gt_rest; trivial.
(* Top_eq_lex *)
subst; assert (Acc_l' : forall s, In s l' -> Acc rpo s).
intros; apply IHn.
apply lt_n_Sm_le; apply lt_le_trans with (size (Term f l')); trivial.
apply size_direct_subterm; trivial.
apply rpo_trans with (Term f l'); trivial; apply Subterm with s; trivial; 
apply Equiv; apply Eq.
apply (IH (f,l')); trivial.
apply Top_eq_lex_rest; trivial.
(* Top_eq_mul *)
subst; assert (Acc_l' : forall s, In s l' -> Acc rpo s).
intros; apply IHn.
apply lt_n_Sm_le; apply lt_le_trans with (size (Term f l')); trivial.
apply size_direct_subterm; trivial.
apply rpo_trans with (Term f l'); trivial; apply Subterm with s; trivial; 
apply Equiv; apply Eq.
apply (IH (f,l')); trivial.
apply Top_eq_mul_rest; trivial.
Qed.

(** ** Main theorem: when the precedence is well-founded, so is the rpo. *)
Lemma wf_rpo : well_founded prec -> well_founded rpo.
Proof.
intro wf_prec;
unfold well_founded; intro t; pattern t; apply term_rec3; clear t.
intro v; apply Acc_intro; intros t lt_t_s; inversion lt_t_s.
intros f l Acc_subterm; apply (acc_build wf_prec (f,l)) ; trivial.
Qed.

(** ** RPO is compatible with the instanciation by a substitution. *)
Lemma equiv_subst :
  forall t s, equiv s t -> 
  forall sigma, equiv (apply_subst sigma s) (apply_subst sigma t).
Proof.
intro t; pattern t; apply term_rec3; clear t.
intros v s v_eq_s; inversion v_eq_s; subst; intro sigma; apply Eq.
intros f l IHl s fl_eq_s sigma; 
inversion fl_eq_s as [ s' 
                               | f' ll l1 l2 Sf E_ll
                               | f' ll l1 l2 Sf E_ll P P' ].
(* Syntacti equality *)
apply Eq.
(* Lex top symbol *)
subst; simpl; 
apply Eq_lex with (map (fun st => 
                                       (apply_subst sigma (fst st), 
                                        apply_subst sigma (snd st))) ll); trivial.
clear fl_eq_s; induction ll as [ | [s t] ll].
contradiction.
simpl; intros t1 t2 [t1t2_eq_s't' | t1t2_in_ll'].
injection t1t2_eq_s't'; intros; subst t1 t2; 
apply IHl; [ idtac | apply E_ll]; left; trivial.
apply IHll; trivial.
intros; apply E_ll; right; trivial.
intros; apply IHl; trivial; right; trivial.
do 2 rewrite map_map; simpl; trivial.
do 2 rewrite map_map; simpl; trivial.
(* Mul top symbol *)
subst; simpl;
apply Eq_mul with (map (fun st => 
                                       (apply_subst sigma (fst st), 
                                        apply_subst sigma (snd st))) ll); trivial.
generalize l IHl l1 P P'; clear l IHl l1 P P' fl_eq_s;
induction ll as [ | [s t] ll].
contradiction.
simpl; intros l IHl l1 P P' t1 t2 [t1t2_eq_s't' | t1t2_in_ll'].
injection t1t2_eq_s't'; intros; subst t1 t2; apply IHl.
setoid_rewrite (in_permut_in t P'); left; trivial.
apply E_ll; left; trivial.
apply IHll with  (map (fun st : term * term => snd st) ll)
    (map (fun st : term * term => fst st) ll); auto.
intros; apply E_ll; right; trivial.
intros u2 u2_in_ll2 u u_eq_u2 tau.
assert (E_ll' : forall t1 t2 : term, In (t1, t2) ll -> equiv t1 t2).
intros; apply E_ll; right; trivial.
destruct (equiv_in_list_2 _ equiv _  ll E_ll' u2_in_ll2) as [u1 [ll1 [ll2 [u2_eq_u1 H]]]].
apply IHl; trivial;
setoid_rewrite (in_permut_in u2 P'); right; subst ll;
rewrite map_app; apply in_or_app; right; left; trivial.
setoid_rewrite (list_permut_map (apply_subst sigma) P);
do 2 rewrite map_map; simpl; auto.
setoid_rewrite (list_permut_map (apply_subst sigma) P');
do 2 rewrite map_map; simpl; auto.
Qed.

Lemma rpo_subst :
  forall t s, rpo s t -> 
  forall sigma, rpo (apply_subst sigma s) (apply_subst sigma t).
Proof.
intro t; pattern t; apply term_rec3; clear t.
intros v s H; inversion H.
intros f l IHl s; simpl; pattern s; apply term_rec3; clear s.
(* case s = Var v; rpo by Subterm *)
intros v R; inversion R as [ f' l' s' t In_t_l R' H1 H2 | | | ]; subst; 
intro sigma; simpl; apply Subterm with (apply_subst sigma t).
apply in_map; trivial.
inversion R' as [ s' R'' | s' t' R'' ]; subst; simpl;
[ inversion H; apply Equiv; apply Eq | apply Lt; refine (IHl _ In_t_l _ R'' sigma) ].
(* case s = Term f l *)
intros g k IHl' R sigma; 
inversion R as [ f' l' s' t In_t_l R' 
                       | f' g' l' l'' R' R'' 
                       | g' l' k' f_lex Rlex R' H2 H3
                       | g' l' k' f_mul Rmul R' H2 ]; subst.
(* case Subterm *)
apply Subterm with (apply_subst sigma t).
apply in_map; trivial.
inversion R' as [ s' R'' | s' t' R'' ]; subst; 
[ apply Equiv; apply equiv_subst | apply Lt; apply IHl ]; trivial .
(* case Top_gt *)
simpl; apply Top_gt; trivial.
intros s' In_s'; setoid_rewrite in_map_iff in In_s';
destruct In_s' as [s [H1 H2]]; subst.
apply IHl'; trivial; apply R''; trivial.
(* case Top_eq_lex *)
simpl; apply Top_eq_lex; trivial.
generalize l Rlex IHl; clear l R Rlex R' IHl IHl';
induction k as [ | s' k ]; intros l Rlex IHl; inversion Rlex; subst; simpl.
apply List_gt; [ apply IHl; trivial; left | do 2 (rewrite length_map) ]; trivial.
apply List_eq.
apply equiv_subst; trivial.
apply IHk; trivial;
intros; apply IHl; trivial; right; trivial.
intros s' In_s'; setoid_rewrite in_map_iff in In_s'; 
destruct In_s' as [s [In_s H1]]; subst; apply IHl'; trivial;
apply rpo_trans with (Term f k); trivial;
apply Subterm with s; trivial; apply Equiv; apply Eq.
(* case Top_eq_mul *)
simpl; apply Top_eq_mul; trivial;
inversion Rmul as [ a lg ls llc l0 k0 E_llc Pk Pl ls_lt_alg]; subst.
apply (List_mul (apply_subst sigma a) (map (apply_subst sigma) lg)
(map (apply_subst sigma) ls) 
(map (fun st => (apply_subst sigma (fst st), apply_subst sigma (snd st))) llc)).
intros t1 t2 t1t2_in_llc'; rewrite in_map_iff in t1t2_in_llc';
destruct t1t2_in_llc' as [[u v] [H uv_in_llc]];
injection H; intros; subst t1 t2; apply equiv_subst; apply E_llc; trivial.
setoid_rewrite (list_permut_map (apply_subst sigma) Pk);
rewrite map_app; setoid_rewrite <- permut_app1;
do 2 rewrite map_map; simpl; auto.
setoid_rewrite (list_permut_map (apply_subst sigma) Pl); simpl;
rewrite map_app; setoid_rewrite <- permut_cons;
setoid_rewrite <- permut_app1;
do 2 rewrite map_map; simpl; auto.
intros b' In_b'; setoid_rewrite in_map_iff in In_b'.
destruct In_b' as [b [H1 In_b]]; subst; elim (ls_lt_alg b In_b); intros a' [ In_a' H1 ];
exists (apply_subst sigma a'); split.
generalize (in_map (apply_subst sigma) _ _ In_a'); simpl; trivial.
apply IHl; trivial; setoid_rewrite (in_permut_in a' Pl);
elim In_a'; clear In_a'; intro In_a'; subst.
left; trivial.
right; apply in_or_app; left; trivial.
Qed.

(** ** RPO is compatible with adding context. *)
Lemma rpo_add_context :
 forall p ctx s t, rpo s t -> is_a_pos ctx p = true -> 
  rpo (replace_at_pos ctx s p) (replace_at_pos ctx t p).
Proof.
intro p; induction p as [ | i p ]; intros ctx s t R H; trivial;
destruct ctx as [ v | f l ].
discriminate.
assert (Status : forall g, g = f -> status g = status f).
intros; subst; trivial.
do 2 (rewrite replace_at_pos_unfold);
destruct (status f); generalize (Status f (refl_equal _)); clear Status; 
intro Status.
apply Top_eq_lex; trivial.
generalize l i H; clear l i H; induction l as [ | u1 l ]; intros i H.
destruct i; discriminate.
destruct i; simpl in H; simpl.
apply List_gt; trivial; apply IHp; trivial.
apply List_eq; [apply Eq | apply IHl; trivial].
intros s' In_s'; 
assert (H' : exists t', In t' (replace_at_pos_list l t i p) /\ rpo_eq s' t').
generalize l i H In_s'; clear l i H In_s'; 
induction l as [ | u1 l ]; intros i H In_s'.
destruct i; discriminate.
destruct i; elim In_s'; clear In_s'; intro In_s'; subst.
exists (replace_at_pos u1 t p); split; [ left | apply Lt; apply IHp ]; trivial.
exists s'; split; [ right; trivial | apply Equiv; apply Eq ].
exists s'; split; [ left; trivial | apply Equiv; apply Eq ].
elim (IHl i H In_s'); intros t' [H1 H2]; exists t'; split; [ right | idtac ]; trivial.
elim H'; clear H'; intros t' [H1 H2]; apply Subterm with t'; trivial.

apply Top_eq_mul; trivial.
assert (H' : exists l1, exists ui, exists l2, l = l1 ++ ui :: l2 /\ length l1 = i).
generalize i H; clear i H; induction l as [ | u1 l ]; intros i H.
destruct i; discriminate.
destruct i as [ | i ].
exists (nil (A:= term)); exists u1; exists l; split; trivial.
elim (IHl i H); intros l1 [ ui [ l2 [ H1 H2]]].
exists (u1 :: l1); exists ui; exists l2; split; subst; trivial.
elim H'; clear H';  intros l1 [ ui [ l2 [ H1 H' ]]]; subst l.
do 2 (rewrite replace_at_pos_list_replace_at_pos_in_subterm; trivial).
assert (map_id : forall l, map (fun x : term => x) l = l).
induction l as [ | u l]; auto.
simpl; rewrite IHl; trivial.
apply (List_mul (replace_at_pos ui t p) nil (replace_at_pos ui s p :: nil) 
          (map (fun s => (s,s)) (l1 ++ l2))).
generalize (l1 ++ l2); intro l; induction l as [ | u l].
contradiction.
intros t1 t2 [t1t2_eq_uu | t1t1_in_ll].
injection t1t2_eq_uu; intros; subst t1 t2; apply Eq.
apply IHl; trivial.
rewrite map_map; simpl; apply list_permut_sym; 
setoid_rewrite <- permut_cons_inside; rewrite map_id; auto.
simpl; apply list_permut_sym; setoid_rewrite <- permut_cons_inside;
rewrite map_map; simpl; rewrite map_id; auto.
intros b In_b; elim In_b; clear In_b; intro In_b; subst.
exists (replace_at_pos ui t p); split; [ left | apply IHp ]; trivial.
induction l1 as [ | u1 l1 ]; trivial; apply IHl1; trivial.
contradiction.
Qed.

Function var_in_term_list (x : variable) (l : list term) 
{ measure (list_size size) l } : bool :=
  match l with
    | nil => false
    | (Var y) :: l => if (X.eq_variable_dec x y) then true else (var_in_term_list x l) 
    | (Term _ ll) :: l => orb (var_in_term_list x ll) (var_in_term_list x l)
end.
Proof.
intros x l t ll y H1 H2 H _; simpl; auto with arith.
intros  _ l t ll f kk H1 H2; simpl; auto with arith.
intros _ l t ll f kk H1 H2; simpl.
subst; apply lt_le_trans with (size (Term f kk)).
rewrite size_unfold; simpl; auto with arith.
simpl; auto with arith.
Qed.

Definition prec_eval f1 f2 :=
  if (F.eq_symbol_dec f1 f2) 
  then Equivalent
  else 
     if prec_dec f1 f2 then Less_than
     else 
        if prec_dec f2 f1 then Greater_than
        else Uncomparable.
    
Lemma prec_eval_is_sound :  
  forall f1 f2, 
  match prec_eval f1 f2 with
  | Equivalent => f1 = f2
  | Less_than => prec f1 f2
  | Greater_than => prec f2 f1 
  | Uncomparable => f1 <> f2 /\ ~prec f1 f2 /\ ~prec f2 f1
  end.
Proof.
intros f1 f2; unfold prec_eval;
destruct (F.eq_symbol_dec f1 f2) as [f1_eq_f2 | f1_diff_f2]; trivial.
destruct (prec_dec f1 f2) as [f1_lt_f2 | not_f1_lt_f2]; trivial.
destruct (prec_dec f2 f1) as [f2_lt_f1 | not_f2_lt_f1]; trivial.
repeat split; trivial.
Qed.

Inductive result (A : Set) : Set := 
  | Not_found : result A
  | Not_finished : result A
  | Found : A -> result A.

Function remove_equiv (p : term -> term -> option bool) 
    (t : term) (l : list term) {struct l} : result (list term) :=
     match l with
     | nil => @Not_found _
     | a :: l => 
            match p t a with
            | Some true => (Found _ l)
            | Some false =>
               match remove_equiv p t l  with
               | Found l' => Found _ (a :: l')
               | Not_found => @Not_found _
               | Not_finished => @Not_finished _
               end
             | None => @Not_finished _
             end 
            end.

Function  remove_equiv_list (p : term -> term -> option bool) (l1 l2 : list term) 
  {struct l1} : option ((list term) * (list term)):=
    match l1, l2 with
    | _, nil => Some (l1,l2)
    | nil, _ => Some (l1,l2)
    | a1 :: l1, l2 =>
          match remove_equiv p a1 l2 with
          | Found l2' =>  remove_equiv_list p l1 l2' 
          | Not_found =>
                      match remove_equiv_list p l1 l2 with
                      | Some (l1',l2') => Some (a1 :: l1', l2')
                      | None => None
                      end
          | Not_finished => None
          end
     end.

Function equiv_eval_list (p : term -> term -> option bool) (l1 l2 : list term) 
{struct l1} : option bool := 
     match l1, l2 with
    | nil, nil => Some true
    | (a :: l), (b :: l') => 
              match p a b with
              | Some true => equiv_eval_list p l l'
              | Some false => Some false
              | None => None
              end
         | _, _ => Some false
    end.

Fixpoint equiv_eval (n : nat) (t1 t2 : term) {struct n} : option bool := 
   match n with
   | 0 => None
   | S m =>
     match t1, t2 with
     | Var v1, Var v2 => 
         if X.eq_variable_dec v1 v2 then Some true else Some false
     | Term f1 l1, Term f2 l2 =>
         if F.eq_symbol_dec f1 f2 
         then 
             match status f1 with
             | Lex =>  equiv_eval_list (equiv_eval m) l1 l2
             | Mul => 
                        match remove_equiv_list (equiv_eval m) l1 l2 with
                        | Some (nil,nil) => Some true
                        | Some _ => Some false
                        | None => None
                        end
            end
        else Some false
     | _, _ => Some false
     end
  end.

Lemma equiv_eval_equation :
  forall n t1 t2, equiv_eval n t1 t2 =
   match n with
   | 0 => None
   | S m =>
     match t1, t2 with
     | Var v1, Var v2 => 
         if X.eq_variable_dec v1 v2 then Some true else Some false
     | Term f1 l1, Term f2 l2 =>
         if F.eq_symbol_dec f1 f2 
         then 
             match status f1 with
             | Lex =>  equiv_eval_list (equiv_eval m) l1 l2
             | Mul => 
                        match remove_equiv_list (equiv_eval m) l1 l2 with
                        | Some (nil,nil) => Some true
                        | Some _ => Some false
                        | None => None
                        end
            end
        else Some false
     | _, _ => Some false
     end
  end.
Proof.
intros [ | n] [v1 | f1 l1] [v2 | f2 l2];
unfold equiv_eval; simpl; trivial.
Qed.

Lemma equiv_eval_list_is_sound :
  forall p l1 l2, match equiv_eval_list p l1 l2 with
      | Some true => length l1 = length l2 /\ 
                              (forall t1 t2, In (t1,t2) (combine l1 l2) -> p t1 t2 = Some true)
      | Some false => length l1 <> length l2 \/
                               (exists t1, exists t2, 
                                 In (t1,t2) (combine l1 l2) /\ p t1 t2 = Some false)
      | None => exists t1, exists t2, In (t1,t2) (combine l1 l2) /\
                            p t1 t2 = None
      end.
Proof.
intros p l1; induction l1 as [ | t1 l1]; intros [ | t2 l2]; simpl.
split; trivial; intros; contradiction.
left; discriminate.
left; discriminate.
assert (H : forall u2, u2 = t2 -> p t1 u2 = p t1 t2).
intros; subst; trivial.
destruct (p t1 t2) as [ [ | ] | ]; generalize (H _ (refl_equal _)); clear H; intro H.
generalize (IHl1 l2); destruct (equiv_eval_list p l1 l2) as [ [ | ] | ].
intros [L H']; repeat split.
rewrite L; trivial.
intros t3 t4 [t3t4_eq_t1t2 | t3t4_in_ll].
injection t3t4_eq_t1t2; intros; subst; trivial.
apply H'; trivial.
intros [L | [t3 [t4 [H' H'']]]].
left; intros H''; apply L; injection H''; intros; trivial.
right; exists t3; exists t4; split; trivial; right; trivial.
intros [t3 [t4 [H' H'']]]; exists t3; exists t4; split; trivial; right; trivial.
right; exists t1; exists t2; split; trivial; left; trivial.
exists t1; exists t2; split; trivial; left; trivial.
Qed.

Lemma remove_equiv_is_sound : 
  forall p t l, 
      match remove_equiv p t l with
      | Found l' => exists t', p t t' = Some true /\ list_permut l (t' :: l')
      | Not_found => forall t', In t' l -> p t t' = Some false
      | Not_finished => exists t', In t' l /\ p t t' = None
      end.
intros p t l; 
functional induction (remove_equiv p t l) as 
  [ 
  | H1 a l t' H 
  | H1 a l t' H IH l' H' 
  | H1 a l t' H IH H' 
  | H1 a l t' H IH H' 
  | H1 a l t' H].
simpl; intros; contradiction.
exists a; split; auto.
rewrite H' in IH; destruct IH as [t' [H'' P]]; exists t'; split; trivial;
setoid_rewrite <- (permut_cons_inside a l (t' :: nil) l'); trivial.
rewrite H' in IH; intros t' [a_eq_t' | t'_in_l]; [subst | apply IH]; trivial.
rewrite H' in IH; destruct IH as [t' [t'_in_l ptt'_eq_none]]; 
exists t'; split; trivial; right; trivial.
exists a; split; trivial; left; trivial.
Qed.

Lemma remove_equiv_list_is_sound :
  forall p l1 l2, 
    match remove_equiv_list p l1 l2 with
         | Some (l1',l2') => 
              exists ll, (forall t1 t2, In (t1,t2) ll -> p t1 t2 = Some true) /\
                list_permut l1 (l1' ++ (map (fun st => fst st) ll)) /\
                list_permut l2 (l2' ++ (map (fun st => snd st) ll)) /\
                (forall t1 t2, In t1 l1' -> In t2 l2' -> p t1 t2 = Some false)
         | None => exists t1, exists t2, In t1 l1 /\ In t2 l2 /\ p t1 t2 = None
    end.
Proof.
intros p l1 l2; 
functional induction (remove_equiv_list p l1 l2) as
[ l1
| H1 l2 H2 H3 H4 H'
| H1 l2 t1 l1 H2 H3 H4 H' l2' H IH 
| H1 l2 t1 l1 H2 H3 H4 H' H IH l1' l2' R
| H1 l2 t1 l1 H2 H3 H4 H' H IH R
| H1 l2 t1 l1 H2 H3 H4 H' H].
(* 1/ 6 *)
exists (@nil (term * term)); simpl; intuition; intros; 
rewrite <- app_nil_end; auto.
(* 1/5 *)
destruct l2 as [ | t2 l2].
contradiction.
exists (@nil (term * term)); simpl; intuition; intros; 
rewrite <- app_nil_end; auto.
(* 1/4 *)
assert (K := remove_equiv_is_sound p t1 l2); rewrite H in K;
destruct K as [t2' [pt1t2_eq_true P]];
destruct (remove_equiv_list p l1 l2') as [ [l1'' l2''] | ].
destruct IH as [ll [E_ll [P1 [P2 F]]]]; exists ((t1,t2') :: ll); repeat split.
intros u1 u2 [I | I].
injection I; intros; subst; trivial.
apply E_ll; trivial.
simpl; setoid_rewrite <- permut_cons_inside; trivial.
setoid_rewrite P; simpl; setoid_rewrite <- permut_cons_inside; trivial.
trivial.
destruct IH as [t1' [t2 [t1_in_l1 [t2_in_l2 p1p2_eq_none]]]];
exists t1'; exists t2; repeat split; trivial.
right; trivial.
setoid_rewrite (in_permut_in t2 P); right; trivial.
(* 1/3 *)
assert (K := remove_equiv_is_sound p t1 l2); rewrite H in K.
rewrite R in IH; destruct IH as [ll [E_ll [P1 [P2 F]]]]; 
exists ll; repeat split; auto.
simpl; setoid_rewrite <- permut_cons; trivial.
simpl; intros u1 u2 [u1_eq_t1 | u1_in_l1'] u2_in_l2'.
subst u1; apply K; setoid_rewrite (in_permut_in u2 P2);
apply in_or_app; left; trivial.
apply F; trivial.
(* 1/2 *)
rewrite R in IH; destruct IH as [u1 [u2 [u1_in_l1 [u2_in_l2 pu1u2_eq_none]]]];
exists u1; exists u2; repeat split; trivial; right; trivial.
(* 1/1 *)
assert (K := remove_equiv_is_sound p t1 l2); rewrite H in K.
destruct K as [t2 [t2_in_l2 pt1t2_eq_none]];
exists t1; exists t2; repeat split; trivial; left; trivial.
Qed.

Lemma equiv_eval_is_sound_weak :
  forall n t1 t2, equiv_eval n t1 t2 = Some true -> equiv t1 t2.
intro n; induction n as [ | n].
intros; discriminate.
destruct t1 as [v1 | f1 l1]; destruct t2 as [v2 | f2 l2]; simpl.
destruct (X.eq_variable_dec v1 v2) as [v1_eq_v2 | v1_diff_v2].
subst; intuition; apply Eq.
intro; discriminate.
intro; discriminate.
intro; discriminate.
destruct (F.eq_symbol_dec f1 f2) as [f1_eq_f2 | f1_diff_f2].
assert (Sf1 := f_equal status f1_eq_f2); 
destruct (status f2); subst f2; rewrite Sf1.
intro H; assert (H' := equiv_eval_list_is_sound (equiv_eval n) l1 l2);
rewrite H in H'; destruct H' as [L H'].
apply Eq_lex with (combine l1 l2); trivial.
intros; apply IHn; apply H'; trivial.
generalize l1 l2 L; intro l; induction l as [ | t l]; intros [ | t' l'].
trivial.
intros; discriminate.
intros; discriminate.
intros L'; injection L'; clear L'; intro L'; simpl; rewrite <- (IHl l' L'); trivial.
generalize l2 l1 L; intro l; induction l as [ | t l]; intros [ | t' l'].
trivial.
intros; discriminate.
intros; discriminate.
intros L'; injection L'; clear L'; intro L'; simpl; rewrite <- (IHl l' L'); trivial.
intro H; assert (H' := remove_equiv_list_is_sound (equiv_eval n) l1 l2);
destruct (remove_equiv_list (equiv_eval n) l1 l2) as [ [l1' l2'] | ].
destruct H' as [ll [E_ll [P1 [P2 H']]]];
apply Eq_mul with ll; trivial.
intros; apply IHn; apply E_ll; trivial.
destruct l1'; destruct l2'; try discriminate; simpl in P1; trivial.
destruct l1'; destruct l2'; try discriminate; simpl in P2; trivial.
discriminate.
intro; discriminate.
Qed.

Lemma equiv_eval_list_fully_evaluates :
  forall p l1 l2, (forall t1 t2, In t1 l1 -> In t2 l2 -> p t1 t2 <> None) ->
  equiv_eval_list p l1 l2 <> None.
Proof.
intros p l1 l2 E;
functional induction (equiv_eval_list p l1 l2) as 
[ 
| H1 H2 t1 l1 H3 t2 l2 H4 H IH
| H1 H2 t1 l1 H3 t2 l2 H4 H IH
| H1 H2 t1 l1 H3 t2 l2 H4 H IH
| l1 l2 H1 H2 H3 H4 H].
(* 1/5 *) 
discriminate.
(* 1/4 *)
apply IH; intros u1 u2 u1_in_l1 u2_in_l2; apply E; right; trivial.
(* 1/3 *)
discriminate.
(* 1/2 *)
assert (E' := E t1 t2); rewrite H in E'; apply E'; left; trivial.
(* 1/1 *)
discriminate.
Qed.

Lemma remove_equiv_fully_evaluates :
  forall p t l, (forall t', In t' l -> p t t' <> None) -> 
  remove_equiv p t l <> @Not_finished _.
Proof.
intros p t l E;
functional induction (remove_equiv p t l) as 
  [ 
  | H1 a l t' H 
  | H1 a l t' H IH l' H' 
  | H1 a l t' H IH H' 
  | H1 a l t' H IH H' 
  | H1 a l t' H].
discriminate.
discriminate.
discriminate.
discriminate.
rewrite H' in IH; apply IH; intros; apply E; right; trivial.
assert (H' := E _ (or_introl _ (refl_equal _)) H); contradiction.
Qed.

Lemma remove_equiv_list_fully_evaluates :
  forall p l1 l2, (forall t1 t2, In t1 l1 -> In t2 l2 -> p t1 t2 <> None) ->
  remove_equiv_list p l1 l2 <> None.
Proof.
intros p l1 l2 E; 
functional induction (remove_equiv_list p l1 l2) as
[ l1
| H1 l2 H2 H3 H4 H'
| H1 l2 t1 l1 H2 H3 H4 H' l2' H IH 
| H1 l2 t1 l1 H2 H3 H4 H' H IH l1' l2' R
| H1 l2 t1 l1 H2 H3 H4 H' H IH R
| H1 l2 t1 l1 H2 H3 H4 H' H].
(* 1/6 *)
discriminate.
(* 1/5 *)
discriminate.
(* 1/4 *)
assert (K := remove_equiv_is_sound p t1 l2); rewrite H in K;
destruct K as [t2 [t2_in_l2 P2]];
apply IH; intros u1 u2 u1_in_l1 u2_in_l2'; apply E; trivial.
right; trivial.
setoid_rewrite (in_permut_in u2 P2); right; trivial.
(* 1/3 *)
discriminate.
(* 1/2 *)
rewrite R in IH; apply IH; intros u1 u2 u1_in_l1 u2_in_l2; apply E; trivial;
right; trivial.
(* 1/1 *)
assert (K := remove_equiv_fully_evaluates p t1 l2);
rewrite H in K; intros _; apply K; trivial.
intros t2 t2_in_l2; apply E; trivial; left; trivial.
Qed.

Lemma equiv_eval_terminates :
  forall n t1 t2, size t1 + size t2 <= n -> equiv_eval n t1 t2 <> None.
Proof.
intro n; induction n as [ | n].
(* base case *)
intros t1 t2 St1; 
absurd (1 <= 0); auto with arith; 
apply le_trans with (size t1 + size t2); trivial;
apply le_trans with (1 + size t2);
[apply le_plus_l | apply plus_le_compat_r; apply size_ge_one].
(* induction step *)
intros t1 t2 St; rewrite equiv_eval_equation.
destruct t1 as [ v1 | f1 l1]; destruct t2 as [ v2 | f2 l2].
destruct (X.eq_variable_dec v1 v2); intros; discriminate.
intros; discriminate.
intros; discriminate.
destruct (F.eq_symbol_dec f1 f2) as [f1_eq_f2 | f1_diff_f2]; 
[idtac | intros; discriminate].
assert (H : forall t1 t2 : term, In t1 l1 -> In t2 l2 -> equiv_eval n t1 t2 <> None).
intros t1 t2 t1_in_l1 t2_in_l2; apply IHn.
rewrite size_unfold in St; rewrite <- plus_assoc in St.
rewrite size_unfold in St; simpl in St.
refine (le_trans _ _ _ _ (le_S_n _ _ St)); apply plus_le_compat.
generalize (size_direct_subterm t1 (Term f1 l1) t1_in_l1);
rewrite (size_unfold (Term f1 l1)); simpl; auto with arith.
generalize (size_direct_subterm t2 (Term f1 l2) t2_in_l2);
rewrite (size_unfold (Term f1 l2)); simpl; auto with arith.
assert (Sf1 := f_equal status f1_eq_f2); 
destruct (status f2); subst f2; rewrite Sf1.
apply equiv_eval_list_fully_evaluates; trivial.
assert (H':= remove_equiv_list_fully_evaluates (equiv_eval n) l1 l2 H);
destruct (remove_equiv_list (equiv_eval n) l1 l2) as [ [[ | t1' l1'] [ | t2' l2']] | ];
try discriminate.
intros _; apply H'; trivial.
Qed.

Lemma equiv_eval_is_complete_true :
  forall n t1 t2, size t1 + size t2 <= n -> equiv t1 t2 -> 
     equiv_eval n t1 t2 = Some true.
Proof.
intro n; induction n as [ | n ].
(* base case *)
intros t1 t2 St; absurd (1 <= 0); auto with arith.
refine (le_trans _ _ _ _ St); apply le_trans with (size t1);
[ apply size_ge_one | apply le_plus_l].
(* induction step *)
intros t1 t2 St t1_eq_t2; 
inversion t1_eq_t2 as 
[ t
| f ll l1 l2 Sf E_ll
| f ll l1 l2 Sf E_ll P1 P2]; subst.
(* 1/3 syntactic equality *)
destruct t2 as [v2 | f2 l2]; simpl.
destruct (X.eq_variable_dec v2 v2) as [v2_eq_v2 | v2_diff_v2]; trivial.
absurd (v2 = v2); trivial.
destruct (F.eq_symbol_dec f2 f2) as [f2_eq_f2 | f2_diff_f2].
assert (H : forall t2, In t2 l2 -> equiv_eval n t2 t2 = Some true).
intros t2 t2_in_l2; apply IHn.
apply le_S_n; refine (le_trans _ _ _ _ St);
replace (S (size t2 + size t2)) with (S (size t2) + size t2); trivial;
apply plus_le_compat; [idtac | apply lt_le_weak];
apply size_direct_subterm; trivial.
apply Eq.
destruct (status f2).
(* 1/5 f2 has a Lex status *)
clear St t1_eq_t2; induction l2 as [ | t2 l2]; simpl; trivial.
rewrite (H t2); [rewrite IHl2 | left]; trivial; intros; apply H; right; trivial.
(* 1/4 f2 has a Mul status *)
assert (H' : remove_equiv_list (equiv_eval n) l2 l2 = Some (nil,nil)).
clear St t1_eq_t2; induction l2 as [ | t2 l2]; simpl; trivial.
rewrite (H t2); [rewrite IHl2 | left]; trivial; intros; apply H; right; trivial.
rewrite H'; trivial.
absurd (f2 = f2); trivial.
(* 1/2 Eq_lex *)
assert (H : forall t1 t2, In (t1,t2) ll -> equiv_eval n t1 t2 = Some true).
intros t1 t2 t1t2_in_ll; apply IHn.
apply le_S_n; refine (le_trans _ _ _ _ St);
replace (S (size t1 + size t2)) with (S (size t1) + size t2); trivial;
apply plus_le_compat; [idtac | apply lt_le_weak];
apply size_direct_subterm; simpl; 
rewrite in_map_iff; exists (t1,t2); simpl; split; trivial.
apply E_ll; trivial.
rewrite equiv_eval_equation; rewrite Sf;
destruct (F.eq_symbol_dec f f) as [_ | f_diff_f].
clear E_ll St t1_eq_t2; induction ll as [ | [t1 t2] ll]; simpl; trivial.
rewrite H; [rewrite IHll | left]; trivial; intros; apply H; right; trivial.
absurd (f = f); trivial.
(* 1/1 Eq_mul *)
assert (St' : forall t1 t2, In t1 l1 -> In t2 l2 -> size t1 + size t2 <= n).
intros t1 t2 t1_in_l1 t2_in_l2; apply le_S_n; refine (le_trans _ _ _ _ St);
replace (S (size t1 + size t2)) with (S (size t1) + size t2); trivial;
apply plus_le_compat; [idtac | apply lt_le_weak];
apply size_direct_subterm; simpl; trivial.
assert (T : forall t1 t2, In t1 l1 -> In t2 l2 -> equiv_eval n t1 t2 <> None).
intros t1 t2 t1_in_l1 t2_in_l2; apply equiv_eval_terminates; apply St'; trivial.
assert (H : forall t1 t2, In (t1,t2) ll -> equiv_eval n t1 t2 = Some true).
intros t1 t2 t1t2_in_ll; apply IHn.
apply St';
[ setoid_rewrite (in_permut_in t1 P1) | setoid_rewrite (in_permut_in t2 P2)];
rewrite in_map_iff; exists (t1,t2); simpl; split; trivial.
apply E_ll; trivial.
assert (H' : remove_equiv_list (equiv_eval n) l1 l2 = Some (nil,nil)).
generalize l2 ll H E_ll P1 P2 St' T; clear l2 ll H P1 P2 E_ll St t1_eq_t2 St' T;
induction l1 as [ | t1 l1]; intros l2 ll H E_ll P1 P2 St' T.
destruct ll as [ | [u1 u2] ll].
rewrite (list_permut_nil P2); simpl; trivial.
assert (H' := list_permut_length P1); simpl in H'; discriminate.
simpl; destruct l2 as [ | t2 l2].
destruct ll as [ | [u1 u2] ll].
assert (H' := list_permut_length P1); simpl in H'; discriminate.
assert (H' := list_permut_length P2); simpl in H'; discriminate.
assert (H' := remove_equiv_is_sound (equiv_eval n) t1 (t2 :: l2)).
assert (H'' : exists t2', In t2' (t2 :: l2) /\ equiv_eval n t1 t2' = Some true).
assert (t1_in_ll1 : In t1 (t1 :: l1)).
left; trivial.
setoid_rewrite (in_permut_in t1 P1) in t1_in_ll1.
destruct (equiv_in_list _ equiv t1 ll) as [t2' [ll1 [ll2 [t1_eq_t2' H'']]]]; trivial.
exists t2'; split.
setoid_rewrite (in_permut_in t2' P2); setoid_rewrite in_map_iff;
exists (t1,t2'); simpl; split; trivial; subst ll; apply in_or_app; right; left; trivial.
apply H; subst ll; apply in_or_app; right; left; trivial.
assert (T' := remove_equiv_fully_evaluates (equiv_eval n) t1 (t2 :: l2)).
destruct (remove_equiv (equiv_eval n) t1 (t2 :: l2)) as [ | | l2'].
destruct H'' as [t2' [t2'_in_l2 t1_eq_t2']].
rewrite (H' t2') in t1_eq_t2'; trivial; discriminate.
assert False; [idtac | contradiction].
apply T'; trivial.
intros; apply T; trivial; left; trivial.
destruct H' as [t2' [t1_eq_t2' P2']].
destruct (split_equiv_aux equiv_equiv ll t1 t2' E_ll) as [ll' [E_ll' [Q1 Q2]]]; trivial.
setoid_rewrite <- (in_permut_in t1 P1); left; trivial.
setoid_rewrite <- (in_permut_in t2' P2); 
setoid_rewrite (in_permut_in t2' P2'); left; trivial.
apply (equiv_eval_is_sound_weak n); trivial.
apply (IHl1 l2' ll'); trivial.
intros u1 u2 u1u2_in_ll'; apply IHn.
apply St'.
setoid_rewrite (in_permut_in u1 P1);
setoid_rewrite (in_permut_in u1 Q1); right;
setoid_rewrite in_map_iff; exists (u1,u2); simpl; split; trivial.
setoid_rewrite (in_permut_in u2 P2);
setoid_rewrite (in_permut_in u2 Q2); right;
setoid_rewrite in_map_iff; exists (u1,u2); simpl; split; trivial.
apply E_ll'; trivial.
setoid_rewrite (permut_cons t1); setoid_rewrite P1; trivial.
setoid_rewrite (permut_cons t2');
setoid_rewrite <- P2'; setoid_rewrite P2; trivial.
intros u1 u2 u1_in_l1 u2_in_l2'; apply St'.
right; trivial.
setoid_rewrite (in_permut_in u2 P2'); right; trivial.
intros u1 u2 u1_in_l1 u2_in_l2'; apply T.
right; trivial.
setoid_rewrite (in_permut_in u2 P2'); right; trivial.
rewrite equiv_eval_equation; rewrite Sf; rewrite H'.
destruct (F.eq_symbol_dec f f) as [f_eq_f | f_diff_f]; trivial;
absurd (f = f); trivial.
Qed.

Lemma equiv_eval_is_sound :
  forall n t1 t2, size t1 +size t2 <= n ->
     match equiv_eval n t1 t2 with
     | Some true => equiv t1 t2
     | Some false => ~equiv t1 t2
     | None => False
     end.
Proof.
intros n t1 t2 St;
assert (H := equiv_eval_is_sound_weak n t1 t2);
assert (T := equiv_eval_terminates n t1 t2 St);
assert (H' := equiv_eval_is_complete_true n t1 t2 St);
destruct (equiv_eval n t1 t2) as [ [ | ] | ].
apply H; trivial.
intro t1_eq_t2; assert (H'' := H' t1_eq_t2); discriminate.
apply T; trivial.
Qed.

Fixpoint lexico_eval (p : term -> term -> option comp) (s1 s2 : term)
   (l1 l2 : list term) {struct l1} : option comp :=
    match l1, l2 with
    | nil, nil => Some Equivalent
    | (t1 :: l1'), (t2 :: l2') =>
          match p t1 t2 with
          | Some Equivalent => lexico_eval p s1 s2 l1' l2'
          | Some Greater_than => 
             let check_s1_gt_l2 := 
              list_forall_option
		(fun t2 => 
		   match p s1 t2 with
		     | Some Greater_than => Some true
		     | Some _ => Some false
                     | None => None
                    end) l2 in
            match check_s1_gt_l2 with
            | Some true => Some Greater_than
            | Some false => Some Uncomparable
            | None => None
            end
          | Some Less_than =>
              let check_s2_gt_l1 := 
                list_forall_option
		(fun t1 =>
		   match p t1 s2 with
		     | Some Less_than => Some true
		     | Some _ => Some false
                     | None => None
                    end) l1 in
              match check_s2_gt_l1 with
              | Some true => Some Less_than
              | Some false => Some Uncomparable
              | None => None
              end
         | Some Uncomparable => Some Uncomparable
         | None => None
     end
   | _, _ => Some Uncomparable
end.

Lemma lexico_eval_equation :
  forall p s1 s2 l1 l2, lexico_eval p s1 s2 l1 l2 =
    match l1, l2 with
    | nil, nil => Some Equivalent
    | (t1 :: l1'), (t2 :: l2') =>
          match p t1 t2 with
          | Some Equivalent => lexico_eval p s1 s2 l1' l2'
          | Some Greater_than => 
             let check_s1_gt_l2 := 
              list_forall_option
		(fun t2 => 
		   match p s1 t2 with
		     | Some Greater_than => Some true
		     | Some _ => Some false
                     | None => None
                    end) l2 in
            match check_s1_gt_l2 with
            | Some true => Some Greater_than
            | Some false => Some Uncomparable
            | None => None
            end
          | Some Less_than =>
              let check_s2_gt_l1 := 
                list_forall_option
		(fun t1 =>
		   match p t1 s2 with
		     | Some Less_than => Some true
		     | Some _ => Some false
                     | None => None
                    end) l1 in
              match check_s2_gt_l1 with
              | Some true => Some Less_than
              | Some false => Some Uncomparable
              | None => None
              end
         | Some Uncomparable => Some Uncomparable
         | None => None
     end
   | _, _ => Some Uncomparable
end.
Proof.
intros p s1 s2 [ | t1 l1] [ | t2 l2]; apply refl_equal.
Qed.

Definition mult_eval (p : term -> term -> option comp) (l1 l2 : list elt)  : option comp :=
         let check_l1_gt_l2 :=
           list_forall_option 
	      (fun t2 => 
		 list_exists_option 
		   (fun t1 => 
		      match p t1 t2 with
			| Some Greater_than => Some true
			| Some _ => Some false
                        | None => None
                      end) l1) l2 in
         match check_l1_gt_l2 with
         | None => None
        | Some true => Some Greater_than
	| Some false =>
	  let check_l1_lt_l2 :=
             list_forall_option 
		(fun t1 =>
		   list_exists_option
		     (fun t2 =>
			match p t1 t2 with
			  | Some Less_than => Some true
			  | Some _ => Some false
                          | None => None
                         end) l2) l1 in
          match check_l1_lt_l2 with
	  | Some true => Some Less_than
	  | Some false => Some Uncomparable
          | None => None
          end
end.

Fixpoint rpo_eval (n : nat) (t1 t2 : term) {struct n} : option comp :=
  (match equiv_eval n t1 t2 with
  | None => None
  | Some true => Some Equivalent
  | Some false =>
     (match t1, t2 with
     | Var _, Var _ => Some Uncomparable

     | Var x, (Term _ l) =>
     	    if var_in_term_list x l
     	    then Some Less_than
     	    else Some Uncomparable

     | (Term _ l), Var x =>
     	    if var_in_term_list x l
     	    then Some Greater_than
     	    else Some Uncomparable

     | (Term f1 l1), (Term f2 l2) =>
       (match n with
       | 0 => None
       | S m => 
         let check_l1_gt_t2 :=
                     list_exists_option 
         		  (fun t => match rpo_eval m t t2 with 
                                    | Some Equivalent 
                                    | Some Greater_than => Some true
                                    | Some _ => Some false
	                            | None => None
                                   end) l1 in
          (match check_l1_gt_t2 with
          | None => None
          | Some true => Some Greater_than
          | Some false =>
            let check_l2_gt_t1 :=
                   list_exists_option 
		        (fun t => match rpo_eval m t1 t with
                                       | Some Equivalent 
                                       | Some Less_than => Some true
                                       | Some _ => Some false
                                       | None  => None
                                     end) l2 in
          (match check_l2_gt_t1 with
          | None => None
          | Some true => Some Less_than
          | Some false =>
             (match prec_eval f1 f2 with
		  | Uncomparable => Some Uncomparable
		  | Greater_than =>
		       let check_l2_lt_t1 :=
                          list_forall_option
			      (fun t => match rpo_eval m t1 t with
                                               | Some Greater_than => Some true
                                               | Some _ => Some false
        		                       | None => None
                                               end) l2 in
                     (match check_l2_lt_t1 with
                     | None => None
                     | Some true => Some Greater_than
                     | Some false => Some Uncomparable
                    end)
		  | Less_than =>
                      let check_l1_lt_t2 :=
		          list_forall_option
			    (fun t => match rpo_eval m t t2 with
                                               | Some Less_than => Some true
                                               | Some _ => Some false
                                               | None => None
                                               end) l1 in
                      (match check_l1_lt_t2 with
                      | None => None
		      | Some true => Some Less_than
		      | Some false => Some Uncomparable
                    end)
		  | Equivalent =>
			(match status f1 with
			  | Mul => 
                                match remove_equiv_list (equiv_eval m) l1 l2 with
                                | None => None
                                | Some (nil, nil) => Some Equivalent
                                | Some (nil, _ :: _) => Some Less_than
                                | Some (_ :: _,nil) => Some Greater_than
                                | Some (l1, l2) => mult_eval (rpo_eval m) l1 l2
                                end
			  | Lex => lexico_eval (rpo_eval m) t1 t2 l1 l2
                       end) 
          end)
        end)
      end)
    end)
  end)
end).

Lemma rpo_eval_equation :
  forall n t1 t2, rpo_eval n t1 t2 =
  match equiv_eval n t1 t2 with
  | None => None
  | Some true => Some Equivalent
  | Some false =>
     (match t1, t2 with
     | Var _, Var _ => Some Uncomparable

     | Var x, (Term _ l) =>
     	    if var_in_term_list x l
     	    then Some Less_than
     	    else Some Uncomparable

     | (Term _ l), Var x =>
     	    if var_in_term_list x l
     	    then Some Greater_than
     	    else Some Uncomparable

     | (Term f1 l1), (Term f2 l2) =>
       (match n with
       | 0 => None
       | S m => 
         let check_l1_gt_t2 :=
                     list_exists_option 
         		  (fun t => match rpo_eval m t t2 with 
                                    | Some Equivalent 
                                    | Some Greater_than => Some true
                                    | Some _ => Some false
	                            | None => None
                                   end) l1 in
          (match check_l1_gt_t2 with
          | None => None
          | Some true => Some Greater_than
          | Some false =>
            let check_l2_gt_t1 :=
                   list_exists_option 
		        (fun t => match rpo_eval m t1 t with
                                       | Some Equivalent 
                                       | Some Less_than => Some true
                                       | Some _ => Some false
                                       | None  => None
                                     end) l2 in
          (match check_l2_gt_t1 with
          | None => None
          | Some true => Some Less_than
          | Some false =>
             (match prec_eval f1 f2 with
		  | Uncomparable => Some Uncomparable
		  | Greater_than =>
		       let check_l2_lt_t1 :=
                          list_forall_option
			      (fun t => match rpo_eval m t1 t with
                                               | Some Greater_than => Some true
                                               | Some _ => Some false
        		                       | None => None
                                               end) l2 in
                     (match check_l2_lt_t1 with
                     | None => None
                     | Some true => Some Greater_than
                     | Some false => Some Uncomparable
                    end)
		  | Less_than =>
                      let check_l1_lt_t2 :=
		          list_forall_option
			    (fun t => match rpo_eval m t t2 with
                                               | Some Less_than => Some true
                                               | Some _ => Some false
                                               | None => None
                                               end) l1 in
                      (match check_l1_lt_t2 with
                      | None => None
		      | Some true => Some Less_than
		      | Some false => Some Uncomparable
                    end)
		  | Equivalent =>
			(match status f1 with
			  | Mul => 
                                match remove_equiv_list (equiv_eval m) l1 l2 with
                                | None => None
                                | Some (nil, nil) => Some Equivalent
                                | Some (nil, _ :: _) => Some Less_than
                                | Some (_ :: _,nil) => Some Greater_than
                                | Some (l1, l2) => mult_eval (rpo_eval m) l1 l2
                                end
			  | Lex => lexico_eval (rpo_eval m) t1 t2 l1 l2
                       end) 
          end)
        end)
      end)
    end)
  end)
end.
Proof.
intros [ | n] [v1 | f1 l1] [v2 | f2 l2];
unfold rpo_eval; simpl; trivial.
Qed.

Lemma lexico_eval_is_sound :
  forall (p : term -> term -> option comp) s1 s2 l1 l2,
           length l1 = length l2 -> 
           match lexico_eval p s1 s2 l1 l2 with
           | Some Equivalent => 
             (exists ll, (forall t1 t2, In (t1,t2) ll -> p t1 t2 = Some Equivalent) /\
                            l1 = map (fun st => fst st) ll /\
                            l2 = map (fun st => snd st) ll) 
             |  Some Less_than => 
                 (exists ll, exists t1, exists t2, exists l1', exists l2',
                   (forall t1 t2, In (t1,t2) ll -> p t1 t2 = Some Equivalent) /\
                   p t1 t2 = Some Less_than /\
                   (forall t1, In t1 l1 -> 
                   ((exists t2, In t2 l2 /\ (p t1 t2 = Some Equivalent \/
                                                     p t1 t2 = Some Less_than)) \/ 
                                                  p t1 s2 = Some Less_than)) /\
                   l1 = map (fun st => fst st) ll ++ t1 :: l1' /\
                   l2 = map (fun st => snd st) ll ++ t2 :: l2')
             |  Some Greater_than => 
                 (exists ll, exists t1, exists t2, exists l1', exists l2',
                   (forall t1 t2, In (t1,t2) ll -> p t1 t2 = Some Equivalent) /\
                   p t1 t2 = Some Greater_than /\
                   (forall t2, In t2 l2 -> 
                   ((exists t1, In t1 l1 /\ (p t1 t2 = Some Equivalent \/
                                                     p t1 t2 = Some Greater_than)) \/ 
                                                  p s1 t2 = Some Greater_than)) /\
                   l1 = map (fun st => fst st) ll ++ t1 :: l1' /\
                   l2 = map (fun st => snd st) ll ++ t2 :: l2')
            | _ => True
end.
Proof. 
intros p s1 s2 l1; induction l1 as [ | t1 l1]; intros [ | t2 l2] L.
simpl; exists (@nil (term * term)); simpl; intuition.
simpl; trivial.
simpl; trivial.
simpl in L; injection L; clear L; intro L; simpl.
assert (P : forall u1, u1 = t1 -> p u1 t2 = p t1 t2).
intros; subst; trivial.
destruct (p t1 t2) as [ [ | | | ] | ];
generalize (P _ (refl_equal _)); clear P; intro P.
generalize (IHl1 _ L); 
destruct (lexico_eval p s1 s2 l1 l2) as [ [ | | | ] | ].
intros [ll [E_ll [H1 H2]]]; exists ((t1,t2) :: ll); simpl; repeat split.
intros t3 t4 [H | H].
injection H; intros; subst; trivial.
apply E_ll; trivial.
rewrite H1; trivial.
rewrite H2; trivial.
intros [ll [u1 [u2 [l1' [l2' [E_ll [P' [H [H1 H2]]]]]]]]].
exists ((t1,t2) :: ll); exists u1; exists u2; exists l1'; exists l2';
repeat split; trivial.
intros t3 t4 [H' | H'].
injection H'; intros; subst; trivial.
apply E_ll; trivial.
intros t3 [t1_eq_t3 | t3_in_l1].
subst t3.
left; exists t2; split; trivial; left; trivial.
destruct (H _ t3_in_l1) as [[t4 [t4_in_l2 H']] | H'].
left; exists t4; split; trivial; right; trivial.
right; trivial.
simpl; rewrite H1; trivial.
simpl; rewrite H2; trivial.
intros [ll [u1 [u2 [l1' [l2' [E_ll [P' [H [H1 H2]]]]]]]]].
exists ((t1,t2) :: ll); exists u1; exists u2; exists l1'; exists l2';
repeat split; trivial.
intros t3 t4 [H' | H'].
injection H'; intros; subst; trivial.
apply E_ll; trivial.
intros t3 [t1_eq_t3 | t3_in_l1].
subst t3.
left; exists t1; split; trivial; left; trivial.
destruct (H _ t3_in_l1) as [[t4 [t4_in_l2 H']] | H'].
left; exists t4; split; trivial; right; trivial.
right; trivial.
simpl; rewrite H1; trivial.
simpl; rewrite H2; trivial.
trivial.
trivial.
assert (P' : forall u1, u1 = t1 -> p u1 s2 = p t1 s2).
intros; subst; trivial.
destruct (p t1 s2) as [ [ | | | ] | ];
generalize (P' _ (refl_equal _)); clear P'; intro P';
generalize (list_forall_option_is_sound
        (fun t3 : term =>
         match p t3 s2 with
         | Some Equivalent => Some false
         | Some Less_than => Some true
         | Some Greater_than => Some false
         | Some Uncomparable => Some false
         | None => None (A:=bool)
         end) l1);
destruct (list_forall_option
        (fun t3 : term =>
         match p t3 s2 with
         | Some Equivalent => Some false
         | Some Less_than => Some true
         | Some Greater_than => Some false
         | Some Uncomparable => Some false
         | None => None (A:=bool)
         end) l1) as [ [ | ] | ]; trivial.
intros H.
exists (@nil (term * term)); exists t1; exists t2; exists l1; exists l2;
repeat split; trivial.
simpl; tauto.
intros u1 [u1_eq_t1 | u1_in_l1].
left; exists t2; split.
left; trivial.
subst; right; trivial.
right; generalize (H _ u1_in_l1);
destruct (p u1 s2) as [ [ | | | ] | ]; trivial; intros; discriminate.
assert (P' : forall u2, u2 = t2 -> p s1 u2 = p s1 t2).
intros; subst; trivial.
destruct (p s1 t2) as [ [ | | | ] | ];
generalize (P' _ (refl_equal _)); clear P'; intro P';
generalize (list_forall_option_is_sound
        (fun t3 : term =>
         match p s1 t3 with
         | Some Equivalent => Some false
         | Some Less_than => Some false
         | Some Greater_than => Some true
         | Some Uncomparable => Some false
         | None => None (A:=bool)
         end) l2);
destruct (list_forall_option
        (fun t3 : term =>
         match p s1 t3 with
         | Some Equivalent => Some false
         | Some Less_than => Some false
         | Some Greater_than => Some true
         | Some Uncomparable => Some false
         | None => None (A:=bool)
         end) l2) as [ [ | ] | ]; trivial.
intros H.
exists (@nil (term * term)); exists t1; exists t2; exists l1; exists l2;
repeat split; trivial.
simpl; tauto.
intros u2 [u2_eq_t2 | u2_in_l2].
left; exists t1; split.
left; trivial.
subst; right; trivial.
right; generalize (H _ u2_in_l2);
destruct (p s1 u2) as [ [ | | | ] | ]; trivial; intros; discriminate.
trivial.
trivial.
Qed.

Lemma mult_eval_is_sound_weak :
  forall p l1 l2, 
   match mult_eval p l1 l2 with
     | Some Equivalent => False
     | Some Less_than =>  
       forall t1, In t1 l1 -> exists t2, In t2 l2 /\ p t1 t2 = Some Less_than
     | Some Greater_than =>  
       forall t2, In t2 l2 -> exists t1, In t1 l1 /\ p t1 t2 = Some Greater_than
     | _ => True
     end.
Proof.
intros p l1 l2; unfold mult_eval.
generalize (list_forall_option_is_sound
      (fun t2 : term =>
       list_exists_option
         (fun t1 : term =>
          match p t1 t2 with
          | Some Equivalent => Some false
          | Some Less_than => Some false
          | Some Greater_than => Some true
          | Some Uncomparable => Some false
          | None => None (A:=bool)
          end) l1) l2);
destruct (list_forall_option
      (fun t2 : term =>
       list_exists_option
         (fun t1 : term =>
          match p t1 t2 with
          | Some Equivalent => Some false
          | Some Less_than => Some false
          | Some Greater_than => Some true
          | Some Uncomparable => Some false
          | None => None (A:=bool)
          end) l1) l2) as [ [ | ] | ].
intros H t2 t2_in_l2;
generalize (H _ t2_in_l2)
(list_exists_option_is_sound
   (fun t1 : term =>
    match p t1 t2 with
    | Some Equivalent => Some false
    | Some Less_than => Some false
    | Some Greater_than => Some true
    | Some Uncomparable => Some false
    | None => None (A:=bool)
    end) l1);
destruct (list_exists_option
   (fun t1 : term =>
    match p t1 t2 with
    | Some Equivalent => Some false
    | Some Less_than => Some false
    | Some Greater_than => Some true
    | Some Uncomparable => Some false
    | None => None (A:=bool)
    end) l1) as [ [ | ] | ].
intros _ [t1 [t1_in_l1 H']]; exists t1; split; trivial.
destruct (p t1 t2) as [ [ | | | ] | ]; trivial; discriminate.
intros; discriminate.
intros; discriminate.
intros _;
generalize (list_forall_option_is_sound
      (fun t1 : term =>
       list_exists_option
         (fun t2 : term =>
          match p t1 t2 with
          | Some Equivalent => Some false
          | Some Less_than => Some true
          | Some Greater_than => Some false
          | Some Uncomparable => Some false
          | None => None (A:=bool)
          end) l2) l1);
destruct (list_forall_option
      (fun t1 : term =>
       list_exists_option
         (fun t2 : term =>
          match p t1 t2 with
          | Some Equivalent => Some false
          | Some Less_than => Some true
          | Some Greater_than => Some false
          | Some Uncomparable => Some false
          | None => None (A:=bool)
          end) l2) l1) as [ [ | ] | ].
intros H t1 t1_in_l1;
generalize (H _ t1_in_l1)
(list_exists_option_is_sound
   (fun t2 : term =>
    match p t1 t2 with
    | Some Equivalent => Some false
    | Some Less_than => Some true
    | Some Greater_than => Some false
    | Some Uncomparable => Some false
    | None => None (A:=bool)
    end) l2);
destruct (list_exists_option
   (fun t2 : term =>
    match p t1 t2 with
    | Some Equivalent => Some false
    | Some Less_than => Some true
    | Some Greater_than => Some false
    | Some Uncomparable => Some false
    | None => None (A:=bool)
    end) l2) as [ [ | ] | ].
intros _ [t2 [t2_in_l2 H']]; exists t2; split; trivial.
destruct (p t1 t2) as [ [ | | | ] | ]; trivial; discriminate.
intros; discriminate.
intros; discriminate.
trivial.
trivial.
trivial.
Qed.

Lemma well_formed_list_is_well_formed_list :
  forall l, well_formed_list l -> (forall t, In t l -> well_formed t).
Proof.
intro l; induction l as [ | t l].
intros; contradiction.
simpl; intros [Wt Wl] u [u_eq_t | u_in_l].
subst; trivial.
apply IHl; trivial.
Qed.

Lemma rpo_eval_is_sound_weak :
  forall n t1 t2, well_formed t1 -> well_formed t2 ->
                match rpo_eval n t1 t2 with
                     | Some Equivalent => equiv t1 t2
                     | Some Greater_than => rpo t2 t1
                     | Some Less_than => rpo t1 t2
                     | _ => True
                     end.
Proof.
intro n; induction n as [ | n].
intros t1 t2; simpl; trivial.
(* induction step *)
intros t1 t2 W1 W2; rewrite rpo_eval_equation.
assert (E1 := equiv_eval_is_sound_weak (S n) t1 t2); 
destruct (equiv_eval (S n) t1 t2) as [ [ | ] | ].
(* 1/3 t1 and t2 are equivalent *)
apply E1; trivial.
(* 1/2 t1 and t2 are not equivalent *)
clear E1; 
assert (H : forall v f l, var_in_term_list v l = true -> rpo (Var v) (Term f l)).
intros v f l H; apply trans_clos_subterm_rpo; trivial.
assert (H' : (In (Var v) l \/ 
                  (exists t, In t l /\ trans_clos direct_subterm (Var v) t)) ->
                     trans_clos direct_subterm (Var v) (Term f l)).
intros [v_in_l | [t [t_in_l H']]].
apply closure.t_step; trivial.
apply trans_clos_is_trans with t; trivial; apply closure.t_step; trivial.
apply H'; clear H'.
generalize H; clear H; pattern l; refine (list_rec3 size _ _ _); clear l.
intros m; induction m as [ | m]; intros [ | t l] L.
intros H; rewrite var_in_term_list_equation in H; discriminate H.
simpl in L; absurd (1 <= 0); auto with arith;
refine (le_trans _ _ _ _ L); apply le_trans with (size t); auto with arith;
apply size_ge_one.
intros H; rewrite var_in_term_list_equation in H; discriminate H.
simpl in L; assert (Sl : list_size size l <= m).
apply le_S_n; refine (le_trans _ _ _ _ L); 
apply (plus_le_compat_r 1 (size t) (list_size size l));
apply size_ge_one.
destruct t as [v' | f' l']; rewrite var_in_term_list_equation.
destruct (X.eq_variable_dec v v') as [v_eq_v' | v_diff_v'].
intros _; left; subst; left; trivial.
intro H; destruct (IHm _ Sl H) as [v_in_l | [t [t_in_l H']]].
left; right; trivial.
right; exists t; split; trivial; right; trivial.
assert (Sl' : list_size size l' <= m).
apply le_S_n; refine (le_trans _ _ _ _ L); rewrite size_unfold; simpl;
apply le_n_S; auto with arith.
generalize (IHm _ Sl'); destruct (var_in_term_list v l').
intro H; destruct (H (refl_equal _)) as [v_in_l' | [t [t_in_l' H']]].
right; exists (Term f' l'); split.
left; trivial.
apply closure.t_step; trivial.
right; exists (Term f' l'); split.
left; trivial.
apply trans_clos_is_trans with t; trivial; apply closure.t_step; trivial.
intros _; generalize (IHm _ Sl); destruct (var_in_term_list v l).
intro H; destruct (H (refl_equal _)) as [v_in_l' | [t [t_in_l' H']]].
left; right; trivial.
right; exists t; split; trivial; right; trivial.
intros; discriminate.
destruct t1 as [v1 | f1 l1]; destruct t2 as [v2 | f2 l2]; trivial.
generalize (H v1 f2 l2); destruct (var_in_term_list v1 l2); 
intro H'; trivial; apply H'; trivial.
generalize (H v2 f1 l1); destruct (var_in_term_list v2 l1); 
intro H'; trivial; apply H'; trivial.
(* 1/2 t1 = Term f1 l1, t2 = Term f2 l2 *)
generalize (list_exists_option_is_sound 
  (fun t : term =>
        match rpo_eval n t (Term f2 l2) with
        | Some Equivalent => Some true
        | Some Less_than => Some false
        | Some Greater_than => Some true
        | Some Uncomparable => Some false
        | None => None (A:=bool)
        end) l1);
destruct (list_exists_option
     (fun t : term =>
      match rpo_eval n t (Term f2 l2) with
      | Some Equivalent => Some true
      | Some Less_than => Some false
      | Some Greater_than => Some true
      | Some Uncomparable => Some false
      | None => None (A:=bool)
      end) l1) as [ [ | ] | ].
(* 1/4 there is a term in l1 which is greater than (Term f2 l2) *)
intros [t1 [t1_in_l1 t1_gt_f2l2]]; simpl; 
apply Subterm with t1; trivial.
assert (Wt1 : well_formed t1).
destruct (well_formed_unfold _ W1) as [W _]; apply W; trivial.
generalize (IHn t1 (Term f2 l2) Wt1 W2); 
destruct (rpo_eval n t1 (Term f2 l2)) as [ [ | | | ] | ]; try discriminate.
intro H1; apply Equiv; apply (equiv_sym _ _ equiv_equiv); trivial.
intro H1; apply Lt; trivial.
(* 1/3 there are no terms in l1 which are greater than (Term f2 l2) *)
intros _;
generalize (list_exists_option_is_sound 
  (fun t : term =>
        match rpo_eval n (Term f1 l1) t with
        | Some Equivalent => Some true
        | Some Less_than => Some true
        | Some Greater_than => Some false
        | Some Uncomparable => Some false
        | None => None (A:=bool)
        end) l2);
destruct (list_exists_option
     (fun t : term =>
      match rpo_eval n (Term f1 l1) t with
      | Some Equivalent => Some true
      | Some Less_than => Some true
      | Some Greater_than => Some false
      | Some Uncomparable => Some false
      | None => None (A:=bool)
      end) l2) as [ [ | ] | ].
(* 1/5 there is a term in l2 which is greater than (Term f1 l1) *)
intros [t2 [t2_in_l2 t2_gt_f1l1]]; simpl; apply Subterm with t2; trivial.
assert (Wt2 : well_formed t2).
destruct (well_formed_unfold _ W2) as [W _]; apply W; trivial.
generalize (IHn (Term f1 l1) t2 W1 Wt2); 
destruct (rpo_eval n (Term f1 l1) t2) as [ [ | | | ] | ]; try discriminate.
intro; apply Equiv; trivial.
intro; apply Lt; trivial.
(* 1/4 there are no terms in l2 which are greater than (Term f1 l1) *)
assert (Wl1 : well_formed_list l1).
destruct W1; trivial.
assert (Wl2 : well_formed_list l2).
destruct W2; trivial.
intros _;
generalize (prec_eval_is_sound f1 f2); destruct (prec_eval f1 f2).
(* 1/7 f1 = f2 *)
intro f1_eq_f2; assert (Sf1 := f_equal status f1_eq_f2);
destruct (status f2); subst f2; rewrite Sf1.
(* 1/8 f1 has a Lex status *)
assert (L : length l1 = length l2).
destruct (well_formed_unfold _ W1) as [_ Ll1];
destruct (well_formed_unfold _ W2) as [_ Ll2];
destruct (F.arity f1) as [ | | m]; rewrite Ll1; rewrite Ll2; trivial.
simpl; assert (H' := lexico_eval_is_sound (rpo_eval n) (Term f1 l1)
                                (Term f1 l2) l1 l2);
destruct (lexico_eval (rpo_eval n) (Term f1 l1) (Term f1 l2) l1 l2) as [ [ | | | ] | ].
destruct H' as [ll [E_ll [H1 H2]]]; trivial.
apply (Eq_lex f1 ll l1 l2 Sf1); trivial.
intros t1 t2 t1t2_in_ll; 
assert (Wt1 : well_formed t1).
apply (well_formed_list_is_well_formed_list _ (proj1 W1)).
rewrite H1; setoid_rewrite in_map_iff; exists (t1,t2); simpl; split; trivial.
assert (Wt2 : well_formed t2).
apply (well_formed_list_is_well_formed_list _ (proj1 W2)).
rewrite H2; setoid_rewrite in_map_iff; exists (t1,t2); simpl; split; trivial.
assert (H' := IHn t1 t2 Wt1 Wt2).
rewrite (E_ll t1 t2 t1t2_in_ll) in H'; trivial.
destruct (H' L) as [ll [t1 [t2 [l1' [l2' [E_ll [t1_lt_t2 [H'' [H1 H2]]]]]]]]]; 
clear H'; apply Top_eq_lex; trivial.
subst l1 l2; generalize t1 t2 l1' l2' E_ll t1_lt_t2 Wl1 Wl2 L;
clear H'' t1 t2 l1' l2' E_ll t1_lt_t2 Wl1 Wl2 W1 W2 L; 
induction ll as [ | [u1 u2]];
intros t1 t2 l1' l2' E_ll t1_lt_t2 Wl1 Wl2 L.
subst; simpl; apply List_gt.
assert (Wt1 : well_formed t1).
apply (well_formed_list_is_well_formed_list _ Wl1); simpl; left; trivial.
assert (Wt2 : well_formed t2).
apply (well_formed_list_is_well_formed_list _ Wl2); simpl; left; trivial.
assert (H'' := IHn t1 t2 Wt1 Wt2).
rewrite t1_lt_t2 in H''; trivial.
simpl in L; injection L; trivial.
simpl; apply List_eq.
assert (Wu1 : well_formed u1).
apply (well_formed_list_is_well_formed_list _ Wl1).
simpl; left; trivial.
assert (Wu2 : well_formed u2).
apply (well_formed_list_is_well_formed_list _ Wl2).
simpl; left; trivial.
assert (H'' := IHn u1 u2 Wu1 Wu2).
rewrite (E_ll u1 u2) in H''; trivial; left; trivial.
apply IHll; trivial.
intros; apply E_ll; right; trivial.
destruct Wl1; trivial.
destruct Wl2; trivial.
simpl in L; injection L; trivial.
intros u1 u1_in_l1; assert (Wu1 : well_formed u1).
apply (well_formed_list_is_well_formed_list _ Wl1); trivial.
destruct (H'' u1 u1_in_l1) as [[u2 [u2_in_l2 u1_le_u2]] | u1_lt_s2].
assert (Wu2 : well_formed u2).
apply (well_formed_list_is_well_formed_list _ Wl2); trivial.
apply Subterm with u2; trivial.
assert (H''' := IHn u1 u2 Wu1 Wu2).
destruct u1_le_u2 as [u1_eq_u2 | u1_lt_u2].
rewrite u1_eq_u2 in H'''; apply Equiv; trivial.
rewrite u1_lt_u2 in H'''; apply Lt; trivial.
assert (H''' := IHn u1 _ Wu1 W2); rewrite u1_lt_s2 in H'''; trivial.
destruct (H' L) as [ll [t1 [t2 [l1' [l2' [E_ll [t1_lt_t2 [H'' [H1 H2]]]]]]]]]; 
clear H'; apply Top_eq_lex; trivial.
subst l1 l2; generalize t1 t2 l1' l2' E_ll t1_lt_t2 Wl1 Wl2 L;
clear H'' t1 t2 l1' l2' E_ll t1_lt_t2 Wl1 Wl2 W1 W2 L; 
induction ll as [ | [u1 u2]];
intros t1 t2 l1' l2' E_ll t1_lt_t2 Wl1 Wl2 L.
subst; simpl; apply List_gt.
assert (Wt1 : well_formed t1).
apply (well_formed_list_is_well_formed_list _ Wl1); simpl; left; trivial.
assert (Wt2 : well_formed t2).
apply (well_formed_list_is_well_formed_list _ Wl2); simpl; left; trivial.
assert (H'' := IHn t1 t2 Wt1 Wt2).
rewrite t1_lt_t2 in H''; trivial.
simpl in L; injection L; intro; apply sym_eq; trivial.
simpl; apply List_eq.
apply (equiv_sym _ _ equiv_equiv);
assert (Wu1 : well_formed u1).
apply (well_formed_list_is_well_formed_list _ Wl1).
simpl; left; trivial.
assert (Wu2 : well_formed u2).
apply (well_formed_list_is_well_formed_list _ Wl2).
simpl; left; trivial.
assert (H'' := IHn u1 u2 Wu1 Wu2).
rewrite (E_ll u1 u2) in H''; trivial; left; trivial.
apply IHll; trivial.
intros; apply E_ll; right; trivial.
destruct Wl1; trivial.
destruct Wl2; trivial.
simpl in L; injection L; trivial.
intros u2 u2_in_l2; assert (Wu2 : well_formed u2).
apply (well_formed_list_is_well_formed_list _ Wl2); trivial.
destruct (H'' u2 u2_in_l2) as [[u1 [u1_in_l1 u2_le_u1]] | u2_lt_s1].
assert (Wu1 : well_formed u1).
apply (well_formed_list_is_well_formed_list _ Wl1); trivial.
apply Subterm with u1; trivial.
assert (H''' := IHn u1 u2 Wu1 Wu2).
destruct u2_le_u1 as [u2_eq_u1 | u2_lt_u1].
rewrite u2_eq_u1 in H'''; 
apply Equiv; apply (equiv_sym _ _ equiv_equiv); trivial.
rewrite u2_lt_u1 in H'''; apply Lt; trivial.
assert (H''' := IHn _ u2 W1 Wu2); rewrite u2_lt_s1 in H'''; trivial.
trivial.
trivial.
(* 1/7 f1 has a Mul status *)
simpl; assert (H' := remove_equiv_list_is_sound (equiv_eval n) l1 l2);
destruct (remove_equiv_list (equiv_eval n) l1 l2) as [ [[ | t1' l1'] [ | t2' l2']] | ].
destruct H' as [ll [E_ll [P1 [P2 _]]]]; apply (Eq_mul f1 ll l1 l2 Sf1); trivial.
intros t1 t2 t1t2_in_ll; 
assert (H' := equiv_eval_is_sound_weak n t1 t2);
rewrite (E_ll t1 t2 t1t2_in_ll) in H'; apply H'; trivial.
destruct H' as [ll [E_ll [P1 [P2 _]]]].
apply Top_eq_mul; trivial.
apply (List_mul t2' l2' nil ll); trivial.
intros t1 t2 t1t2_in_ll; 
assert (H' := equiv_eval_is_sound_weak n t1 t2);
rewrite (E_ll t1 t2 t1t2_in_ll) in H'; apply H'; trivial.
simpl; tauto.
destruct H' as [ll [E_ll [P1 [P2 _]]]].
assert (E_ll' : forall t1 t2, In (t1,t2) ll -> equiv t1 t2).
intros t1 t2 t1t2_in_ll; 
assert (H' := equiv_eval_is_sound_weak n t1 t2);
rewrite (E_ll t1 t2 t1t2_in_ll) in H'; apply H'; trivial.
destruct (equiv_swap _ equiv ll E_ll') as [ll' [E_ll'' [H1 [H2 H3]]]].
apply Top_eq_mul; trivial.
apply (List_mul t1' l1' nil ll'); trivial.
intros t1 t2; setoid_rewrite <- H3; intro; 
apply (equiv_sym _ _ equiv_equiv); apply E_ll'; trivial.
rewrite <- H2; trivial.
rewrite <- H1; trivial.
simpl; tauto.
assert (H'' := mult_eval_is_sound_weak (rpo_eval n) (t1' :: l1') (t2' :: l2')).
destruct (mult_eval (rpo_eval n) (t1' :: l1') (t2' :: l2')) as [ [ | | | ] | ].
contradiction.
destruct H' as [ll [E_ll [P1 [P2 _]]]].
assert (E_ll' : forall t1 t2, In (t1,t2) ll -> equiv t1 t2).
intros t1 t2 t1t2_in_ll; 
assert (H' := equiv_eval_is_sound_weak n t1 t2);
rewrite (E_ll t1 t2 t1t2_in_ll) in H'; apply H'; trivial.
apply Top_eq_mul; trivial.
apply (List_mul t2' l2' (t1' :: l1') ll); trivial.
intros u1 u1_in_l1'; destruct (H'' _ u1_in_l1') as [u2 [u2_in_l2' u1_lt_u2]];
exists u2; split; trivial.
assert (Wu1 : well_formed u1).
apply (well_formed_list_is_well_formed_list _ Wl1);
setoid_rewrite (in_permut_in u1 P1); apply in_or_app; left; trivial.
assert (Wu2 : well_formed u2).
apply (well_formed_list_is_well_formed_list _ Wl2); 
setoid_rewrite (in_permut_in u2 P2); apply in_or_app; left; trivial.
assert (H''' := IHn u1 u2 Wu1 Wu2).
rewrite u1_lt_u2 in H'''; trivial.
destruct H' as [ll [E_ll [P1 [P2 _]]]].
assert (E_ll' : forall t1 t2, In (t1,t2) ll -> equiv t1 t2).
intros t1 t2 t1t2_in_ll; 
assert (H' := equiv_eval_is_sound_weak n t1 t2);
rewrite (E_ll t1 t2 t1t2_in_ll) in H'; apply H'; trivial.
destruct (equiv_swap _ equiv ll E_ll') as [ll' [E_ll'' [H1 [H2 H3]]]].
apply Top_eq_mul; trivial.
apply (List_mul t1' l1' (t2' :: l2') ll'); trivial.
intros t1 t2; setoid_rewrite <- H3; intro; 
apply (equiv_sym _ _ equiv_equiv); apply E_ll'; trivial.
rewrite <- H2; trivial.
rewrite <- H1; trivial.
intros u2 u2_in_l2'; destruct (H'' _ u2_in_l2') as [u1 [u1_in_l1' u2_lt_u1]];
exists u1; split; trivial.
assert (Wu1 : well_formed u1).
apply (well_formed_list_is_well_formed_list _ Wl1);
setoid_rewrite (in_permut_in u1 P1); apply in_or_app; left; trivial.
assert (Wu2 : well_formed u2).
apply (well_formed_list_is_well_formed_list _ Wl2); 
setoid_rewrite (in_permut_in u2 P2); apply in_or_app; left; trivial.
assert (H''' := IHn u1 u2 Wu1 Wu2).
rewrite u2_lt_u1 in H'''; trivial.
trivial.
trivial.
trivial.
(* 1/6 f1 < f2 *)
intro f1_lt_f2; simpl.
generalize (list_forall_option_is_sound
      (fun t : term =>
       match rpo_eval n t (Term f2 l2) with
       | Some Equivalent => Some false
       | Some Less_than => Some true
       | Some Greater_than => Some false
       | Some Uncomparable => Some false
       | None => None (A:=bool)
       end) l1);
destruct (list_forall_option
      (fun t : term =>
       match rpo_eval n t (Term f2 l2) with
       | Some Equivalent => Some false
       | Some Less_than => Some true
       | Some Greater_than => Some false
       | Some Uncomparable => Some false
       | None => None (A:=bool)
       end) l1) as [ [ | ] | ].
intros l1_lt_s2; apply Top_gt; trivial.
intros t1 t1_in_l1; 
assert (Wt1 : well_formed t1).
apply (well_formed_list_is_well_formed_list _ Wl1); trivial.
assert (H' := l1_lt_s2 t1 t1_in_l1);
assert (H'' := IHn _ _ Wt1 W2);
destruct (rpo_eval n t1 (Term f2 l2)) as [ [ | | | ] | ]; trivial; discriminate.
trivial.
trivial.
(* 1/5 f2 < f1 *)
intro f2_lt_f1; simpl.
generalize (list_forall_option_is_sound
      (fun t : term =>
       match rpo_eval n (Term f1 l1) t with
       | Some Equivalent => Some false
       | Some Less_than => Some false
       | Some Greater_than => Some true
       | Some Uncomparable => Some false
       | None => None (A:=bool)
       end) l2); 
destruct (list_forall_option
      (fun t : term =>
       match rpo_eval n (Term f1 l1) t with
       | Some Equivalent => Some false
       | Some Less_than => Some false
       | Some Greater_than => Some true
       | Some Uncomparable => Some false
       | None => None (A:=bool)
       end) l2) as [ [ | ] | ].
intros s1_gt_l2; apply Top_gt; trivial.
intros t2 t2_in_l2; 
assert (Wt2 : well_formed t2).
apply (well_formed_list_is_well_formed_list _ Wl2); trivial.
assert (H' := s1_gt_l2 t2 t2_in_l2);
assert (H'' := IHn _ _ W1 Wt2);
destruct (rpo_eval n (Term f1 l1) t2) as [ [ | | | ] | ]; trivial; discriminate.
trivial.
trivial.
simpl; trivial.
simpl; trivial.
simpl; trivial.
trivial.
Qed.

Lemma lexico_eval_fully_evaluates :
  forall p s1 s2 l1 l2, 
  (forall t1, In t1 l1 -> p t1 s2 <> None) ->
  (forall t2, In t2 l2 -> p s1 t2 <> None) ->
  (forall t1 t2, In t1 l1 -> In t2 l2 -> p t1 t2 <> None) ->
  lexico_eval p s1 s2 l1 l2 <> None.
Proof.
intros p s1 s2 l1; induction l1 as [ | t1 l1]; intros [ | t2 l2] Es2 Es1 E;
simpl; try discriminate.
assert (H := E t1 t2 (or_introl _ (refl_equal _)) (or_introl _ (refl_equal _)));
destruct (p t1 t2) as [ [ | | | ] | ].
(* 1/5 p t1 t2 = Some Equivalent *)
apply IHl1; intros; [apply Es2 | apply Es1 | apply E]; right; trivial.
(* 1/4 p t1 t2 = Some Less_than *)
generalize (list_forall_option_is_sound
      (fun t3 : term =>
       match p t3 s2 with
       | Some Equivalent => Some false
       | Some Less_than => Some true
       | Some Greater_than => Some false
       | Some Uncomparable => Some false
       | None => None (A:=bool)
       end) l1);
destruct (list_forall_option
      (fun t3 : term =>
       match p t3 s2 with
       | Some Equivalent => Some false
       | Some Less_than => Some true
       | Some Greater_than => Some false
       | Some Uncomparable => Some false
       | None => None (A:=bool)
       end) l1) as [ [ | ] | ].
(* 1/6 all terms in l1 are smaller than s2 *)
clear H; assert (H := Es2 t1 (or_introl _ (refl_equal _)));
destruct (p t1 s2) as [ [ | | | ] | ].
intros _; discriminate.
intros _; discriminate.
intros _; discriminate.
intros _; discriminate.
intros _; apply H; trivial.
(* 1/5 NOT all terms in l1 are smaller than s2 *)
clear H; assert (H := Es2 t1 (or_introl _ (refl_equal _)));
destruct (p t1 s2) as [ [ | | | ] | ].
intros _; discriminate.
intros _; discriminate.
intros _; discriminate.
intros _; discriminate.
intros _; apply H; trivial.
(* 1/4 at least one evaluation does not yield a result *) 
intros [u1 [u1_in_l1 pt1s2_eq_none]];
assert (H' := Es2 _ (or_intror _ u1_in_l1)); 
destruct (p u1 s2) as [ [ | | | ] | ]; trivial.
discriminate.
discriminate.
discriminate.
discriminate.
absurd (@None comp = None); trivial.
(* 1/3 p t1 t2 = Some Greater_than *)
generalize (list_forall_option_is_sound
        (fun t3 : term =>
         match p s1 t3 with
         | Some Equivalent => Some false
         | Some Less_than => Some false
         | Some Greater_than => Some true
         | Some Uncomparable => Some false
         | None => None (A:=bool)
         end) l2);
destruct (list_forall_option
        (fun t3 : term =>
         match p s1 t3 with
         | Some Equivalent => Some false
         | Some Less_than => Some false
         | Some Greater_than => Some true
         | Some Uncomparable => Some false
         | None => None (A:=bool)
         end) l2) as [ [ | ] | ].
(* 1/5 all terms in l2 are smaller than s1 *)
clear H; assert (H := Es1 t2 (or_introl _ (refl_equal _)));
destruct (p s1 t2) as [ [ | | | ] | ].
intros _; discriminate.
intros _; discriminate.
intros _; discriminate.
intros _; discriminate.
intros _; apply H; trivial.
(* 1/4 NOT all terms in l2 are smaller than s1 *)
clear H; assert (H := Es1 t2 (or_introl _ (refl_equal _)));
destruct (p s1 t2) as [ [ | | | ] | ].
intros _; discriminate.
intros _; discriminate.
intros _; discriminate.
intros _; discriminate.
intros _; apply H; trivial.
(* 1/3 at least one evaluation does not yield a result *) 
intros [u2 [u2_in_l2 ps1u2_eq_none]];
assert (H' := Es1 _ (or_intror _ u2_in_l2)); 
destruct (p s1 u2) as [ [ | | | ] | ]; trivial.
discriminate.
discriminate.
discriminate.
discriminate.
absurd (@None comp = None); trivial.
(* 1/2 p t1 t2 = Some Uncomparable *)
discriminate.
(* 1/1 p t1 t2 = None *)
trivial.
Qed.

Lemma mult_eval_fully_evaluates :
  forall p l1 l2, 
  (forall t1 t2, In t1 l1 -> In t2 l2 -> p t1 t2 <> None) ->
  mult_eval p l1 l2 <> None.
Proof.
intros p [ | t1 l1] l2 E; unfold mult_eval.
simpl; clear E; induction l2 as [ | t2 l2]; simpl.
discriminate.
destruct (list_forall_option (fun _ : term => Some false) l2) as [ [ | ] | ];
trivial; discriminate.
generalize (list_forall_option_is_sound
    (fun t2 : term =>
     list_exists_option
       (fun t3 : term =>
        match p t3 t2 with
        | Some Equivalent => Some false
        | Some Less_than => Some false
        | Some Greater_than => Some true
        | Some Uncomparable => Some false
        | None => None (A:=bool)
        end) (t1 :: l1)) l2);
destruct (list_forall_option
    (fun t2 : term =>
     list_exists_option
       (fun t3 : term =>
        match p t3 t2 with
        | Some Equivalent => Some false
        | Some Less_than => Some false
        | Some Greater_than => Some true
        | Some Uncomparable => Some false
        | None => None (A:=bool)
        end) (t1 :: l1)) l2) as [ [ | ] | ].
intros _; discriminate.
intros _;
generalize (list_forall_option_is_sound
    (fun t2 : term =>
     list_exists_option
       (fun t3 : term =>
        match p t2 t3 with
        | Some Equivalent => Some false
        | Some Less_than => Some true
        | Some Greater_than => Some false
        | Some Uncomparable => Some false
        | None => None (A:=bool)
        end) l2) (t1 :: l1));
destruct (list_forall_option
    (fun t2 : term =>
     list_exists_option
       (fun t3 : term =>
        match p t2 t3 with
        | Some Equivalent => Some false
        | Some Less_than => Some true
        | Some Greater_than => Some false
        | Some Uncomparable => Some false
        | None => None (A:=bool)
        end) l2) (t1 :: l1)) as [ [ | ] | ].
intros _; discriminate.
intros _; discriminate.
intros [u1 [u1_in_l1 H]].
assert (H' := list_exists_option_is_sound
      (fun t3 : term =>
       match p u1 t3 with
       | Some Equivalent => Some false
       | Some Less_than => Some true
       | Some Greater_than => Some false
       | Some Uncomparable => Some false
       | None => None (A:=bool)
       end) l2); rewrite H in H'. 
destruct H' as [ [u2 [u2_in_l2 pu1u2_eq_none]] _].
assert (H' := E u1 u2 u1_in_l1 u2_in_l2);
destruct (p u1 u2) as [ [ | | | ] | ]; trivial; discriminate.
intros [u2 [u2_in_l2 H]].
assert (H' := list_exists_option_is_sound
      (fun t3 : term =>
       match p t3 u2 with
       | Some Equivalent => Some false
       | Some Less_than => Some false
       | Some Greater_than => Some true
       | Some Uncomparable => Some false
       | None => None (A:=bool)
       end) (t1 :: l1)); rewrite H in H'. 
destruct H' as [ [u1 [u1_in_l1 pu1u2_eq_none]] _].
assert (H' := E u1 u2 u1_in_l1 u2_in_l2);
destruct (p u1 u2) as [ [ | | | ] | ]; trivial; discriminate.
Qed.

Lemma rpo_eval_terminates :
  forall n t1 t2, size t1 + size t2 <= n -> rpo_eval n t1 t2 <> None.
Proof.
intro n; induction n as [ | n].
(* Base case *)
intros t1 t2 St1; 
absurd (1 <= 0); auto with arith; 
apply le_trans with (size t1 + size t2); trivial;
apply le_trans with (1 + size t2);
[apply le_plus_l | apply plus_le_compat_r; apply size_ge_one].
(* Induction step *)
intros t1 t2 St; rewrite rpo_eval_equation.
assert (T := equiv_eval_terminates (S n) t1 t2 St); 
destruct (equiv_eval (S n) t1 t2) as [ [ | ] | ]; try discriminate.
destruct t1 as [v1 | f1 l1]; destruct t2 as [v2 | f2 l2].
discriminate.
destruct (var_in_term_list v1 l2); discriminate.
destruct (var_in_term_list v2 l1); discriminate.
generalize (list_exists_option_is_sound (fun t : term =>
      match rpo_eval n t (Term f2 l2) with
      | Some Equivalent => Some true
      | Some Less_than => Some false
      | Some Greater_than => Some true
      | Some Uncomparable => Some false
      | None => None (A:=bool)
      end) l1);
destruct (list_exists_option
    (fun t : term =>
     match rpo_eval n t (Term f2 l2) with
     | Some Equivalent => Some true
     | Some Less_than => Some false
     | Some Greater_than => Some true
     | Some Uncomparable => Some false
     | None => None (A:=bool)
     end) l1) as [ [ | ] | ].
intros _; simpl; discriminate.
intros _;
generalize (list_exists_option_is_sound (fun t : term =>
         match rpo_eval n (Term f1 l1) t with
         | Some Equivalent => Some true
         | Some Less_than => Some true
         | Some Greater_than => Some false
         | Some Uncomparable => Some false
         | None => None (A:=bool)
         end) l2);
destruct (list_exists_option
        (fun t : term =>
         match rpo_eval n (Term f1 l1) t with
         | Some Equivalent => Some true
         | Some Less_than => Some true
         | Some Greater_than => Some false
         | Some Uncomparable => Some false
         | None => None (A:=bool)
         end) l2) as [ [ | ] | ].
intros _; simpl; discriminate.
intros _; simpl; destruct (prec_eval f1 f2); try discriminate.
destruct (status f1).
apply lexico_eval_fully_evaluates.
intros t1 t1_in_l1; apply IHn.
apply le_S_n; refine (le_trans _ _ _ _ St);
replace (S (size t1 + size (Term f2 l2))) with 
                (S (size t1) + size (Term f2 l2)); trivial;
apply plus_le_compat_r; apply size_direct_subterm; trivial.
intros t2 t2_in_l2; apply IHn.
apply le_S_n; refine (le_trans _ _ _ _ St); rewrite plus_comm;
replace (S (size t2 + size (Term f1 l1))) with 
                (S (size t2) + size (Term f1 l1)); trivial; rewrite plus_comm;
apply plus_le_compat_l; apply size_direct_subterm; trivial.
intros t1 t2 t1_in_l1 t2_in_l2; apply IHn.
apply le_S_n; refine (le_trans _ _ _ _ St);
replace (S (size t1 + size t2)) with  (S (size t1) + size t2); trivial;
apply plus_le_compat.
apply size_direct_subterm; trivial.
apply lt_le_weak; apply size_direct_subterm; trivial.
generalize (remove_equiv_list_is_sound (equiv_eval n) l1 l2);
destruct (remove_equiv_list (equiv_eval n) l1 l2) as [ [[ | t1' l1'] [ | t2' l2']] | ].
intros; discriminate.
intros; discriminate.
intros; discriminate.
intros [ll [_ [P1 [P2 _]]]]; 
apply mult_eval_fully_evaluates.
intros t1 t2 t1_in_l1' t2_in_l2'; apply IHn.
apply le_S_n; refine (le_trans _ _ _ _ St);
replace (S (size t1 + size t2)) with  (S (size t1) + size t2); trivial;
apply plus_le_compat.
apply size_direct_subterm;
setoid_rewrite (in_permut_in t1 P1); apply in_or_app; left; trivial.
apply lt_le_weak; apply size_direct_subterm;
setoid_rewrite (in_permut_in t2 P2); apply in_or_app; left; trivial.
intros [t1 [t2 [t1_in_l1 [t2_in_l2 H]]]];
assert (H' := equiv_eval_terminates n t1 t2);
destruct (equiv_eval n t1 t2) as [ [ | ] | ].
discriminate.
discriminate.
intros _; apply H'; trivial.
apply le_S_n; refine (le_trans _ _ _ _ St);
replace (S (size t1 + size t2)) with  (S (size t1) + size t2); trivial;
apply plus_le_compat.
apply size_direct_subterm; trivial.
apply lt_le_weak; apply size_direct_subterm; trivial.
generalize (list_forall_option_is_sound
    (fun t : term =>
     match rpo_eval n t (Term f2 l2) with
     | Some Equivalent => Some false
     | Some Less_than => Some true
     | Some Greater_than => Some false
     | Some Uncomparable => Some false
     | None => None (A:=bool)
     end) l1);
destruct (list_forall_option
    (fun t : term =>
     match rpo_eval n t (Term f2 l2) with
     | Some Equivalent => Some false
     | Some Less_than => Some true
     | Some Greater_than => Some false
     | Some Uncomparable => Some false
     | None => None (A:=bool)
     end) l1) as [ [ | ] | ].
intros _; discriminate.
intros _; discriminate.
intros [u1 [u1_in_l1 H]]; assert (H' := IHn u1 (Term f2 l2)); 
destruct (rpo_eval n u1 (Term f2 l2)) as [ [ | | | ] | ].
discriminate.
discriminate.
discriminate.
discriminate.
apply H'.
apply le_S_n; refine (le_trans _ _ _ _ St);
replace (S (size u1 + size (Term f2 l2))) with 
                (S (size u1) + size (Term f2 l2)); trivial;
apply plus_le_compat_r; apply size_direct_subterm; trivial.
generalize  (list_forall_option_is_sound
    (fun t : term =>
     match rpo_eval n (Term f1 l1) t with
     | Some Equivalent => Some false
     | Some Less_than => Some false
     | Some Greater_than => Some true
     | Some Uncomparable => Some false
     | None => None (A:=bool)
     end) l2);
destruct (list_forall_option
    (fun t : term =>
     match rpo_eval n (Term f1 l1) t with
     | Some Equivalent => Some false
     | Some Less_than => Some false
     | Some Greater_than => Some true
     | Some Uncomparable => Some false
     | None => None (A:=bool)
     end) l2) as [ [ | ] | ].
intros _; discriminate.
intros _; discriminate.
intros [u2 [u2_in_l2 H]]; assert (H' := IHn (Term f1 l1) u2); 
destruct (rpo_eval n (Term f1 l1) u2) as [ [ | | | ] | ].
discriminate.
discriminate.
discriminate.
discriminate.
apply H'.
apply le_S_n; refine (le_trans _ _ _ _ St); rewrite plus_comm;
replace (S (size u2 + size (Term f1 l1))) with 
                (S (size u2) + size (Term f1 l1)); trivial; rewrite plus_comm;
apply plus_le_compat_l; apply size_direct_subterm; trivial.
intros [[u2 [u2_in_l2 H]] _]; assert (H' := IHn (Term f1 l1) u2); 
destruct (rpo_eval n (Term f1 l1) u2) as [ [ | | | ] | ].
discriminate.
discriminate.
discriminate.
discriminate.
apply H'.
apply le_S_n; refine (le_trans _ _ _ _ St); rewrite plus_comm;
replace (S (size u2 + size (Term f1 l1))) with 
                (S (size u2) + size (Term f1 l1)); trivial; rewrite plus_comm;
apply plus_le_compat_l; apply size_direct_subterm; trivial.
intros [[u1 [u1_in_l1 H]] _]; assert (H' := IHn u1 (Term f2 l2)); 
destruct (rpo_eval n u1 (Term f2 l2)) as [ [ | | | ] | ].
discriminate.
discriminate.
discriminate.
discriminate.
apply H'.
apply le_S_n; refine (le_trans _ _ _ _ St);
replace (S (size u1 + size (Term f2 l2))) with 
                (S (size u1) + size (Term f2 l2)); trivial;
apply plus_le_compat_r; apply size_direct_subterm; trivial.
intros _; apply T; trivial.
Qed.

Lemma rpo_eval_is_complete_equivalent :
  forall n t1 t2, size t1 + size t2 <= n -> equiv t1 t2 ->
        rpo_eval n t1 t2 = Some Equivalent.
Proof.
intros n t1 t2 St t1_eq_t2; assert (H := equiv_eval_is_sound _ _ _ St).
rewrite rpo_eval_equation.
destruct (equiv_eval n t1 t2) as [ [ | ] | ].
trivial.
absurd (equiv t1 t2); trivial.
contradiction.
Qed.

Lemma var_case2 : 
  forall v f l, rpo (Var v) (Term f l) -> var_in_term_list v l = true .
Proof.
intros v f l v_lt_t; 
inversion v_lt_t as [ f2 l2 u s s_in_l v_le_s | | | ]; subst.
generalize s s_in_l v_le_s; clear v_lt_t s s_in_l v_le_s; 
pattern l; apply (list_rec3 size); clear l.
intro n; induction n as [ | n]; intros [ | t l] Sl s s_in_l v_le_s.
contradiction.
simpl in Sl; absurd (1 <= 0); auto with arith.
refine (le_trans _ _ _ _ Sl); apply le_trans with (size t).
apply size_ge_one.
apply le_plus_l.
contradiction.
simpl in s_in_l; destruct s_in_l as [s_eq_t | s_in_l].
subst s; inversion v_le_s; subst.
inversion H; subst;
rewrite var_in_term_list_equation; simpl;
destruct (X.eq_variable_dec v v) as [_ | v_diff_v]; trivial;
absurd (v = v); trivial.
inversion H as [g k u1 s' s'_in_k v_le_s' | | |]; subst;
rewrite var_in_term_list_equation; simpl.
assert (Sk : list_size size k <= n).
apply le_S_n; refine (le_trans _ _ _ _ Sl); simpl; apply le_n_S;
rewrite (list_size_fold size k); auto with arith.
rewrite (IHn k Sk s' s'_in_k v_le_s'); simpl; trivial.
assert (Sl' : list_size size l <= n).
apply le_S_n; refine (le_trans _ _ _ _ Sl); simpl;
apply (plus_le_compat_r 1 (size t) (list_size size l));
apply size_ge_one.
rewrite var_in_term_list_equation; 
rewrite (IHn l Sl' s s_in_l v_le_s); destruct t as [v' | g k].
destruct (X.eq_variable_dec v v'); trivial.
destruct (var_in_term_list v k); trivial.
Qed.

Lemma rpo_eval_is_complete_less_greater :
  forall n t1 t2, size t1 + size t2 <= n -> 
        well_formed t1 -> well_formed t2 -> rpo t1 t2 ->
        rpo_eval n t1 t2 = Some Less_than /\
        rpo_eval n t2 t1 = Some Greater_than.
Proof.
intro n; induction n as [ | n]; intros t1 t2 St W1 W2 t1_lt_t2. 
absurd (1 <= 0); auto with arith; 
apply le_trans with (size t1 + size t2); trivial;
apply le_trans with (1 + size t2);
[apply le_plus_l | apply plus_le_compat_r; apply size_ge_one].
assert (TE := equiv_eval_terminates _ _ _ St);
assert (R := rpo_eval_is_sound_weak (S n) t1 t2 W1 W2);
assert (T := rpo_eval_terminates (S n) t1 t2 St).
rewrite plus_comm in St;
assert (TE' := equiv_eval_terminates _ _ _ St);
assert (R' := rpo_eval_is_sound_weak (S n) t2 t1 W2 W1);
assert (T' := rpo_eval_terminates (S n) t2 t1 St).
rewrite rpo_eval_equation;
rewrite rpo_eval_equation in R; 
rewrite rpo_eval_equation in T;
destruct (equiv_eval (S n) t1 t2) as [ [ | ] | ].
absurd (rpo t1 t1).
intro; apply (rpo_antirefl t1); trivial.
setoid_rewrite <- (equiv_rpo_equiv_1 _ _ R) in t1_lt_t2; trivial.
rewrite rpo_eval_equation;
rewrite rpo_eval_equation in R'; 
rewrite rpo_eval_equation in T';
destruct (equiv_eval (S n) t2 t1) as [ [ | ] | ].
absurd (rpo t1 t1).
intro; apply (rpo_antirefl t1); trivial.
setoid_rewrite (equiv_rpo_equiv_1 _ _ R') in t1_lt_t2; trivial.
clear TE TE';
destruct t1 as [v1 | f1 l1]; destruct t2 as [v2 | f2 l2].
inversion t1_lt_t2.
rewrite (var_case2 _ _ _ t1_lt_t2); split; trivial.
inversion t1_lt_t2.
assert (Sl : forall t1 t2, In t1 l1 -> In t2 l2 -> size t1 + size t2 <= n).
intros t1 t2 t1_in_l1 t2_in_l2; apply le_S_n; refine (le_trans _ _ _ _ St);
replace (S (size t1 + size t2)) with (S (size t1) + size t2); trivial;
rewrite plus_comm; apply plus_le_compat.
apply lt_le_weak; apply size_direct_subterm; trivial.
apply size_direct_subterm; trivial.
destruct (well_formed_unfold (Term f1 l1) W1) as [Wl1 _].
destruct (well_formed_unfold (Term f2 l2) W2) as [Wl2 _].
assert (IHl1l2 : forall t1 t2, In t1 l1 -> In t2 l2 -> rpo t1 t2 ->
     rpo_eval n t1 t2 = Some Less_than /\
      rpo_eval n t2 t1 = Some Greater_than).
intros; apply IHn; trivial.
apply Sl; trivial.
apply Wl1; trivial.
apply Wl2; trivial.
assert (IHl2l1 : forall t1 t2, In t1 l1 -> In t2 l2 -> rpo t2 t1 ->
     rpo_eval n t2 t1 = Some Less_than /\
      rpo_eval n t1 t2 = Some Greater_than).
intros; apply IHn; trivial.
rewrite plus_comm; apply Sl; trivial.
apply Wl2; trivial.
apply Wl1; trivial.
assert (IHl1s2 : forall t1, In t1 l1 -> rpo t1 (Term f2 l2) ->
     rpo_eval n t1 (Term f2 l2) = Some Less_than /\
      rpo_eval n (Term f2 l2) t1 = Some Greater_than).
intros; apply IHn; trivial.
apply le_S_n; refine (le_trans _ _ _ _ St);
replace (S (size t1 + size (Term f2 l2))) with (S (size t1) + size (Term f2 l2)); trivial;
rewrite plus_comm; apply plus_le_compat_l.
apply size_direct_subterm; trivial.
apply Wl1; trivial.
assert (IHs2l1 : forall t1, In t1 l1 -> rpo (Term f2 l2) t1 ->
     rpo_eval n (Term f2 l2) t1 = Some Less_than /\
      rpo_eval n t1 (Term f2 l2) = Some Greater_than).
intros; apply IHn; trivial.
rewrite plus_comm; apply le_S_n; refine (le_trans _ _ _ _ St);
replace (S (size t1 + size (Term f2 l2))) with (S (size t1) + size (Term f2 l2)); trivial;
rewrite plus_comm; apply plus_le_compat_l.
apply size_direct_subterm; trivial.
apply Wl1; trivial.
assert (IHs1l2 : forall t2, In t2 l2 -> rpo (Term f1 l1) t2 ->
     rpo_eval n (Term f1 l1) t2 = Some Less_than /\
      rpo_eval n t2 (Term f1 l1) = Some Greater_than).
intros; apply IHn; trivial.
rewrite plus_comm;
apply le_S_n; refine (le_trans _ _ _ _ St);
replace (S (size t2 + size (Term f1 l1))) with (S (size t2) + size (Term f1 l1)); trivial;
apply plus_le_compat_r.
apply size_direct_subterm; trivial.
apply Wl2; trivial.
assert (IHl2s1 : forall t2, In t2 l2 -> rpo t2 (Term f1 l1) ->
     rpo_eval n t2 (Term f1 l1) = Some Less_than /\
      rpo_eval n (Term f1 l1) t2 = Some Greater_than).
intros; apply IHn; trivial.
apply le_S_n; refine (le_trans _ _ _ _ St);
replace (S (size t2 + size (Term f1 l1))) with (S (size t2) + size (Term f1 l1)); trivial;
apply plus_le_compat_r.
apply size_direct_subterm; trivial.
apply Wl2; trivial.
inversion t1_lt_t2 as
[ f2' l2' u1 s2 s2_in_l2 u1_le_s2
| f2' f1' l2' l1' f1_lt_f2 H
| f l2' l1' Sf l1_lt_l2 H
| f l2' l1' Sf l1_lt_l2]; subst.
(* 1/6 Subterm *)
split.
destruct (list_exists_option
           (fun t : term =>
            match rpo_eval n t (Term f2 l2) with
            | Some Equivalent => Some true
            | Some Less_than => Some false
            | Some Greater_than => Some true
            | Some Uncomparable => Some false
            | None => None (A:=bool)
            end) l1) as [ [ | ] | ].
simpl in R; 
destruct (rpo_closure (Term f1 l1) (Term f2 l2) (Term f2 l2)) as [_ [Antisym _]].
assert False; [apply Antisym; trivial | contradiction].
assert (t1_lt_l2 : list_exists_option
          (fun t : term =>
           match rpo_eval n (Term f1 l1) t with
           | Some Equivalent => Some true
           | Some Less_than => Some true
           | Some Greater_than => Some false
           | Some Uncomparable => Some false
           | None => None (A:=bool)
           end) l2 = Some true).
apply list_exists_option_is_complete_true.
exists s2; split; trivial.
inversion u1_le_s2; subst.
rewrite (rpo_eval_is_complete_equivalent n (Term f1 l1) s2); trivial.
rewrite plus_comm;
apply le_S_n; refine (le_trans _ _ _ _ St);
replace (S (size s2 + size (Term f1 l1))) with (S (size s2) + size (Term f1 l1)); trivial;
apply plus_le_compat_r.
apply size_direct_subterm; trivial.
destruct (IHs1l2 s2 s2_in_l2 H) as [H1 _]; rewrite H1; trivial.
rewrite t1_lt_l2; simpl; trivial.
simpl in T; absurd (@None comp = None); trivial.
assert (l2_gt_t1 : list_exists_option
            (fun t : term =>
             match rpo_eval n t (Term f1 l1) with
             | Some Equivalent => Some true
             | Some Less_than => Some false
             | Some Greater_than => Some true
             | Some Uncomparable => Some false
             | None => None (A:=bool)
             end) l2 = Some true).
apply list_exists_option_is_complete_true.
exists s2; split; trivial.
inversion u1_le_s2; subst.
rewrite (rpo_eval_is_complete_equivalent n s2 (Term f1 l1)); trivial.
apply le_S_n; refine (le_trans _ _ _ _ St);
replace (S (size s2 + size (Term f1 l1))) with (S (size s2) + size (Term f1 l1)); trivial;
apply plus_le_compat_r.
apply size_direct_subterm; trivial.
apply (equiv_sym _ _ equiv_equiv); trivial.
destruct (IHs1l2 s2 s2_in_l2 H) as [_ H2]; rewrite H2; trivial.
rewrite l2_gt_t1; simpl; trivial.
(* 1/5 Top_gt *)
split.
destruct (list_exists_option
           (fun t : term =>
            match rpo_eval n t (Term f2 l2) with
            | Some Equivalent => Some true
            | Some Less_than => Some false
            | Some Greater_than => Some true
            | Some Uncomparable => Some false
            | None => None (A:=bool)
            end) l1) as [ [ | ] | ].
simpl in R.
destruct (rpo_closure (Term f1 l1) (Term f2 l2) (Term f2 l2)) as [_ [Antisym _]].
assert False; [apply Antisym; trivial | contradiction].
destruct (list_exists_option
               (fun t : term =>
                match rpo_eval n (Term f1 l1) t with
                | Some Equivalent => Some true
                | Some Less_than => Some true
                | Some Greater_than => Some false
                | Some Uncomparable => Some false
                | None => None (A:=bool)
                end) l2) as [ [ | ] | ].
trivial.
generalize (prec_eval_is_sound f1 f2); destruct (prec_eval f1 f2).
intro; subst f2; assert False; [apply (prec_antisym f1); trivial | contradiction].
intros _; assert (H' : list_forall_option
    (fun t : term =>
     match rpo_eval n t (Term f2 l2) with
     | Some Equivalent => Some false
     | Some Less_than => Some true
     | Some Greater_than => Some false
     | Some Uncomparable => Some false
     | None => None (A:=bool)
     end) l1 = Some true).
apply list_forall_option_is_complete_true.
intros u1 u1_in_l1; destruct (IHl1s2 u1 u1_in_l1) as [H1 _].
apply H0; trivial.
rewrite H1; trivial.
rewrite H'; trivial.
intro f2_lt_f1; assert False;
[apply (prec_antisym f1); apply prec_transitive with f2; trivial
| contradiction].
intros [_ [not_f1_lt_f2 _]]; absurd (prec f1 f2); trivial.
simpl in T; absurd (@None comp = None); trivial.
simpl in T; absurd (@None comp = None); trivial.
clear R T;
destruct (list_exists_option
             (fun t : term =>
             match rpo_eval n t (Term f1 l1) with
             | Some Equivalent => Some true
             | Some Less_than => Some false
             | Some Greater_than => Some true
             | Some Uncomparable => Some false
             | None => None (A:=bool)
             end) l2) as [ [ | ] | ].
trivial.
destruct (list_exists_option
         (fun t : term =>
          match rpo_eval n (Term f2 l2) t with
          | Some Equivalent => Some true
          | Some Less_than => Some true
          | Some Greater_than => Some false
          | Some Uncomparable => Some false
          | None => None (A:=bool)
          end) l1) as [ [ | ] | ].
simpl in R'.
destruct (rpo_closure (Term f1 l1) (Term f2 l2) (Term f2 l2)) as [_ [Antisym _]].
assert False; [apply Antisym; trivial | contradiction].
generalize (prec_eval_is_sound f2 f1); destruct (prec_eval f2 f1).
intro; subst f2; assert False; [apply (prec_antisym f1); trivial | contradiction].
intro f2_lt_f1; assert False;
[apply (prec_antisym f1); apply prec_transitive with f2; trivial
| contradiction].
intros _; assert (H' : list_forall_option
    (fun t : term =>
     match rpo_eval n (Term f2 l2) t with
     | Some Equivalent => Some false
     | Some Less_than => Some false
     | Some Greater_than => Some true
     | Some Uncomparable => Some false
     | None => None (A:=bool)
     end) l1 = Some true).
apply list_forall_option_is_complete_true.
intros u1 u1_in_l1; destruct (IHl1s2 u1 u1_in_l1) as [_ H2].
apply H0; trivial.
rewrite H2; trivial.
rewrite H'; trivial.
intros [_ [_ not_f1_lt_f2]]; absurd (prec f1 f2); trivial.
simpl in T'; absurd (@None comp = None); trivial.
simpl in T'; absurd (@None comp = None); trivial.
(* 1/4 Top_eq_lex *)
split.
destruct (list_exists_option
     (fun t : term =>
      match rpo_eval n t (Term f2 l2) with
      | Some Equivalent => Some true
      | Some Less_than => Some false
      | Some Greater_than => Some true
      | Some Uncomparable => Some false
      | None => None (A:=bool)
      end) l1) as [ [ | ] | ].
simpl in R.
destruct (rpo_closure (Term f2 l1) (Term f2 l2) (Term f2 l2)) as [_ [Antisym _]].
assert False; [apply Antisym; trivial | contradiction].
destruct (list_exists_option
         (fun t : term =>
          match rpo_eval n (Term f2 l1) t with
          | Some Equivalent => Some true
          | Some Less_than => Some true
          | Some Greater_than => Some false
          | Some Uncomparable => Some false
          | None => None (A:=bool)
          end) l2) as [ [ | ] | ].
trivial.
simpl; generalize (prec_eval_is_sound f2 f2); destruct (prec_eval f2 f2).
intros _; rewrite Sf; rewrite Sf in R; rewrite Sf in T; simpl in R; simpl in T.
assert (L : length l1 = length l2).
destruct W1 as [_ L1]; destruct W2 as [_ L2];
destruct (F.arity f2) as [ | | m]; rewrite L1; rewrite L2; trivial.
generalize (Term f2 l1) (Term f2 l2) t1_lt_t2 H0 St 
IHl1s2 IHs2l1 IHs1l2 IHl2s1 R T.
clear W1 W2 H0 St IHl1s2 IHs2l1 IHs1l2 IHl2s1 R T R' T'.
generalize l2 L l1_lt_l2 IHl1l2 IHl2l1 Sl; 
clear l2 L l1_lt_l2 IHl1l2 IHl2l1 Sl Wl1 Wl2 t1_lt_t2.
induction l1 as [ | t1 l1]; intros [ | t2 l2] L l1_lt_l2 IHl1l2 IHl2l1 Sl;
intros s1 s2 s1_lt_s2 H0 St IHl1s2 IHs2l1 IHs1l2 IHl2s1 R T.
simpl in R; absurd (rpo s1 s1).
intro; apply (rpo_antirefl s1); trivial.
setoid_rewrite <- (equiv_rpo_equiv_1 _ _ R) in s1_lt_s2; trivial.
discriminate.
discriminate.
inversion l1_lt_l2; subst.
simpl; destruct (IHl1l2 t1 t2 (or_introl _ (refl_equal _)) 
                 (or_introl _ (refl_equal _)) H3) as [H1 _]; rewrite H1; clear H1.
destruct (IHl1s2 t1 (or_introl _ (refl_equal _)) 
                  (H0 t1 (or_introl _ (refl_equal _)))) as [H1 _]; rewrite H1; clear H1.
assert (H' : list_forall_option
    (fun t3 : term =>
     match rpo_eval n t3 s2 with
     | Some Equivalent => Some false
     | Some Less_than => Some true
     | Some Greater_than => Some false
     | Some Uncomparable => Some false
     | None => None (A:=bool)
     end) l1 = Some true).
apply list_forall_option_is_complete_true.
intros u1 u1_in_l1; 
destruct (IHl1s2 u1 (or_intror _ u1_in_l1) 
                  (H0 u1 (or_intror _ u1_in_l1))) as [H1 _]; rewrite H1; trivial.
rewrite H'; trivial.
simpl; assert (H' := rpo_eval_is_complete_equivalent n t1 t2).
generalize (H' (Sl _ _ (or_introl _ (refl_equal _)) 
                        (or_introl _ (refl_equal _))) H3);
clear H'; intro H'; rewrite H'.
apply IHl1; trivial.
injection L; intros; trivial.
intros; apply IHl1l2; trivial; right; trivial.
intros; apply IHl2l1; trivial; right; trivial.
intros; apply Sl; right; trivial.
intros; apply H0; right; trivial.
intros; apply IHl1s2; trivial; right; trivial.
intros; apply IHs2l1; trivial; right; trivial.
intros; apply IHs1l2; trivial; right; trivial.
intros; apply IHl2s1; trivial; right; trivial.
simpl in R; rewrite H' in R; trivial.
simpl in T; rewrite H' in T; trivial.
intro; assert False; [apply (prec_antisym f2); trivial | contradiction].
intro; assert False; [apply (prec_antisym f2); trivial | contradiction].
intros [f2_diff_f2 _]; absurd (f2 = f2); trivial.
simpl in T; absurd (@None comp = None); trivial.
simpl in T; absurd (@None comp = None); trivial.
destruct (list_exists_option
     (fun t : term =>
      match rpo_eval n t (Term f2 l1) with
      | Some Equivalent => Some true
      | Some Less_than => Some false
      | Some Greater_than => Some true
      | Some Uncomparable => Some false
      | None => None (A:=bool)
      end) l2) as [ [ | ] | ].
trivial.
destruct (list_exists_option
         (fun t : term =>
          match rpo_eval n (Term f2 l2) t with
          | Some Equivalent => Some true
          | Some Less_than => Some true
          | Some Greater_than => Some false
          | Some Uncomparable => Some false
          | None => None (A:=bool)
          end) l1) as [ [ | ] | ].
simpl in R'.
destruct (rpo_closure (Term f2 l1) (Term f2 l2) (Term f2 l2)) as [_ [Antisym _]].
assert False; [apply Antisym; trivial | contradiction].
simpl; generalize (prec_eval_is_sound f2 f2); destruct (prec_eval f2 f2).
intros _; rewrite Sf; rewrite Sf in R'; rewrite Sf in T'; simpl in R'; simpl in T'.
assert (L : length l1 = length l2).
destruct W1 as [_ L1]; destruct W2 as [_ L2];
destruct (F.arity f2) as [ | | m]; rewrite L1; rewrite L2; trivial.
generalize (Term f2 l1) (Term f2 l2) t1_lt_t2 H0 St 
IHl1s2 IHs2l1 IHs1l2 IHl2s1 R' T'.
clear W1 W2 H0 St IHl1s2 IHs2l1 IHs1l2 IHl2s1 R T R' T'.
generalize l2 L l1_lt_l2 IHl1l2 IHl2l1 Sl; 
clear l2 L l1_lt_l2 IHl1l2 IHl2l1 Sl Wl1 Wl2 t1_lt_t2.
induction l1 as [ | t1 l1]; intros [ | t2 l2] L l1_lt_l2 IHl1l2 IHl2l1 Sl;
intros s1 s2 s1_lt_s2 H0 St IHl1s2 IHs2l1 IHs1l2 IHl2s1 R' T'.
simpl in R'; absurd (rpo s1 s1).
intro; apply (rpo_antirefl s1); trivial.
setoid_rewrite (equiv_rpo_equiv_1 _ _ R') in s1_lt_s2; trivial.
discriminate.
discriminate.
inversion l1_lt_l2; subst.
simpl; destruct (IHl1l2 t1 t2 (or_introl _ (refl_equal _)) 
                 (or_introl _ (refl_equal _)) H3) as [_ H2]; rewrite H2; clear H2.
destruct (IHl1s2 t1 (or_introl _ (refl_equal _)) 
                  (H0 t1 (or_introl _ (refl_equal _)))) as [_ H2]; rewrite H2; clear H2.
assert (H' : list_forall_option
    (fun t3 : term =>
     match rpo_eval n s2 t3 with
     | Some Equivalent => Some false
     | Some Less_than => Some false
     | Some Greater_than => Some true
     | Some Uncomparable => Some false
     | None => None (A:=bool)
     end) l1 = Some true).
apply list_forall_option_is_complete_true.
intros u1 u1_in_l1; 
destruct (IHl1s2 u1 (or_intror _ u1_in_l1) 
                  (H0 u1 (or_intror _ u1_in_l1))) as [_ H2]; rewrite H2; trivial.
rewrite H'; trivial.
simpl; assert (H' := rpo_eval_is_complete_equivalent n t2 t1).
rewrite plus_comm in H';
generalize (H' (Sl _ _ (or_introl _ (refl_equal _)) 
                        (or_introl _ (refl_equal _))) 
                      (equiv_sym _ _ equiv_equiv _ _ H3));
clear H'; intro H'; rewrite H'.
apply IHl1; trivial.
injection L; intros; trivial.
intros; apply IHl1l2; trivial; right; trivial.
intros; apply IHl2l1; trivial; right; trivial.
intros; apply Sl; right; trivial.
intros; apply H0; right; trivial.
intros; apply IHl1s2; trivial; right; trivial.
intros; apply IHs2l1; trivial; right; trivial.
intros; apply IHs1l2; trivial; right; trivial.
intros; apply IHl2s1; trivial; right; trivial.
simpl in R'; rewrite H' in R'; trivial.
simpl in T'; rewrite H' in T'; trivial.
intro; assert False; [apply (prec_antisym f2); trivial | contradiction].
intro; assert False; [apply (prec_antisym f2); trivial | contradiction].
intros [f2_diff_f2 _]; absurd (f2 = f2); trivial.
simpl in T; absurd (@None comp = None); trivial.
simpl in T; absurd (@None comp = None); trivial.
(* 1/3 Top_eq_mul *)
split.
destruct (list_exists_option
     (fun t : term =>
      match rpo_eval n t (Term f2 l2) with
      | Some Equivalent => Some true
      | Some Less_than => Some false
      | Some Greater_than => Some true
      | Some Uncomparable => Some false
      | None => None (A:=bool)
      end) l1) as [ [ | ] | ].
simpl in R.
destruct (rpo_closure (Term f2 l1) (Term f2 l2) (Term f2 l2)) as [_ [Antisym _]].
assert False; [apply Antisym; trivial | contradiction].
destruct (list_exists_option
         (fun t : term =>
          match rpo_eval n (Term f2 l1) t with
          | Some Equivalent => Some true
          | Some Less_than => Some true
          | Some Greater_than => Some false
          | Some Uncomparable => Some false
          | None => None (A:=bool)
          end) l2) as [ [ | ] | ].
trivial.
simpl; generalize (prec_eval_is_sound f2 f2); destruct (prec_eval f2 f2).
intros _; rewrite Sf; rewrite Sf in R; rewrite Sf in T; simpl in R; simpl in T.
case_eq (remove_equiv_list (equiv_eval n) l1 l2).
intros [[ | t1' l1'] [ | t2' l2']] H.
rewrite H in R; absurd (rpo (Term f2 l1) (Term f2 l1)).
intro; apply (rpo_antirefl (Term f2 l1)); trivial.
setoid_rewrite <- (equiv_rpo_equiv_1 _ _ R) in t1_lt_t2; trivial.
trivial.
rewrite H in R; absurd (rpo (Term f2 l1) (Term f2 l1)).
intro; apply (rpo_antirefl (Term f2 l1)); trivial.
apply rpo_trans with (Term f2 l2); trivial.
assert (H' := remove_equiv_list_is_sound (equiv_eval n) l1 l2);
rewrite H in H'; destruct H' as [ll [E_ll [P1 [P2 F]]]].
assert (H' : rpo_mul (t1' :: l1') (t2' :: l2')).
assert (E_ll' : forall t1 t2, In (t1,t2) ll -> equiv t1 t2).
intros t1 t2 t1t2_in_ll; apply (equiv_eval_is_sound_weak n); 
apply E_ll; trivial.
generalize ll l1 l2 t1' l1' t2' l2' l1_lt_l2 E_ll' P1 P2; 
clear t1' l1' t2' l2' l1_lt_l2 ll E_ll' E_ll P1 P2 F H;
induction ll as [ | [u1 u2] ll]; intros k1 k2 t1' l1' t2' l2' k1_lt_k2 E_ll P1 P2.
inversion k1_lt_k2; subst.
apply (List_mul a lg ls llc); trivial.
simpl in P1; rewrite <- app_nil_end in P1; setoid_rewrite <- P1; trivial.
simpl in P2; rewrite <- app_nil_end in P2; setoid_rewrite <- P2; trivial.
apply (rpo_mul_remove_equiv_aux (t1' :: l1') (t2' :: l2') u1 u2).
intros t _ t_lt_t; apply (rpo_antirefl t); trivial.
apply E_ll; left; trivial.
apply (IHll k1 k2 u1 (t1' :: l1') u2 (t2' :: l2') k1_lt_k2).
intros; apply E_ll; right; trivial.
setoid_rewrite P1; simpl; rewrite app_comm_cons;
apply list_permut_sym; setoid_rewrite <- permut_cons_inside;
simpl; auto.
setoid_rewrite P2; simpl; rewrite app_comm_cons;
apply list_permut_sym; setoid_rewrite <- permut_cons_inside;
simpl; auto.
rewrite H in R; rewrite H in T;
unfold mult_eval; unfold mult_eval in R; unfold mult_eval in T;
destruct (list_forall_option
          (fun t2 : term =>
           list_exists_option
             (fun t1 : term =>
              match rpo_eval n t1 t2 with
              | Some Equivalent => Some false
              | Some Less_than => Some false
              | Some Greater_than => Some true
              | Some Uncomparable => Some false
              | None => None (A:=bool)
              end) (t1' :: l1')) (t2' :: l2')) as [ [ | ] | ].
absurd (rpo (Term f2 l1) (Term f2 l1)).
intro; apply (rpo_antirefl (Term f2 l1)); trivial.
apply rpo_trans with (Term f2 l2); trivial.
assert (list_forall_option
    (fun t1 : term =>
     list_exists_option
       (fun t2 : term =>
        match rpo_eval n t1 t2 with
        | Some Equivalent => Some false
        | Some Less_than => Some true
        | Some Greater_than => Some false
        | Some Uncomparable => Some false
        | None => None (A:=bool)
        end) (t2' :: l2')) (t1' :: l1') = Some true).
apply list_forall_option_is_complete_true.
intros u1 u1_in_l1'; apply list_exists_option_is_complete_true.
inversion H'; subst.
setoid_rewrite (in_permut_in u1 H1) in u1_in_l1'.
destruct (in_app_or _ _ _ u1_in_l1') as [u1_in_ls | u1_in_llc].
destruct (H3 _ u1_in_ls) as [a' [a'_in_lg u1_lt_a']]; exists a'; split.
setoid_rewrite (in_permut_in a' H2); rewrite app_comm_cons;
apply in_or_app; left; trivial.
destruct (IHl1l2 u1 a') as [H1' _]; trivial.
setoid_rewrite (in_permut_in u1 P1); apply in_or_app; left; 
setoid_rewrite (in_permut_in u1 H1); apply in_or_app; left; trivial.
setoid_rewrite (in_permut_in a' P2); apply in_or_app; left; 
rewrite app_comm_cons in H2;
setoid_rewrite (in_permut_in a' H2); apply in_or_app; left; trivial.
rewrite H1'; trivial.
destruct (equiv_in_list _ equiv u1 llc H0 u1_in_llc) as 
        [u2 [ll1 [ll2 [u1_eq_u2 H4]]]].
setoid_rewrite <- (in_permut_in u1 H1) in u1_in_l1'.
assert (u2_in_l2' : In u2 (t2' :: l2')).
setoid_rewrite (in_permut_in u2 H2); rewrite app_comm_cons;
apply in_or_app; right; setoid_rewrite in_map_iff;
exists (u1,u2); simpl; split; trivial; subst; apply in_or_app; right; left; trivial.
generalize (equiv_eval_is_sound n u1 u2);
rewrite (F u1 u2 u1_in_l1' u2_in_l2');
intro; absurd (equiv u1 u2); trivial.
apply H5; apply le_S_n; refine (le_trans _ _ _ _ St);
replace (S (size u1 + size u2)) with (S (size u1) + size u2); trivial;
rewrite plus_comm; apply plus_le_compat.
apply lt_le_weak; apply size_direct_subterm; 
setoid_rewrite (in_permut_in u2 P2); apply in_or_app; left; trivial. 
apply size_direct_subterm; 
setoid_rewrite (in_permut_in u1 P1); apply in_or_app; left; trivial.
rewrite H0; trivial.
assert False; [apply T; trivial | contradiction].
intro H; rewrite H in T; assert False; [apply T; trivial | contradiction].
intro; assert False; [apply (prec_antisym f2); trivial | contradiction].
intro; assert False; [apply (prec_antisym f2); trivial | contradiction].
intros [f2_diff_f2 _]; absurd (f2 = f2); trivial.
assert False; [apply T; trivial | contradiction].
assert False; [apply T; trivial | contradiction].
destruct (list_exists_option
     (fun t : term =>
      match rpo_eval n t (Term f2 l1) with
      | Some Equivalent => Some true
      | Some Less_than => Some false
      | Some Greater_than => Some true
      | Some Uncomparable => Some false
      | None => None (A:=bool)
      end) l2) as [ [ | ] | ].
trivial.
destruct (list_exists_option
         (fun t : term =>
          match rpo_eval n (Term f2 l2) t with
          | Some Equivalent => Some true
          | Some Less_than => Some true
          | Some Greater_than => Some false
          | Some Uncomparable => Some false
          | None => None (A:=bool)
          end) l1) as [ [ | ] | ].
simpl in R.
destruct (rpo_closure (Term f2 l1) (Term f2 l2) (Term f2 l2)) as [_ [Antisym _]].
assert False; [apply Antisym; trivial | contradiction].
simpl; generalize (prec_eval_is_sound f2 f2); destruct (prec_eval f2 f2).
intros _; rewrite Sf; rewrite Sf in R'; rewrite Sf in T'; simpl in R'; simpl in T'.
case_eq (remove_equiv_list (equiv_eval n) l2 l1).
intros [[ | t2' l2'] [ | t1' l1']] H.
rewrite H in R'; absurd (rpo (Term f2 l1) (Term f2 l1)).
intro; apply (rpo_antirefl (Term f2 l1)); trivial.
setoid_rewrite (equiv_rpo_equiv_1 _ _ R') in t1_lt_t2; trivial.
rewrite H in R'; absurd (rpo (Term f2 l1) (Term f2 l1)).
intro; apply (rpo_antirefl (Term f2 l1)); trivial.
apply rpo_trans with (Term f2 l2); trivial.
trivial.
assert (H' := remove_equiv_list_is_sound (equiv_eval n) l2 l1);
rewrite H in H'; destruct H' as [ll [E_ll [P2 [P1 F]]]].
assert (H' : rpo_mul (t1' :: l1') (t2' :: l2')).
assert (E_ll' : forall t1 t2, In (t1,t2) ll -> equiv t1 t2).
intros t1 t2 t1t2_in_ll; apply (equiv_eval_is_sound_weak n); 
apply E_ll; trivial.
generalize ll l1 l2 t1' l1' t2' l2' l1_lt_l2 E_ll' P1 P2; 
clear t1' l1' t2' l2' l1_lt_l2 ll E_ll' E_ll P1 P2 F H;
induction ll as [ | [u1 u2] ll]; intros k1 k2 t1' l1' t2' l2' k1_lt_k2 E_ll P1 P2.
inversion k1_lt_k2; subst.
apply (List_mul a lg ls llc); trivial.
simpl in P1; rewrite <- app_nil_end in P1; setoid_rewrite <- P1; trivial.
simpl in P2; rewrite <- app_nil_end in P2; setoid_rewrite <- P2; trivial.
apply (rpo_mul_remove_equiv_aux (t1' :: l1') (t2' :: l2') u2 u1).
intros t _ t_lt_t; apply (rpo_antirefl t); trivial.
apply (equiv_sym _ _ equiv_equiv); apply E_ll; left; trivial.
apply (IHll k1 k2 u2 (t1' :: l1') u1 (t2' :: l2') k1_lt_k2).
intros; apply E_ll; right; trivial.
setoid_rewrite P1; simpl; rewrite app_comm_cons; 
apply list_permut_sym; setoid_rewrite <- permut_cons_inside;
simpl; auto.
setoid_rewrite P2; simpl; rewrite app_comm_cons;
apply list_permut_sym; setoid_rewrite <- permut_cons_inside;
simpl; auto.
rewrite H in R'; rewrite H in T';
unfold mult_eval; unfold mult_eval in R'; unfold mult_eval in T'.
assert (list_forall_option
    (fun t2 : term =>
     list_exists_option
       (fun t1 : term =>
        match rpo_eval n t1 t2 with
        | Some Equivalent => Some false
        | Some Less_than => Some false
        | Some Greater_than => Some true
        | Some Uncomparable => Some false
        | None => None (A:=bool)
        end) (t2' :: l2')) (t1' :: l1') = Some true).
apply list_forall_option_is_complete_true.
intros u1 u1_in_l1'; apply list_exists_option_is_complete_true.
inversion H'; subst.
setoid_rewrite (in_permut_in u1 H1) in u1_in_l1'.
destruct (in_app_or _ _ _ u1_in_l1') as [u1_in_ls | u1_in_llc].
destruct (H3 _ u1_in_ls) as [a' [a'_in_lg u1_lt_a']]; exists a'; split.
setoid_rewrite (in_permut_in a' H2); rewrite app_comm_cons;
apply in_or_app; left; trivial.
destruct (IHl1l2 u1 a') as [_ H2']; trivial.
setoid_rewrite (in_permut_in u1 P1); apply in_or_app; left; 
setoid_rewrite (in_permut_in u1 H1); apply in_or_app; left; trivial.
setoid_rewrite (in_permut_in a' P2); apply in_or_app; left; 
rewrite app_comm_cons in H2;
setoid_rewrite (in_permut_in a' H2); apply in_or_app; left; trivial.
rewrite H2'; trivial.
destruct (equiv_in_list _ equiv u1 llc H0 u1_in_llc) as 
        [u2 [ll1 [ll2 [u1_eq_u2 H4]]]].
setoid_rewrite <- (in_permut_in u1 H1) in u1_in_l1'.
assert (u2_in_l2' : In u2 (t2' :: l2')).
setoid_rewrite (in_permut_in u2 H2); rewrite app_comm_cons;
apply in_or_app; right; setoid_rewrite in_map_iff;
exists (u1,u2); simpl; split; trivial; subst; apply in_or_app; right; left; trivial.
generalize (equiv_eval_is_sound n u2 u1);
rewrite (F u2 u1 u2_in_l2' u1_in_l1');
intro; absurd (equiv u2 u1); trivial.
apply H5; apply le_S_n; refine (le_trans _ _ _ _ St);
replace (S (size u2 + size u1)) with (S (size u2) + size u1); trivial;
apply plus_le_compat.
apply size_direct_subterm; 
setoid_rewrite (in_permut_in u2 P2); apply in_or_app; left; trivial. 
apply lt_le_weak; apply size_direct_subterm; 
setoid_rewrite (in_permut_in u1 P1); apply in_or_app; left; trivial.
apply (equiv_sym _ _ equiv_equiv); trivial.
rewrite H0; trivial.
intro H; rewrite H in T'; assert False; [apply T'; trivial | contradiction].
intro; assert False; [apply (prec_antisym f2); trivial | contradiction].
intro; assert False; [apply (prec_antisym f2); trivial | contradiction].
intros [f2_diff_f2 _]; absurd (f2 = f2); trivial.
assert False; [apply T'; trivial | contradiction].
assert False; [apply T'; trivial | contradiction].
assert False; [apply T'; trivial | contradiction].
assert False; [apply T; trivial | contradiction].
Qed.

Lemma rpo_eval_is_complete :
  forall n t1 t2, size t1 + size t2 <= n -> 
        well_formed t1 -> well_formed t2 -> 
        match rpo_eval n t1 t2 with
        | Some Equivalent => equiv t1 t2
        | Some Less_than => rpo t1 t2
        | Some Greater_than => rpo t2 t1
        | Some Uncomparable => ~equiv t1 t2 /\ ~ rpo t1 t2 /\ ~rpo t2 t1
        | None => False
        end.
intros n t1 t2 St W1 W2;
generalize (rpo_eval_is_sound_weak n t1 t2 W1 W2)
(rpo_eval_is_complete_equivalent n t1 t2 St)
(rpo_eval_is_complete_less_greater n t1 t2 St W1 W2)
(rpo_eval_terminates n t1 t2 St);
rewrite plus_comm in St;
generalize (rpo_eval_is_complete_less_greater n t2 t1 St W2 W1).
destruct (rpo_eval n t1 t2) as [ [ | | | ] | ]; trivial.
intros not_t2_lt_t1 _ t1_diff_t2 not_t1_lt_t2 T; repeat split; intro H.
generalize (t1_diff_t2 H); discriminate.
destruct (not_t1_lt_t2 H); discriminate.
destruct (not_t2_lt_t1 H); discriminate.
intros _ _ _ _ H; apply H; trivial.
Qed.

End Make.








